@@ -1,5 +1,6 @@
 
-from GridCalEngine.IO.cim.cim16.cim_parser import CIMImport, CIMExport
+
+from GridCalEngine.IO.cim import cgmesProfile, CIMImport, CIMExport
 from GridCalEngine.IO.dgs.dgs_parser import dgs_to_circuit
 from GridCalEngine.IO.others.dpx_parser import load_dpx
 from GridCalEngine.IO.others.ipa_parser import load_iPA
@@ -7,4 +8,4 @@ from GridCalEngine.IO.gridcal.json_parser import save_json_file_v3, parse_json_d
 from GridCalEngine.IO.matpower.matpower_parser import parse_matpower_file, get_matpower_case_data
 from GridCalEngine.IO.raw.raw_parser_legacy import PSSeParser
 from GridCalEngine.IO.gridcal.excel_interface import interpret_excel_v3, interprete_excel_v2
-from GridCalEngine.IO.file_handler import FileOpen, FileSave
+from GridCalEngine.IO.file_handler import FileOpen, FileSave, FileSavingOptions
