@@ -1,5 +1,5 @@
 # GridCal
-# Copyright (C) 2015 - 2023 Santiago Peñate Vera
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
 # 
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -21,9 +21,10 @@ from typing import List, AnyStr, Dict
 
 import numpy as np
 
-from GridCalEngine.Core.Devices.multi_circuit import MultiCircuit
+from GridCalEngine.Devices.multi_circuit import MultiCircuit
+from GridCalEngine.enumerations import DeviceType
 from GridCalEngine.basic_structures import Logger
-import GridCalEngine.Core.Devices as dev
+import GridCalEngine.Devices as dev
 
 
 def interpret_line(line, splitter=','):
@@ -232,7 +233,7 @@ def parse_buses(data_lst: List[List], substations_dict: Dict[int, dev.Substation
         buses_dict[code] = dev.Bus(name=raw[1],
                                    idtag=None,
                                    code=code,
-                                   vnom=raw[2],
+                                   Vnom=raw[2],
                                    vmin=0.9,
                                    vmax=1.1,
                                    angle_min=-6.28, angle_max=6.28, r_fault=0.0, x_fault=0.0,
@@ -274,7 +275,7 @@ def parse_dc_buses(data_lst: List[List]):
         buses_dict[code] = dev.Bus(name=raw[1],
                                    idtag=None,
                                    code=code,
-                                   vnom=raw[7],
+                                   Vnom=raw[7],
                                    vmin=0.9,
                                    vmax=1.1,
                                    angle_min=-6.28, angle_max=6.28, r_fault=0.0, x_fault=0.0,
@@ -322,6 +323,7 @@ def parse_transformers(data_lst: List[List], buses_dict: Dict[int, dev.Bus]):
                                 g=1e-20,
                                 b=1e-20,
                                 rate=rate,
+                                nominal_power=rate,
                                 tap_module=1.0,
                                 tap_module_max=1.2,
                                 tap_module_min=0.5,
@@ -449,8 +451,7 @@ def parse_loads(data_lst: List[List], buses_dict: Dict[int, dev.Bus]):
                        P=P,
                        Q=Q,
                        active=True,)
-        bus_f.add_device(elm)
-
+        elm.bus = bus_f
         data.append(elm)
 
     return data
@@ -490,7 +491,7 @@ def parse_generators(data_lst: List[List], buses_dict: Dict[int, dev.Bus], bus_v
                             Snom=np.sqrt(Pmax*Pmax+Qmax*Qmax),
                             Sbase=Sbase,
                             active=True, )
-        bus_f.add_device(elm)
+        elm.bus = bus_f
 
         data.append(elm)
 
@@ -632,30 +633,37 @@ class PowerWorldParser:
         buses_dict, bus_volt = parse_buses(data_dict['bus data']['data'], substations_dict, area_dict, zone_dict)
 
         # create devices
-        grid.buses = list(buses_dict.values())
+        grid.set_elements_list_by_type(device_type=DeviceType.BusDevice, devices=list(buses_dict.values()))
 
         if 'branch data' in data_dict.keys():
-            grid.lines = parse_branches(data_dict['branch data']['data'], buses_dict)
+            lines = parse_branches(data_dict['branch data']['data'], buses_dict)
+            grid.set_elements_list_by_type(device_type=DeviceType.LineDevice, devices=lines)
 
         if 'transformer data' in data_dict.keys():
-            grid.transformers2w = parse_transformers(data_dict['transformer data']['data'], buses_dict)
+            transformers2w = parse_transformers(data_dict['transformer data']['data'], buses_dict)
+            grid.set_elements_list_by_type(device_type=DeviceType.Transformer2WDevice, devices=transformers2w)
 
         if 'load data' in data_dict.keys():
-            parse_loads(data_dict['load data']['data'], buses_dict)
+            loads = parse_loads(data_dict['load data']['data'], buses_dict)
+            grid.set_elements_list_by_type(device_type=DeviceType.LoadDevice, devices=loads)
 
         if 'generator data' in data_dict.keys():
-            parse_generators(data_dict['generator data']['data'], buses_dict, bus_volt)
+            generators = parse_generators(data_dict['generator data']['data'], buses_dict, bus_volt)
+            grid.set_elements_list_by_type(device_type=DeviceType.GeneratorDevice, devices=generators)
 
         if 'dc bus data' in data_dict.keys():
             # augments buses_dict and bus_volt
             dc_buses_dict, dc_bus_volt = parse_dc_buses(data_dict['dc bus data']['data'])
-            grid.buses += list(dc_buses_dict.values())
+            for elm in dc_buses_dict.values():
+                grid.add_bus(elm)
 
             if 'dc line data' in data_dict.keys():
-                grid.dc_lines = parse_dc_lines(data_dict['dc line data']['data'], dc_buses_dict)
+                dc_lines = parse_dc_lines(data_dict['dc line data']['data'], dc_buses_dict)
+                grid.set_elements_list_by_type(device_type=DeviceType.DCLineDevice, devices=dc_lines)
 
             if 'dc converter data' in data_dict.keys():
-                grid.vsc_devices = parse_dc_converters(data_dict['dc converter data']['data'], buses_dict, dc_buses_dict)
+                vsc_devices = parse_dc_converters(data_dict['dc converter data']['data'], buses_dict, dc_buses_dict)
+                grid.set_elements_list_by_type(device_type=DeviceType.VscDevice, devices=vsc_devices)
 
         logger += grid.fill_xy_from_lat_lon()
 
