@@ -1,6 +1,6 @@
 """
 GridCal
-# Copyright (C) 2015 - 2023 Santiago Peñate Vera
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
 # 
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -18,9 +18,12 @@ GridCal
 """
 from typing import Dict, Tuple, List, Union
 import numpy as np
-import GridCalEngine.basic_structures as bs
-from GridCalEngine.Core.Devices.multi_circuit import MultiCircuit
-import GridCalEngine.Core.Devices as dev
+import pandas as pd
+
+from GridCalEngine.basic_structures import Logger
+from GridCalEngine.enumerations import ConverterControlType
+from GridCalEngine.Devices.multi_circuit import MultiCircuit
+import GridCalEngine.Devices as dev
 import GridCalEngine.IO.matpower.matpower_branch_definitions as matpower_branches
 import GridCalEngine.IO.matpower.matpower_bus_definitions as matpower_buses
 import GridCalEngine.IO.matpower.matpower_gen_definitions as matpower_gen
@@ -45,16 +48,13 @@ def find_between(s, first, last):
         return ""
 
 
-def txt2mat(txt, line_splitter=';', col_splitter='\t', to_float=True):
+def txt2mat(txt: str, line_splitter=';', to_float=True):
     """
 
-    Args:
-        txt:
-        line_splitter:
-        col_splitter:
-
-    Returns:
-
+    :param txt:
+    :param line_splitter:
+    :param to_float:
+    :return:
     """
     lines = txt.strip().split('\n')
     # del lines[-1]
@@ -73,7 +73,7 @@ def txt2mat(txt, line_splitter=';', col_splitter='\t', to_float=True):
     for i, line in enumerate(lines2):
 
         if ';' in line:
-            line2 = line.split(';')[0]
+            line2 = line.split(line_splitter)[0]
         else:
             line2 = line
 
@@ -85,7 +85,7 @@ def txt2mat(txt, line_splitter=';', col_splitter='\t', to_float=True):
             if to_float:
                 arr = np.zeros((nrows, ncols))
             else:
-                arr = np.zeros((nrows, ncols), dtype=np.object)
+                arr = np.zeros((nrows, ncols), dtype=object)
 
         # fill-in the data
         for j, val in enumerate(vec):
@@ -97,7 +97,9 @@ def txt2mat(txt, line_splitter=';', col_splitter='\t', to_float=True):
     return np.array(arr)
 
 
-def parse_areas_data(circuit: MultiCircuit, data, logger: bs.Logger):
+def parse_areas_data(circuit: MultiCircuit,
+                     data: Dict[str, np.ndarray],
+                     logger: Logger):
     """
     Parse Matpower / FUBM Matpower area data into GridCal
     :param circuit: MultiCircuit instance
@@ -127,7 +129,10 @@ def parse_areas_data(circuit: MultiCircuit, data, logger: bs.Logger):
     return area_idx_dict
 
 
-def parse_buses_data(circuit: MultiCircuit, data, area_idx_dict, logger: bs.Logger):
+def parse_buses_data(circuit: MultiCircuit,
+                     data: Dict[str, np.ndarray],
+                     area_idx_dict,
+                     logger: Logger):
     """
     Parse Matpower / FUBM Matpower bus data into GridCal
     :param circuit: MultiCircuit instance
@@ -169,13 +174,13 @@ def parse_buses_data(circuit: MultiCircuit, data, area_idx_dict, logger: bs.Logg
             if ref_idx == bus_idx:
                 is_slack = True
         else:
-            area = circuit.default_area
+            area = None
 
         code = str(bus_idx)
 
         bus = dev.Bus(name=names[i],
                       code=code,
-                      vnom=table[i, matpower_buses.BASE_KV],
+                      Vnom=table[i, matpower_buses.BASE_KV],
                       vmax=table[i, matpower_buses.VMAX],
                       vmin=table[i, matpower_buses.VMIN],
                       area=area,
@@ -189,25 +194,26 @@ def parse_buses_data(circuit: MultiCircuit, data, area_idx_dict, logger: bs.Logg
         # determine if the bus is set as slack manually
         bus.is_slack = table[i, matpower_buses.BUS_TYPE] == matpower_buses.REF
 
+        # Add the bus to the circuit buses
+        circuit.add_bus(bus)
+
         # Add the load
         if table[i, matpower_buses.PD] != 0 or table[i, matpower_buses.QD] != 0:
             load = dev.Load(P=table[i, matpower_buses.PD], Q=table[i, matpower_buses.QD])
-            load.bus = bus
-            bus.loads.append(load)
+            circuit.add_load(bus=bus, api_obj=load)
 
         # Add the shunt
         if table[i, matpower_buses.GS] != 0 or table[i, matpower_buses.BS] != 0:
             shunt = dev.Shunt(G=table[i, matpower_buses.GS], B=table[i, matpower_buses.BS])
-            shunt.bus = bus
-            bus.shunts.append(shunt)
-
-        # Add the bus to the circuit buses
-        circuit.add_bus(bus)
+            circuit.add_shunt(bus=bus, api_obj=shunt)
 
     return bus_idx_dict
 
 
-def parse_generators(circuit: MultiCircuit, data, bus_idx_dict, logger: bs.Logger):
+def parse_generators(circuit: MultiCircuit,
+                     data: Dict[str, np.ndarray],
+                     bus_idx_dict,
+                     logger: Logger):
     """
     Parse Matpower / FUBM Matpower generator data into GridCal
     :param circuit: MultiCircuit instance
@@ -219,7 +225,7 @@ def parse_generators(circuit: MultiCircuit, data, bus_idx_dict, logger: bs.Logge
 
     # Generators
     table = data['gen']
-    n = len(table)
+    n = table.shape[0]
     # load profiles
     if 'Gprof' in data.keys():
         Gprof = data['Gprof']
@@ -234,23 +240,23 @@ def parse_generators(circuit: MultiCircuit, data, bus_idx_dict, logger: bs.Logge
         names = ['gen ' + str(i) for i in range(n)]
 
     gen_dict = dict()
-    for i in range(len(table)):
+    for i in range(n):
         bus_idx = bus_idx_dict[int(table[i, matpower_gen.GEN_BUS])]
         # TODO: Calculate pf based on reactive_power
         gen = dev.Generator(name=names[i],
-                            P=table[i, matpower_gen.PG],
-                            vset=table[i, matpower_gen.VG],
-                            Qmax=table[i, matpower_gen.QMAX],
-                            Qmin=table[i, matpower_gen.QMIN],
-                            Pmin=table[i, matpower_gen.PMIN],
-                            Pmax=table[i, matpower_gen.PMAX]
+                            P=float(table[i, matpower_gen.PG]),
+                            vset=float(table[i, matpower_gen.VG]),
+                            Qmax=float(table[i, matpower_gen.QMAX]),
+                            Qmin=float(table[i, matpower_gen.QMIN]),
+                            Pmin=float(table[i, matpower_gen.PMIN]),
+                            Pmax=float(table[i, matpower_gen.PMAX])
                             )
 
         gen_dict[i] = gen
 
         # Add the generator to the bus
         gen.bus = circuit.buses[bus_idx]
-        circuit.buses[bus_idx].generators.append(gen)
+        circuit.add_generator(bus=circuit.buses[bus_idx], api_obj=gen)
 
     if 'gencost' in data:
         # parse the OPF data
@@ -258,15 +264,15 @@ def parse_generators(circuit: MultiCircuit, data, bus_idx_dict, logger: bs.Logge
 
         for i in range(opf_table.shape[0]):
             curve_model = opf_table[i, 0]
-            startup_cost = opf_table[i, 1]
-            shutdown_cost = opf_table[i, 2]
+            gen_dict[i].StartupCost = opf_table[i, 1]
+            gen_dict[i].ShutdownCost = opf_table[i, 2]
             n_cost = opf_table[i, 3]
             points = opf_table[i, 4:]
             if curve_model == 2:
                 if len(points) > 1:
-                    gen_dict[i].Cost0 = points[0]
+                    gen_dict[i].Cost0 = points[2]
                     gen_dict[i].Cost = points[1]
-                    gen_dict[i].Cost2 = points[2]
+                    gen_dict[i].Cost2 = points[0]
             elif curve_model == 1:
                 # fit a quadratic curve
                 x = points[0::1]
@@ -280,7 +286,10 @@ def parse_generators(circuit: MultiCircuit, data, bus_idx_dict, logger: bs.Logge
             #     gen_dict[i].enabled_dispatch = False
 
 
-def parse_branches_data(circuit: MultiCircuit, data, bus_idx_dict, logger: bs.Logger):
+def parse_branches_data(circuit: MultiCircuit,
+                        data: Dict[str, np.ndarray],
+                        bus_idx_dict,
+                        logger: Logger):
     """
     Parse Matpower / FUBM Matpower branch data into GridCal
     :param circuit: MultiCircuit instance
@@ -292,7 +301,7 @@ def parse_branches_data(circuit: MultiCircuit, data, bus_idx_dict, logger: bs.Lo
 
     # Branches
     table = data['branch']
-    n = len(table)
+    n = table.shape[0]
 
     if table.shape[1] == 37:  # FUBM model
         logger.add_info('It is a FUBM model')
@@ -301,9 +310,11 @@ def parse_branches_data(circuit: MultiCircuit, data, bus_idx_dict, logger: bs.Lo
         names = data['branch_names']
     else:
         names = ['branch ' + str(i) for i in range(n)]
-    for i in range(len(table)):
-        f = circuit.buses[bus_idx_dict[int(table[i, matpower_branches.F_BUS])]]
-        t = circuit.buses[bus_idx_dict[int(table[i, matpower_branches.T_BUS])]]
+    for i in range(n):
+        f_idx = int(table[i, matpower_branches.F_BUS])
+        t_idx = int(table[i, matpower_branches.T_BUS])
+        f = circuit.buses[bus_idx_dict[f_idx]]
+        t = circuit.buses[bus_idx_dict[t_idx]]
 
         if table.shape[1] == 37:  # FUBM model
 
@@ -332,37 +343,45 @@ def parse_branches_data(circuit: MultiCircuit, data, bus_idx_dict, logger: bs.Lo
                     if Pfset != 0.0:
 
                         if Qtset != 0.0:
-                            control_mode = dev.ConverterControlType.type_I_2
+                            control_mode = ConverterControlType.type_I_2
 
                         elif Vac_set != 0.0:
-                            control_mode = dev.ConverterControlType.type_I_3
+                            control_mode = ConverterControlType.type_I_3
 
                         else:
-                            control_mode = dev.ConverterControlType.type_I_1
+                            control_mode = ConverterControlType.type_I_1
 
                     else:
-                        control_mode = dev.ConverterControlType.type_0_free
+                        control_mode = ConverterControlType.type_0_free
 
                 elif matpower_converter_mode == 2:
 
                     if Vac_set == 0.0:
-                        control_mode = dev.ConverterControlType.type_II_4
+                        control_mode = ConverterControlType.type_II_4
                     else:
-                        control_mode = dev.ConverterControlType.type_II_5
+                        control_mode = ConverterControlType.type_II_5
 
                 elif matpower_converter_mode == 3:
-                    control_mode = dev.ConverterControlType.type_III_6
+                    control_mode = ConverterControlType.type_III_6
 
                 elif matpower_converter_mode == 4:
-                    control_mode = dev.ConverterControlType.type_III_7
+                    control_mode = ConverterControlType.type_III_7
 
                 else:
-                    control_mode = dev.ConverterControlType.type_0_free
+                    control_mode = ConverterControlType.type_0_free
 
                 rate = max(table[i, [matpower_branches.RATE_A, matpower_branches.RATE_B, matpower_branches.RATE_C]])
 
+                if rate == 0.0:
+                    # in matpower rate=0 means not limited by rating
+                    rate = 10000
+                    monitor_loading = False
+                else:
+                    monitor_loading = True
+
                 branch = dev.VSC(bus_from=f,
                                  bus_to=t,
+                                 code="{0}_{1}_1".format(f_idx, t_idx),
                                  name='VSC' + str(len(circuit.vsc_devices) + 1),
                                  active=bool(table[i, matpower_branches.BR_STATUS]),
                                  r=table[i, matpower_branches.BR_R],
@@ -370,7 +389,7 @@ def parse_branches_data(circuit: MultiCircuit, data, bus_idx_dict, logger: bs.Lo
                                  tap_module=m,
                                  tap_module_max=table[i, matpower_branches.MA_MAX],
                                  tap_module_min=table[i, matpower_branches.MA_MIN],
-                                 tap_phase=table[i, matpower_branches.SHIFT],
+                                 tap_phase=np.deg2rad(table[i, matpower_branches.SHIFT]),  # * np.pi / 180,
                                  tap_phase_max=np.deg2rad(table[i, matpower_branches.ANGMAX]),
                                  tap_phase_min=np.deg2rad(table[i, matpower_branches.ANGMIN]),
                                  G0sw=table[i, matpower_branches.GSW],
@@ -387,8 +406,9 @@ def parse_branches_data(circuit: MultiCircuit, data, bus_idx_dict, logger: bs.Lo
                                  Vdc_set=Vdc_set if Vdc_set > 0 else 1.0,
                                  alpha1=table[i, matpower_branches.ALPHA1],
                                  alpha2=table[i, matpower_branches.ALPHA2],
-                                 alpha3=table[i, matpower_branches.ALPHA3])
-                circuit.add_vsc(branch)
+                                 alpha3=table[i, matpower_branches.ALPHA3],
+                                 monitor_loading=monitor_loading)
+                circuit.add_vsc(obj=branch)
 
                 logger.add_info('Branch as converter', 'Branch {}'.format(str(i + 1)))
 
@@ -399,62 +419,108 @@ def parse_branches_data(circuit: MultiCircuit, data, bus_idx_dict, logger: bs.Lo
                          table[i, matpower_branches.TAP] != 0) or
                         table[i, matpower_branches.SHIFT] != 0.0):
 
+                    rate = table[i, matpower_branches.RATE_A]
+
+                    if rate == 0.0:
+                        # in matpower rate=0 means not limited by rating
+                        rate = 10000
+                        monitor_loading = False
+                    else:
+                        monitor_loading = True
+
                     branch = dev.Transformer2W(bus_from=f,
                                                bus_to=t,
+                                               code="{0}_{1}_1".format(f_idx, t_idx),
                                                name=names[i],
-                                               r=table[i, matpower_branches.BR_R],
-                                               x=table[i, matpower_branches.BR_X],
-                                               g=0,
-                                               b=table[i, matpower_branches.BR_B],
-                                               rate=table[i, matpower_branches.RATE_A],
-                                               tap_module=table[i, matpower_branches.TAP],
-                                               tap_phase=table[i, matpower_branches.SHIFT],
+                                               r=float(table[i, matpower_branches.BR_R]),
+                                               x=float(table[i, matpower_branches.BR_X]),
+                                               g=0.0,
+                                               b=float(table[i, matpower_branches.BR_B]),
+                                               rate=rate,
+                                               monitor_loading=monitor_loading,
+                                               tap_module=float(table[i, matpower_branches.TAP]),
+                                               tap_phase=np.deg2rad(table[i, matpower_branches.SHIFT]),
+                                               # * np.pi / 180,
                                                active=bool(table[i, matpower_branches.BR_STATUS]))
-                    circuit.add_transformer2w(branch)
+                    circuit.add_transformer2w(obj=branch)
                     logger.add_info('Branch as 2w transformer', 'Branch {}'.format(str(i + 1)))
 
                 else:
+                    rate = table[i, matpower_branches.RATE_A]
+
+                    if rate == 0.0:
+                        # in matpower rate=0 means not limited by rating
+                        rate = 10000
+                        monitor_loading = False
+                    else:
+                        monitor_loading = True
+
                     branch = dev.Line(bus_from=f,
                                       bus_to=t,
+                                      code="{0}_{1}_1".format(f_idx, t_idx),
                                       name=names[i],
                                       r=table[i, matpower_branches.BR_R],
                                       x=table[i, matpower_branches.BR_X],
                                       b=table[i, matpower_branches.BR_B],
-                                      rate=table[i, matpower_branches.RATE_A],
+                                      rate=rate,
+                                      monitor_loading=monitor_loading,
                                       active=bool(table[i, matpower_branches.BR_STATUS]))
-                    circuit.add_line(branch, logger=logger)
+                    circuit.add_line(obj=branch, logger=logger)
                     logger.add_info('Branch as line', 'Branch {}'.format(str(i + 1)))
 
         else:
 
-            if f.Vnom != t.Vnom or \
-                    (table[i, matpower_branches.TAP] != 1.0 and table[i, matpower_branches.TAP] != 0) or \
-                    table[i, matpower_branches.SHIFT] != 0.0:
+            if (f.Vnom != t.Vnom or
+                    (table[i, matpower_branches.TAP] != 1.0 and table[i, matpower_branches.TAP] != 0) or
+                    table[i, matpower_branches.SHIFT] != 0.0):
+
+                rate = table[i, matpower_branches.RATE_A]
+
+                if rate == 0.0:
+                    # in matpower rate=0 means not limited by rating
+                    rate = 10000
+                    monitor_loading = False
+                else:
+                    monitor_loading = True
 
                 branch = dev.Transformer2W(bus_from=f,
                                            bus_to=t,
+                                           code="{0}_{1}_1".format(f_idx, t_idx),
                                            name=names[i],
-                                           r=table[i, matpower_branches.BR_R],
-                                           x=table[i, matpower_branches.BR_X],
-                                           g=0,
-                                           b=table[i, matpower_branches.BR_B],
-                                           rate=table[i, matpower_branches.RATE_A],
-                                           tap_module=table[i, matpower_branches.TAP],
-                                           tap_phase=table[i, matpower_branches.SHIFT],
+                                           r=float(table[i, matpower_branches.BR_R]),
+                                           x=float(table[i, matpower_branches.BR_X]),
+                                           g=0.0,
+                                           b=float(table[i, matpower_branches.BR_B]),
+                                           rate=rate,
+                                           monitor_loading=monitor_loading,
+                                           tap_module=float(table[i, matpower_branches.TAP]),
+                                           tap_phase=np.deg2rad(table[i, matpower_branches.SHIFT]),  # * np.pi / 180,
                                            active=bool(table[i, matpower_branches.BR_STATUS]))
-                circuit.add_transformer2w(branch)
+                circuit.add_transformer2w(obj=branch)
                 logger.add_info('Branch as 2w transformer', 'Branch {}'.format(str(i + 1)))
 
             else:
+
+                rate = table[i, matpower_branches.RATE_A]
+
+                if rate == 0.0:
+                    # in matpower rate=0 means not limited by rating
+                    rate = 10000
+                    monitor_loading = False
+                else:
+                    monitor_loading = True
+
                 branch = dev.Line(bus_from=f,
                                   bus_to=t,
+                                  code="{0}_{1}_1".format(f_idx, t_idx),
                                   name=names[i],
                                   r=table[i, matpower_branches.BR_R],
                                   x=table[i, matpower_branches.BR_X],
                                   b=table[i, matpower_branches.BR_B],
-                                  rate=table[i, matpower_branches.RATE_A],
+                                  rate=rate,
+                                  monitor_loading=monitor_loading,
                                   active=bool(table[i, matpower_branches.BR_STATUS]))
-                circuit.add_line(branch, logger=logger)
+                circuit.add_line(obj=branch, logger=logger)
                 logger.add_info('Branch as line', 'Branch {}'.format(str(i + 1)))
 
     # convert normal lines into DC-lines if needed
@@ -463,22 +529,24 @@ def parse_branches_data(circuit: MultiCircuit, data, bus_idx_dict, logger: bs.Lo
         if line.bus_to.is_dc and line.bus_from.is_dc:
             dc_line = dev.DcLine(bus_from=line.bus_from,
                                  bus_to=line.bus_to,
+                                 code=line.code,
                                  name=line.name,
                                  active=line.active,
                                  rate=line.rate,
-                                 r=line.R,
-                                 active_prof=line.active_prof,
-                                 rate_prof=line.rate_prof)
+                                 r=line.R)
+
+            dc_line.active_prof = line.active_prof
+            dc_line.rate_prof = line.rate_prof
 
             # add device to the circuit
-            circuit.add_dc_line(dc_line)
+            circuit.add_dc_line(obj=dc_line)
 
             # delete the line from the circuit
             circuit.delete_line(line)
             logger.add_info('Converted to DC line', line.name)
 
 
-def interpret_data_v1(circuit: MultiCircuit, data, logger: bs.Logger) -> MultiCircuit:
+def interpret_data_v1(circuit: MultiCircuit, data, logger: Logger) -> MultiCircuit:
     """
     Pass the loaded table-like data to the  structures
     :param circuit:
@@ -509,12 +577,12 @@ def interpret_data_v1(circuit: MultiCircuit, data, logger: bs.Logger) -> MultiCi
 
     # add the profiles
     if master_time_array is not None:
-        circuit.format_profiles(master_time_array)
+        circuit.format_profiles(pd.DatetimeIndex(master_time_array))
 
     return circuit
 
 
-def read_matpower_file(filename: str) -> [MultiCircuit, bs.Logger]:
+def read_matpower_file(filename: str) -> [MultiCircuit, Logger]:
     """
 
     :param filename:
@@ -566,7 +634,7 @@ def read_matpower_file(filename: str) -> [MultiCircuit, bs.Logger]:
     return data
 
 
-def parse_matpower_file(filename, export=False) -> [MultiCircuit, bs.Logger]:
+def parse_matpower_file(filename, export=False) -> [MultiCircuit, Logger]:
     """
 
     Args:
@@ -580,7 +648,7 @@ def parse_matpower_file(filename, export=False) -> [MultiCircuit, bs.Logger]:
     # declare circuit
     circuit = MultiCircuit()
 
-    logger = bs.Logger()
+    logger = Logger()
 
     data = read_matpower_file(filename)
 
@@ -610,7 +678,7 @@ def get_matpower_case_data(filename, force_linear_cost=False) -> Dict:
     :param force_linear_cost: Force linear cost when costs are found?
     :return: Matpower case data dictionary
     """
-    logger = bs.Logger()
+    logger = Logger()
 
     data = read_matpower_file(filename)
 
@@ -668,20 +736,23 @@ def get_buses(circuit: MultiCircuit) -> Tuple[List[Dict[str, float]], Dict[dev.B
 
     bus_dict = dict()
 
+    inj_per_bus = circuit.get_injection_devices_grouped_by_bus()
+
     for i, elm in enumerate(circuit.buses):
 
         Pd = 0.0
         Qd = 0.0
         Gs = 0.0
         Bs = 0.0
+        injection_devices = inj_per_bus.get(elm, None)
+        for child in injection_devices:
 
-        for child in (elm.loads + elm.static_generators + elm.external_grids):
-            Pd += child.P
-            Qd += child.Q
-
-        for child in elm.shunts:
-            Gs += child.G
-            Bs += child.B
+            if isinstance(child, dev.Shunt):
+                Gs += child.G
+                Bs += child.B
+            elif isinstance(child, (dev.Load, dev.StaticGenerator)):
+                Pd += child.P
+                Qd += child.Q
 
         data.append({
             'bus_i': i + 1,  # matlab starts at 1
@@ -704,8 +775,9 @@ def get_buses(circuit: MultiCircuit) -> Tuple[List[Dict[str, float]], Dict[dev.B
     return data, bus_dict
 
 
-def get_generation(circuit: MultiCircuit, bus_dict: Dict[dev.Bus, int]) -> Tuple[List[Dict[str, float]],
-                                                                                 List[Dict[str, float]]]:
+def get_generation(
+        circuit: MultiCircuit,
+        bus_dict: Dict[dev.Bus, int]) -> Tuple[List[Dict[str, float]], List[Dict[str, float]]]:
     """
     Get generation and generation cost data
     :param circuit:
@@ -796,7 +868,7 @@ def get_branches(circuit: MultiCircuit, bus_dict: Dict[dev.Bus, int]) -> List[Di
     return data
 
 
-def to_matpower(circuit: MultiCircuit, logger: bs.Logger = bs.Logger()) -> Dict[str, Union[float, List[Dict[str, float]]]]:
+def to_matpower(circuit: MultiCircuit, logger: Logger = Logger()) -> Dict[str, Union[float, List[Dict[str, float]]]]:
     """
 
     :param circuit:
