@@ -1,5 +1,5 @@
 # GridCal
-# Copyright (C) 2015 - 2023 Santiago Peñate Vera
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -21,8 +21,8 @@ import numpy as np
 from enum import Enum
 import zipfile
 from xml.etree import cElementTree as ElementTree
-from GridCalEngine.Core.Devices import Bus, Generator, Branch, Load, BranchType
-from GridCalEngine.Core.Devices.multi_circuit import MultiCircuit
+from GridCalEngine.Devices import Bus, Generator, Load, Transformer2W, Line
+from GridCalEngine.Devices.multi_circuit import MultiCircuit
 
 
 class XmlDictConfig(dict):
@@ -37,12 +37,14 @@ class XmlDictConfig(dict):
     >>> xmldict = XmlDictConfig(root)
     And then use xmldict for what it is... a dict.
     """
+
     def __init__(self, parent_element, text_to_remove=''):
         """
 
         :param parent_element:
         :param text_to_remove:
         """
+        dict.__init__(self)
         self.text_to_remove = text_to_remove
 
         if parent_element.items():
@@ -58,12 +60,12 @@ class XmlDictConfig(dict):
 
                 self.update_shim({tag: a_dict})
 
-            elif element.items():    # items() is especialy for attributes
+            elif element.items():  # items() is especialy for attributes
 
                 element_attrib = element.items()
 
                 if element.text:
-                    element_attrib.append((tag, element.text))     # add tag:text if there exist
+                    element_attrib.append((tag, element.text))  # add tag:text if there exist
 
                 self.update_shim({tag: dict(element_attrib)})
             else:
@@ -75,7 +77,7 @@ class XmlDictConfig(dict):
         :param a_dict:
         :return:
         """
-        for og_key in a_dict.keys():   # keys() includes tag and attributes
+        for og_key in a_dict.keys():  # keys() includes tag and attributes
 
             key = og_key.replace(self.text_to_remove, '')
 
@@ -620,7 +622,7 @@ class PlxModel:
                     self.nodes[name].zone = self.zones[member]
 
             elif cls == 'Generator':
-                if row['collection'] == 'Nodes':
+                if row['collection'] == 'Schema':
                     self.generators[name].node = self.nodes[member]
                     self.nodes[member].generators.append(self.generators[name])
 
@@ -797,7 +799,7 @@ def get_st_generation_sent_out(plexos_results_folder):
     """
     fname = os.path.join(plexos_results_folder, 'Interval', 'ST Generator.Generation Sent Out.csv')
 
-    df = pd.read_csv(fname,index_col='DATETIME')
+    df = pd.read_csv(fname, index_col='DATETIME')
     df = pd.read_csv(fname, index_col='DATETIME')
 
     return df
@@ -807,6 +809,7 @@ def get_st_node_load(plexos_results_folder, parse_dates=False):
     """
     Get the node load use from a PLEXOS results folder
     :param plexos_results_folder: PLEXOS results folder
+    :param parse_dates: Parse the dates?
     :return: pandas DataFrame with the node load
     """
     fname = os.path.join(plexos_results_folder, 'Interval', 'ST Node.Load.csv')
@@ -854,7 +857,7 @@ def plx_to_gridcal(mdl: PlxModel, plexos_results_folder, time_indices=None, text
     # add the buses and the loads (in plexos there is only one load per bus)
     bus_dict = dict()
     for name, elm in mdl.nodes.items():
-        bus = Bus(name=name, vnom=elm.voltage,
+        bus = Bus(name=name, Vnom=elm.voltage,
                   latitude=elm.longitude,
                   longitude=-elm.latitude,
                   area=elm.region.name,
@@ -877,12 +880,11 @@ def plx_to_gridcal(mdl: PlxModel, plexos_results_folder, time_indices=None, text
             if (load_profile != 0).any():
                 load = Load(name='Load@' + name,
                             P=elm.load,
-                            Q=elm.load * 0.8,
-                            P_prof=load_profile,
-                            Q_prof=load_profile * 0.8,)
-
+                            Q=elm.load * 0.8, )
+                load.P_prof = load_profile
+                load.Q_prof = load_profile * 0.8
                 load.ensure_profiles_exist(circuit.time_profile)
-                bus.add_device(load)
+                circuit.add_load(bus, load)
 
         if text_func is not None:
             text_func("Creating GridCal model: Buses")
@@ -901,13 +903,12 @@ def plx_to_gridcal(mdl: PlxModel, plexos_results_folder, time_indices=None, text
                 gen_profile = gen_df[name].values
 
             gen = Generator(name=name,
-                            P_prof=gen_profile,
                             Pmin=elm.p_min,
                             Pmax=elm.p_max)
-
+            gen.P_prof = gen_profile
             bus = bus_dict[elm.node.name]
             gen.ensure_profiles_exist(circuit.time_profile)
-            bus.add_device(gen)
+            circuit.add_generator(bus, gen)
 
         if text_func is not None:
             text_func("Creating GridCal model: Buses")
@@ -928,17 +929,16 @@ def plx_to_gridcal(mdl: PlxModel, plexos_results_folder, time_indices=None, text
             profile = None
             rating = elm.rate_max
 
-        br = Branch(bus_from=bus_f,
-                    bus_to=bus_t,
-                    name=name,
-                    active=elm.units,
-                    r=elm.R,
-                    x=elm.X,
-                    rate=rating,
-                    branch_type=BranchType.Line)
+        br = Line(bus_from=bus_f,
+                  bus_to=bus_t,
+                  name=name,
+                  active=elm.units,
+                  r=elm.R,
+                  x=elm.X,
+                  rate=rating)
         br.rate_prof = profile
         br.ensure_profiles_exist(circuit.time_profile)
-        circuit.add_branch(br)
+        circuit.add_line(br)
 
         if text_func is not None:
             text_func("Creating GridCal model: Buses")
@@ -959,17 +959,16 @@ def plx_to_gridcal(mdl: PlxModel, plexos_results_folder, time_indices=None, text
             profile = None
             rating = elm.rate_max
 
-        br = Branch(bus_from=bus_f,
-                    bus_to=bus_t,
-                    name=name,
-                    active=elm.units,
-                    r=elm.R,
-                    x=elm.X,
-                    rate=rating,
-                    branch_type=BranchType.Transformer)
+        br = Transformer2W(bus_from=bus_f,
+                           bus_to=bus_t,
+                           name=name,
+                           active=elm.units,
+                           r=elm.R,
+                           x=elm.X,
+                           rate=rating)
         br.rate_prof = profile
         br.ensure_profiles_exist(circuit.time_profile)
-        circuit.add_branch(br)
+        circuit.add_transformer2w(br)
 
         if text_func is not None:
             text_func("Creating GridCal model: Buses")
@@ -979,4 +978,3 @@ def plx_to_gridcal(mdl: PlxModel, plexos_results_folder, time_indices=None, text
         nn += 1
 
     return circuit
-
