@@ -1,5 +1,5 @@
 # GridCal
-# Copyright (C) 2015 - 2023 Santiago Peñate Vera
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -17,8 +17,8 @@
 import numpy as np
 from typing import Dict, List, Tuple, Union
 from GridCalEngine.basic_structures import Logger
-import GridCalEngine.Core.Devices as dev
-from GridCalEngine.Core.Devices.multi_circuit import MultiCircuit
+import GridCalEngine.Devices as dev
+from GridCalEngine.Devices.multi_circuit import MultiCircuit
 from GridCalEngine.IO.raw.devices.area import RawArea
 from GridCalEngine.IO.raw.devices.branch import RawBranch
 from GridCalEngine.IO.raw.devices.bus import RawBus
@@ -44,19 +44,20 @@ from GridCalEngine.IO.raw.devices.psse_circuit import PsseCircuit
 def get_gridcal_bus(psse_bus: RawBus,
                     area_dict: Dict[int, dev.Area],
                     zone_dict: Dict[int, dev.Zone],
-                    logger: Logger) -> dev.Bus:
+                    logger: Logger) -> Tuple[dev.Bus, Union[dev.Shunt, None]]:
     """
 
     :return:
     """
 
-    bustype = {1: dev.BusMode.PQ, 2: dev.BusMode.PV, 3: dev.BusMode.Slack, 4: dev.BusMode.PQ}
+    bustype = {1: dev.BusMode.PQ_tpe, 2: dev.BusMode.PV_tpe, 3: dev.BusMode.Slack_tpe, 4: dev.BusMode.PQ_tpe}
+    sh = None
 
     if psse_bus.version >= 33:
         # create bus
         name = psse_bus.NAME.replace("'", "")
         bus = dev.Bus(name=name,
-                      vnom=psse_bus.BASKV, code=str(psse_bus.I), vmin=psse_bus.EVLO, vmax=psse_bus.EVHI, xpos=0, ypos=0,
+                      Vnom=psse_bus.BASKV, code=str(psse_bus.I), vmin=psse_bus.EVLO, vmax=psse_bus.EVHI, xpos=0, ypos=0,
                       active=True,
                       area=area_dict[psse_bus.AREA],
                       zone=zone_dict[psse_bus.ZONE],
@@ -66,7 +67,7 @@ def get_gridcal_bus(psse_bus: RawBus,
     elif psse_bus.version == 32:
         # create bus
         name = psse_bus.NAME
-        bus = dev.Bus(name=name, code=str(psse_bus.I), vnom=psse_bus.BASKV, vmin=psse_bus.NVLO, vmax=psse_bus.NVHI,
+        bus = dev.Bus(name=name, code=str(psse_bus.I), Vnom=psse_bus.BASKV, vmin=psse_bus.NVLO, vmax=psse_bus.NVHI,
                       xpos=0,
                       ypos=0,
                       active=True,
@@ -78,7 +79,7 @@ def get_gridcal_bus(psse_bus: RawBus,
     elif psse_bus.version in [29, 30]:
         # create bus
         name = psse_bus.NAME
-        bus = dev.Bus(name=name, code=str(psse_bus.I), vnom=psse_bus.BASKV, vmin=0.9, vmax=1.1, xpos=0, ypos=0,
+        bus = dev.Bus(name=name, code=str(psse_bus.I), Vnom=psse_bus.BASKV, vmin=0.9, vmax=1.1, xpos=0, ypos=0,
                       active=True,
                       area=area_dict[psse_bus.AREA],
                       zone=zone_dict[psse_bus.ZONE],
@@ -90,14 +91,12 @@ def get_gridcal_bus(psse_bus: RawBus,
                            G=psse_bus.GL, B=psse_bus.BL,
                            active=True)
 
-            bus.shunts.append(sh)
-
     else:
         logger.add_warning('Bus not implemented for version', str(psse_bus.version))
         # create bus (try v33)
         name = psse_bus.NAME.replace("'", "")
         bus = dev.Bus(name=name,
-                      vnom=psse_bus.BASKV, code=str(psse_bus.I), vmin=psse_bus.EVLO, vmax=psse_bus.EVHI, xpos=0, ypos=0,
+                      Vnom=psse_bus.BASKV, code=str(psse_bus.I), vmin=psse_bus.EVLO, vmax=psse_bus.EVHI, xpos=0, ypos=0,
                       active=True,
                       area=area_dict[psse_bus.AREA],
                       zone=zone_dict[psse_bus.ZONE],
@@ -108,12 +107,12 @@ def get_gridcal_bus(psse_bus: RawBus,
     if psse_bus.IDE in bustype.keys():
         bus.type = bustype[psse_bus.IDE]
     else:
-        bus.type = dev.BusMode.PQ
+        bus.type = dev.BusMode.PQ_tpe
 
     if int(psse_bus.IDE) == 4:
         bus.active = False
 
-    if bus.type == dev.BusMode.Slack:
+    if bus.type == dev.BusMode.Slack_tpe:
         bus.is_slack = True
 
     # Ensures unique name
@@ -124,7 +123,7 @@ def get_gridcal_bus(psse_bus: RawBus,
     if bus.name == '':
         bus.name = 'Bus ' + str(psse_bus.I)
 
-    return bus
+    return bus, sh
 
 
 def get_gridcal_load(psse_load: RawLoad, bus: dev.Bus, logger: Logger) -> dev.Load:
@@ -133,7 +132,7 @@ def get_gridcal_load(psse_load: RawLoad, bus: dev.Bus, logger: Logger) -> dev.Lo
     Returns:
         Newton Load object
     """
-    name = str(psse_load.I) + '_' + psse_load.ID.replace("'", "")
+    name = str(psse_load.I) + '_' + str(psse_load.ID).replace("'", "")
     name = name.strip()
 
     # GL and BL come in MW and MVAr
@@ -187,11 +186,17 @@ def get_gridcal_shunt_fixed(psse_elm: RawFixedShunt, bus: dev.Bus, logger: Logge
     return elm
 
 
-def get_gridcal_shunt_switched(psse_elm: RawSwitchedShunt, bus: dev.Bus, logger: Logger):
+def get_gridcal_shunt_switched(psse_elm: RawSwitchedShunt,
+                               bus: dev.Bus,
+                               psse_bus_dict: Dict[int, dev.Bus],
+                               logger: Logger) -> dev.ControllableShunt:
     """
-    Return Newton Load object
-    Returns:
-        Newton Load object
+
+    :param psse_elm:
+    :param bus:
+    :param psse_bus_dict:
+    :param logger:
+    :return:
     """
     name = str(psse_elm.I).replace("'", "")
     name = name.strip()
@@ -203,27 +208,49 @@ def get_gridcal_shunt_switched(psse_elm: RawSwitchedShunt, bus: dev.Bus, logger:
     if vv == 0:
         logger.add_error('Voltage equal to zero in shunt conversion', name)
 
-    g = 0.0
     if psse_elm.MODSW in [1, 2]:
         b = psse_elm.BINIT * psse_elm.RMPCT / 100.0
     else:
         b = psse_elm.BINIT
 
-    elm = dev.Shunt(name='Switched shunt ' + name,
-                    G=g, B=b,
-                    active=bool(psse_elm.STAT),
-                    code=name)
+    vset = (psse_elm.VSWHI + psse_elm.VSWLO) / 2.0
+
+    elm = dev.ControllableShunt(name='Switched shunt ' + name,
+                                active=bool(psse_elm.STAT),
+                                B=b,
+                                vset=vset,
+                                code=name)
+
+    if psse_elm.SWREG > 0:
+        if psse_elm.SWREG != psse_elm.I:
+            elm.control_bus = psse_bus_dict[psse_elm.SWREG]
+
+    n_list = []
+    b_list = []
+
+    for i in range(1, 9):
+        s = getattr(psse_elm, f"S{i}")
+        n = getattr(psse_elm, f"N{i}")
+
+        if s == 1:
+            n_list.append(n)
+            b_list.append(getattr(psse_elm, f"B{i}"))
+
+    elm.set_blocks(n_list, b_list)
 
     return elm
 
 
-def get_gridcal_generator(psse_elm: RawGenerator, logger: Logger) -> dev.Generator:
+def get_gridcal_generator(psse_elm: RawGenerator, psse_bus_dict: Dict[int, dev.Bus], logger: Logger) -> dev.Generator:
     """
-    Return Newton Load object
-    Returns:
-        Newton Load object
+
+    :param psse_elm:
+    :param psse_bus_dict:
+    :param logger:
+    :return:
     """
     name = str(psse_elm.I) + '_' + str(psse_elm.ID).replace("'", "")
+
     elm = dev.Generator(name=name,
                         idtag=None,
                         code=name,
@@ -234,13 +261,19 @@ def get_gridcal_generator(psse_elm: RawGenerator, logger: Logger) -> dev.Generat
                         Snom=psse_elm.MBASE,
                         Pmax=psse_elm.PT,
                         Pmin=psse_elm.PB,
-                        active=bool(psse_elm.STAT))
+                        active=bool(psse_elm.STAT),
+                        power_factor=psse_elm.WPF)
+
+    if psse_elm.IREG > 0:
+        if psse_elm.IREG != psse_elm.I:
+            elm.control_bus = psse_bus_dict[psse_elm.IREG]
 
     return elm
 
 
 def get_gridcal_transformer(psse_elm: RawTransformer,
-                            psse_bus_dict, Sbase,
+                            psse_bus_dict: Dict[int, dev.Bus],
+                            Sbase: float,
                             logger: Logger) -> Tuple[Union[dev.Transformer2W, dev.Transformer3W], int]:
     """
 
@@ -345,20 +378,9 @@ def get_gridcal_transformer(psse_elm: RawTransformer,
         bus_3 = psse_bus_dict[abs(psse_elm.K)]
         code = str(psse_elm.I) + '_' + str(psse_elm.J) + '_' + str(psse_elm.K) + '_' + str(psse_elm.CKT)
 
-        if psse_elm.NOMV1 == 0:
-            V1 = bus_1.Vnom
-        else:
-            V1 = psse_elm.NOMV1
-
-        if psse_elm.NOMV2 == 0:
-            V2 = bus_2.Vnom
-        else:
-            V2 = psse_elm.NOMV2
-
-        if psse_elm.NOMV3 == 0:
-            V3 = bus_3.Vnom
-        else:
-            V3 = psse_elm.NOMV3
+        V1 = bus_1.Vnom if psse_elm.NOMV1 == 0 else psse_elm.NOMV1
+        V2 = bus_2.Vnom if psse_elm.NOMV2 == 0 else psse_elm.NOMV2
+        V3 = bus_3.Vnom if psse_elm.NOMV3 == 0 else psse_elm.NOMV3
 
         """
         PSS/e's randomness:
@@ -426,12 +448,14 @@ def get_gridcal_transformer(psse_elm: RawTransformer,
                                  active=bool(psse_elm.STAT),
                                  r12=r12, r23=r23, r31=r31,
                                  x12=x12, x23=x23, x31=x31,
-                                 rate12=psse_elm.RATE1_1, rate23=psse_elm.RATE2_1, rate31=psse_elm.RATE3_1,
-                                 x=0.0, y=0.0)
+                                 rate12=psse_elm.RATE1_1,
+                                 rate23=psse_elm.RATE2_1,
+                                 rate31=psse_elm.RATE3_1)
 
         tr3w.winding1.tap_phase = psse_elm.ANG1
         tr3w.winding2.tap_phase = psse_elm.ANG2
         tr3w.winding3.tap_phase = psse_elm.ANG3
+        tr3w.compute_delta_to_star()
 
         return tr3w, 3
 
@@ -439,7 +463,10 @@ def get_gridcal_transformer(psse_elm: RawTransformer,
         raise Exception(str(psse_elm.windings) + ' number of windings!')
 
 
-def get_gridcal_line(psse_elm: RawBranch, psse_bus_dict, Sbase, logger: Logger) -> dev.Line:
+def get_gridcal_line(psse_elm: RawBranch,
+                     psse_bus_dict: Dict[int, dev.Bus],
+                     Sbase: float,
+                     logger: Logger) -> dev.Line:
     """
 
     :param psse_elm:
@@ -484,45 +511,61 @@ def get_gridcal_line(psse_elm: RawBranch, psse_bus_dict, Sbase, logger: Logger)
     return branch
 
 
-def get_hvdc_from_vscdc(psse_elm: RawVscDCLine, psse_bus_dict, Sbase, logger: Logger) -> dev.HvdcLine:
+def get_hvdc_from_vscdc(psse_elm: RawVscDCLine,
+                        psse_bus_dict: Dict[int, dev.Bus],
+                        Sbase: float,
+                        logger: Logger) -> Union[dev.HvdcLine, None]:
     """
-    GEt equivalent object
+    Get equivalent object
+    :param psse_elm:
     :param psse_bus_dict:
+    :param Sbase: Base power in MVA
     :param logger:
     :return:
     """
-    bus1 = psse_bus_dict[abs(psse_elm.IBUS1)]
-    bus2 = psse_bus_dict[abs(psse_elm.IBUS2)]
+    IBUS1 = abs(psse_elm.IBUS1)
+    IBUS2 = abs(psse_elm.IBUS2)
 
-    name1 = psse_elm.NAME.replace("'", "").replace('/', '').strip()
-    idtag = str(psse_elm.IBUS1) + '_' + str(psse_elm.IBUS2) + '_1'
+    if IBUS1 > 0 and IBUS2 > 0:
+        bus1 = psse_bus_dict[IBUS1]
+        bus2 = psse_bus_dict[IBUS2]
 
-    Vset_f = psse_elm.ACSET1
-    Vset_t = psse_elm.ACSET2
-    rate = max(psse_elm.SMAX1, psse_elm.SMAX2)
+        name1 = psse_elm.NAME.replace("'", "").replace('/', '').strip()
+        idtag = str(psse_elm.IBUS1) + '_' + str(psse_elm.IBUS2) + '_1'
 
-    # Estimate power
-    # P = dV^2 / R
-    V1 = bus1.Vnom * Vset_f
-    V2 = bus2.Vnom * Vset_t
-    dV = (V1 - V2) * 1000.0  # in V
-    P = dV * dV / psse_elm.RDC if psse_elm.RDC != 0 else 0  # power in W
-    specified_power = P * 1e-6  # power in MW
+        Vset_f = psse_elm.ACSET1
+        Vset_t = psse_elm.ACSET2
+        rate = max(psse_elm.SMAX1, psse_elm.SMAX2)
 
-    obj = dev.HvdcLine(bus_from=bus1,
-                       bus_to=bus2,
-                       name=name1,
-                       idtag=idtag,
-                       Pset=specified_power,
-                       Vset_f=Vset_f,
-                       Vset_t=Vset_t,
-                       rate=rate)
+        # Estimate power
+        # P = dV^2 / R
+        V1 = bus1.Vnom * Vset_f
+        V2 = bus2.Vnom * Vset_t
+        dV = (V1 - V2) * 1000.0  # in V
+        P = dV * dV / psse_elm.RDC if psse_elm.RDC != 0 else 0  # power in W
+        specified_power = P * 1e-6  # power in MW
 
-    return obj
+        obj = dev.HvdcLine(bus_from=bus1,
+                           bus_to=bus2,
+                           name=name1,
+                           idtag=idtag,
+                           Pset=specified_power,
+                           Vset_f=Vset_f,
+                           Vset_t=Vset_t,
+                           rate=rate)
+
+        return obj
+    else:
+
+        logger.add_error("VscDCLine has no bus from or bus to, or is missing both", device=psse_elm.get_id())
+
+        return None
 
 
-def get_hvdc_from_twotermdc(psse_elm: RawTwoTerminalDCLine, psse_bus_dict, Sbase: float,
-                            logger: Logger) -> dev.HvdcLine:
+def get_hvdc_from_twotermdc(psse_elm: RawTwoTerminalDCLine,
+                            psse_bus_dict: Dict[int, dev.Bus],
+                            Sbase: float,
+                            logger: Logger) -> Union[dev.HvdcLine, None]:
     """
 
     :param psse_elm:
@@ -531,52 +574,66 @@ def get_hvdc_from_twotermdc(psse_elm: RawTwoTerminalDCLine, psse_bus_dict, Sbase
     :param logger:
     :return:
     """
-    bus1 = psse_bus_dict[abs(psse_elm.IPR)]
-    bus2 = psse_bus_dict[abs(psse_elm.IPI)]
+    IPR = abs(psse_elm.IPR)
+    IPI = abs(psse_elm.IPI)
 
-    if psse_elm.MDC == 1 or psse_elm.MDC == 0:
-        # SETVL is in MW
-        specified_power = psse_elm.SETVL
-    elif psse_elm.MDC == 2:
-        # SETVL is in A, specified_power in MW
-        specified_power = psse_elm.SETVL * psse_elm.VSCHD / 1000.0
+    if IPR > 0 and IPI > 0:
+        bus1 = psse_bus_dict[IPR]
+        bus2 = psse_bus_dict[IPI]
+
+        if psse_elm.MDC == 1 or psse_elm.MDC == 0:
+            # SETVL is in MW
+            specified_power = psse_elm.SETVL
+        elif psse_elm.MDC == 2:
+            # SETVL is in A, specified_power in MW
+            specified_power = psse_elm.SETVL * psse_elm.VSCHD / 1000.0
+        else:
+            # doesn't say, so zero
+            specified_power = 0.0
+
+        # z_base = psse_elm.VSCHD * psse_elm.VSCHD / Sbase
+        # r_pu = psse_elm.RDC / z_base
+
+        Vset_f = 1.0
+        Vset_t = 1.0
+
+        name1 = psse_elm.NAME.replace("'", "").replace('"', "").replace('/', '').strip()
+        idtag = str(psse_elm.IPR) + '_' + str(psse_elm.IPI) + '_1'
+
+        # set the HVDC line active
+        active = bus1.active and bus2.active
+
+        obj = dev.HvdcLine(bus_from=bus1,  # Rectifier as of PSSe
+                           bus_to=bus2,  # inverter as of PSSe
+                           active=active,
+                           name=name1,
+                           idtag=idtag,
+                           Pset=specified_power,
+                           Vset_f=Vset_f,
+                           Vset_t=Vset_t,
+                           rate=specified_power,
+                           r=psse_elm.RDC,
+                           min_firing_angle_f=np.deg2rad(psse_elm.ANMNR),
+                           max_firing_angle_f=np.deg2rad(psse_elm.ANMXR),
+                           min_firing_angle_t=np.deg2rad(psse_elm.ANMNI),
+                           max_firing_angle_t=np.deg2rad(psse_elm.ANMXI))
+        return obj
     else:
-        # doesn't say, so zero
-        specified_power = 0.0
+        logger.add_error("HVDC2TermDC has no bus from or bus to, or is missing both", device=psse_elm.get_id())
 
-    # z_base = psse_elm.VSCHD * psse_elm.VSCHD / Sbase
-    # r_pu = psse_elm.RDC / z_base
-
-    Vset_f = 1.0
-    Vset_t = 1.0
-
-    name1 = psse_elm.NAME.replace("'", "").replace('"', "").replace('/', '').strip()
-    idtag = str(psse_elm.IPR) + '_' + str(psse_elm.IPI) + '_1'
-
-    # set the HVDC line active
-    active = bus1.active and bus2.active
-
-    obj = dev.HvdcLine(bus_from=bus1,  # Rectifier as of PSSe
-                       bus_to=bus2,  # inverter as of PSSe
-                       active=active,
-                       name=name1,
-                       idtag=idtag,
-                       Pset=specified_power,
-                       Vset_f=Vset_f,
-                       Vset_t=Vset_t,
-                       rate=specified_power,
-                       r=psse_elm.RDC,
-                       min_firing_angle_f=np.deg2rad(psse_elm.ANMNR),
-                       max_firing_angle_f=np.deg2rad(psse_elm.ANMXR),
-                       min_firing_angle_t=np.deg2rad(psse_elm.ANMNI),
-                       max_firing_angle_t=np.deg2rad(psse_elm.ANMXI))
-    return obj
+        return None
 
 
-def get_upfc_from_facts(psse_elm: RawFACTS, psse_bus_dict, Sbase, logger: Logger) -> None:
+def get_upfc_from_facts(psse_elm: RawFACTS,
+                        psse_bus_dict: Dict[int, dev.Bus],
+                        Sbase: float,
+                        logger: Logger,
+                        circuit: MultiCircuit):
     """
-    GEt equivalent object
+    Get equivalent object
+    :param psse_elm:
     :param psse_bus_dict:
+    :param Sbase:
     :param logger:
     :param circuit:
     :return:
@@ -599,74 +656,88 @@ def get_upfc_from_facts(psse_elm: RawFACTS, psse_bus_dict, Sbase, logger: Logger
     if '*' in str(psse_elm.SET1):
         psse_elm.SET1 = 0.0
 
-    if mode == 0:
-        active = False
-    elif mode == 1 and abs(psse_elm.J) > 0:
-        # shunt link
-        logger.add_warning('FACTS mode not implemented', str(mode))
+    if abs(psse_elm.J) == 0:  # STATCOM device
+        if mode == 0:
+            active = False
+        else:
+            active = True
 
-    elif mode == 2:
-        # only shunt device: STATCOM
-        logger.add_warning('FACTS mode not implemented', str(mode))
+        # TODO add STATCOM obj
 
-    elif mode == 3 and abs(psse_elm.J) > 0:  # const Z
-        # series and shunt links operating with series link at constant series impedance
-        # sh = Shunt(name='FACTS:' + name1, B=psse_elm.SHMX)
-        # load_from = Load(name='FACTS:' + name1, P=-psse_elm.PDES, Q=-psse_elm.QDES)
-        # gen_to = Generator(name='FACTS:' + name1, active_power=psse_elm.PDES, voltage_module=psse_elm.VSET)
-        # # branch = Line(bus_from=bus1, bus_to=bus2, name='FACTS:' + name1, x=psse_elm.LINX)
-        # circuit.add_shunt(bus1, sh)
-        # circuit.add_load(bus1, load_from)
-        # circuit.add_generator(bus2, gen_to)
-        # # circuit.add_line(branch)
+    elif abs(psse_elm.J) > 0:  # FACTS series device
 
-        elm = dev.UPFC(name=name1,
-                       bus_from=bus1,
-                       bus_to=bus2,
-                       code=idtag,
-                       rs=psse_elm.SET1,
-                       xs=psse_elm.SET2 + psse_elm.LINX,
-                       rp=0.0,
-                       xp=1.0 / psse_elm.SHMX if psse_elm.SHMX > 0 else 0.0,
-                       vp=psse_elm.VSET,
-                       Pset=psse_elm.PDES,
-                       Qset=psse_elm.QDES,
-                       rate=psse_elm.IMX + 1e-20)
+        if mode == 0:
+            active = False
+        elif mode == 1 and abs(psse_elm.J) > 0:
+            # shunt link
+            sh = dev.Shunt(name='FACTS:' + name1, B=psse_elm.SHMX)
+            circuit.add_shunt(bus1, sh)
+            logger.add_warning('FACTS mode (shunt link) added as shunt', str(mode))
 
-        return elm
+        elif mode == 2:
+            # only shunt device: STATCOM
+            logger.add_warning('FACTS mode (STATCOM) not implemented', str(mode))
 
-    elif mode == 4 and abs(psse_elm.J) > 0:
-        # series and shunt links operating with series link at constant series voltage
-        logger.add_warning('FACTS mode not implemented', str(mode))
+        elif mode == 3 and abs(psse_elm.J) > 0:  # const Z
+            # series and shunt links operating with series link at constant series impedance
+            # sh = Shunt(name='FACTS:' + name1, B=psse_elm.SHMX)
+            # load_from = Load(name='FACTS:' + name1, P=-psse_elm.PDES, Q=-psse_elm.QDES)
+            # gen_to = Generator(name='FACTS:' + name1, active_power=psse_elm.PDES, voltage_module=psse_elm.VSET)
+            # # branch = Line(bus_from=bus1, bus_to=bus2, name='FACTS:' + name1, x=psse_elm.LINX)
+            # circuit.add_shunt(bus1, sh)
+            # circuit.add_load(bus1, load_from)
+            # circuit.add_generator(bus2, gen_to)
+            # # circuit.add_line(branch)
 
-    elif mode == 5 and abs(psse_elm.J) > 0:
-        # master device of an IPFC with P and Q setpoints specified;
-        # another FACTS device must be designated as the slave device
-        # (i.e., its MODE is 6 or 8) of this IPFC.
-        logger.add_warning('FACTS mode not implemented', str(mode))
+            elm = dev.UPFC(name=name1,
+                           bus_from=bus1,
+                           bus_to=bus2,
+                           code=idtag,
+                           rs=psse_elm.SET1,
+                           xs=psse_elm.SET2 + psse_elm.LINX,
+                           rp=0.0,
+                           xp=1.0 / psse_elm.SHMX if psse_elm.SHMX > 0 else 0.0,
+                           vp=psse_elm.VSET,
+                           Pset=psse_elm.PDES,
+                           Qset=psse_elm.QDES,
+                           rate=psse_elm.IMX + 1e-20)
 
-    elif mode == 6 and abs(psse_elm.J) > 0:
-        # 6 slave device of an IPFC with P and Q setpoints specified;
-        #  the FACTS device specified in MNAME must be the master
-        #  device (i.e., its MODE is 5 or 7) of this IPFC. The Q setpoint is
-        #  ignored as the master device dictates the active power
-        #  exchanged between the two devices.
-        logger.add_warning('FACTS mode not implemented', str(mode))
+            circuit.add_upfc(elm)
 
-    elif mode == 7 and abs(psse_elm.J) > 0:
-        # master device of an IPFC with constant series voltage setpoints
-        # specified; another FACTS device must be designated as the slave
-        # device (i.e., its MODE is 6 or 8) of this IPFC
-        logger.add_warning('FACTS mode not implemented', str(mode))
+        elif mode == 4 and abs(psse_elm.J) > 0:
+            # series and shunt links operating with series link at constant series voltage
+            logger.add_warning('FACTS mode (series+shunt links) not implemented', str(mode))
 
-    elif mode == 8 and abs(psse_elm.J) > 0:
-        # slave device of an IPFC with constant series voltage setpoints
-        # specified; the FACTS device specified in MNAME must be the
-        # master device (i.e., its MODE is 5 or 7) of this IPFC. The complex
-        # Vd + jVq setpoint is modified during power flow solutions to reflect
-        # the active power exchange determined by the master device
-        logger.add_warning('FACTS mode not implemented', str(mode))
+        elif mode == 5 and abs(psse_elm.J) > 0:
+            # master device of an IPFC with P and Q setpoints specified;
+            # another FACTS device must be designated as the slave device
+            # (i.e., its MODE is 6 or 8) of this IPFC.
+            logger.add_warning('FACTS mode (IPFC) not implemented', str(mode))
 
+        elif mode == 6 and abs(psse_elm.J) > 0:
+            # 6 slave device of an IPFC with P and Q setpoints specified;
+            #  the FACTS device specified in MNAME must be the master
+            #  device (i.e., its MODE is 5 or 7) of this IPFC. The Q setpoint is
+            #  ignored as the master device dictates the active power
+            #  exchanged between the two devices.
+            logger.add_warning('FACTS mode (IPFC) not implemented', str(mode))
+
+        elif mode == 7 and abs(psse_elm.J) > 0:
+            # master device of an IPFC with constant series voltage setpoints
+            # specified; another FACTS device must be designated as the slave
+            # device (i.e., its MODE is 6 or 8) of this IPFC
+            logger.add_warning('FACTS mode (IPFC) not implemented', str(mode))
+
+        elif mode == 8 and abs(psse_elm.J) > 0:
+            # slave device of an IPFC with constant series voltage setpoints
+            # specified; the FACTS device specified in MNAME must be the
+            # master device (i.e., its MODE is 5 or 7) of this IPFC. The complex
+            # Vd + jVq setpoint is modified during power flow solutions to reflect
+            # the active power exchange determined by the master device
+            logger.add_warning('FACTS mode (IPFC) not implemented', str(mode))
+
+        else:
+            return None
     else:
         return None
 
@@ -707,10 +778,10 @@ def psse_to_gridcal(psse_circuit: PsseCircuit,
             zones_dict[abs(psse_bus.ZONE)] = dev.Zone(name='Z' + str(abs(psse_bus.ZONE)))
             missing_zones = True
 
-        bus = get_gridcal_bus(psse_bus=psse_bus,
-                              area_dict=area_dict,
-                              zone_dict=zones_dict,
-                              logger=logger)
+        bus, bus_shunt = get_gridcal_bus(psse_bus=psse_bus,
+                                         area_dict=area_dict,
+                                         zone_dict=zones_dict,
+                                         logger=logger)
 
         # bus.ensure_area_objects(circuit)
 
@@ -720,6 +791,10 @@ def psse_to_gridcal(psse_circuit: PsseCircuit,
         circuit.add_bus(bus)
         psse_bus_dict[psse_bus.I] = bus
 
+        # legacy PSSe buses may have shunts declared within, so add them
+        if bus_shunt is not None:
+            circuit.add_shunt(bus=bus, api_obj=bus_shunt)
+
     if missing_areas:
         circuit.areas = [v for k, v in area_dict.items()]
 
@@ -733,10 +808,13 @@ def psse_to_gridcal(psse_circuit: PsseCircuit,
 
     # Go through loads
     for psse_load in psse_circuit.loads:
-        bus = psse_bus_dict[psse_load.I]
-        api_obj = get_gridcal_load(psse_load, bus, logger)
+        if psse_load.I in psse_bus_dict:
+            bus = psse_bus_dict[psse_load.I]
+            api_obj = get_gridcal_load(psse_load, bus, logger)
 
-        circuit.add_load(bus, api_obj)
+            circuit.add_load(bus, api_obj)
+        else:
+            logger.add_error("Load bus is missing", psse_load.I, psse_load.I)
 
     # Go through shunts
     for psse_shunt in psse_circuit.fixed_shunts:
@@ -750,17 +828,18 @@ def psse_to_gridcal(psse_circuit: PsseCircuit,
     for psse_shunt in psse_circuit.switched_shunts:
         if psse_shunt.I in psse_bus_dict:
             bus = psse_bus_dict[psse_shunt.I]
-            api_obj = get_gridcal_shunt_switched(psse_shunt, bus, logger)
-            circuit.add_shunt(bus, api_obj)
+            api_obj = get_gridcal_shunt_switched(psse_shunt, bus, psse_bus_dict, logger)
+            circuit.add_controllable_shunt(bus, api_obj)
         else:
             logger.add_error("Switched shunt bus missing", psse_shunt.I, psse_shunt.I)
 
     # Go through generators
     for psse_gen in psse_circuit.generators:
         bus = psse_bus_dict[psse_gen.I]
-        api_obj = get_gridcal_generator(psse_gen, logger)
+        api_obj = get_gridcal_generator(psse_gen, psse_bus_dict, logger)
 
         circuit.add_generator(bus, api_obj)
+        api_obj.is_controlled = psse_gen.WMOD == 0 or psse_gen.WMOD == 1
 
     # Go through Branches
     branches_already_there = set()
@@ -818,40 +897,36 @@ def psse_to_gridcal(psse_circuit: PsseCircuit,
         # get the object
         branch = get_hvdc_from_vscdc(psse_branch, psse_bus_dict, psse_circuit.SBASE, logger)
 
-        if branch.idtag not in branches_already_there:
+        if branch is not None:
+            if branch.idtag not in branches_already_there:
 
-            # Add to the circuit
-            circuit.add_hvdc(branch)
-            branches_already_there.add(branch.idtag)
+                # Add to the circuit
+                circuit.add_hvdc(branch)
+                branches_already_there.add(branch.idtag)
 
-        else:
-            logger.add_warning('The RAW file has a repeated HVDC line device and it is omitted from the model',
-                               str(branch.idtag))
+            else:
+                logger.add_warning('The RAW file has a repeated HVDC line device and it is omitted from the model',
+                                   str(branch.idtag))
 
     for psse_branch in psse_circuit.two_terminal_dc_lines:
         # get the object
         branch = get_hvdc_from_twotermdc(psse_branch, psse_bus_dict, psse_circuit.SBASE, logger)
 
-        if branch.idtag not in branches_already_there:
+        if branch is not None:
+            if branch.idtag not in branches_already_there:
 
-            # Add to the circuit
-            circuit.add_hvdc(branch)
-            branches_already_there.add(branch.idtag)
+                # Add to the circuit
+                circuit.add_hvdc(branch)
+                branches_already_there.add(branch.idtag)
 
-        else:
-            logger.add_warning('The RAW file has a repeated HVDC line device and it is omitted from the model',
-                               str(branch.idtag))
+            else:
+                logger.add_warning('The RAW file has a repeated HVDC line device and it is omitted from the model',
+                                   str(branch.idtag))
 
     # Go through facts
     for psse_elm in psse_circuit.facts:
         # since these may be shunt or series or both, pass the circuit so that the correct device is added
         if psse_elm.is_connected():
-            elm = get_upfc_from_facts(psse_elm, psse_bus_dict, psse_circuit.SBASE, logger)
-
-            if elm is not None:
-                circuit.add_upfc(elm)
-            else:
-                code = str(psse_elm.I) + '_' + str(psse_elm.J) + '_1'
-                logger.add_warning('FACTS device was not converted', str(code))
+            get_upfc_from_facts(psse_elm, psse_bus_dict, psse_circuit.SBASE, logger, circuit=circuit)
 
     return circuit
