@@ -1,5 +1,5 @@
 # GridCal
-# Copyright (C) 2015 - 2023 Santiago Peñate Vera
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
 # 
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -18,15 +18,14 @@
 import numpy as np
 import numba as nb
 
-from GridCalEngine.Core.Devices.multi_circuit import MultiCircuit
-from GridCalEngine.Core.DataStructures.numerical_circuit import compile_numerical_circuit_at
+from GridCalEngine.Devices.multi_circuit import MultiCircuit
+from GridCalEngine.DataStructures.numerical_circuit import compile_numerical_circuit_at
 from GridCalEngine.Simulations.LinearFactors.linear_analysis import LinearAnalysis
-from GridCalEngine.Simulations.driver_types import SimulationTypes
-from GridCalEngine.Simulations.result_types import ResultTypes
 from GridCalEngine.Simulations.results_table import ResultsTable
 from GridCalEngine.Simulations.results_template import ResultsTemplate
 from GridCalEngine.Simulations.driver_template import DriverTemplate
-from GridCalEngine.enumerations import StudyResultsType, AvailableTransferMode
+from GridCalEngine.Simulations.ATC.available_transfer_capacity_options import AvailableTransferCapacityOptions
+from GridCalEngine.enumerations import StudyResultsType, AvailableTransferMode, ResultTypes, DeviceType, SimulationTypes
 
 
 @nb.njit()
@@ -454,68 +453,17 @@ class AvailableTransferCapacityResults(ResultsTemplate):
             data = np.array(self.report)
             y_label = ''
             title, _ = result_type.value
-            index = self.report_indices
-            labels = self.report_headers
+
+            return ResultsTable(data=np.array(self.report),
+                                index=self.report_indices,
+                                columns=self.report_headers,
+                                title=title,
+                                ylabel=y_label,
+                                cols_device_type=DeviceType.NoDevice,
+                                idx_device_type=DeviceType.NoDevice)
         else:
             raise Exception('Result type not understood:' + str(result_type))
 
-        # assemble model
-        mdl = ResultsTable(data=data,
-                           index=index,
-                           columns=labels,
-                           title=title,
-                           ylabel=y_label)
-        return mdl
-
-
-class AvailableTransferCapacityOptions:
-
-    def __init__(self, distributed_slack=True, correct_values=True, use_provided_flows=False,
-                 bus_idx_from=list(), bus_idx_to=list(), idx_br=list(), sense_br=list(), Pf=None,
-                 idx_hvdc_br=list(), sense_hvdc_br=list(), Pf_hvdc=None,
-                 dT=100.0, threshold=0.02, mode: AvailableTransferMode = AvailableTransferMode.Generation,
-                 max_report_elements=-1, use_clustering=False, cluster_number=200):
-        """
-
-        :param distributed_slack:
-        :param correct_values:
-        :param use_provided_flows:
-        :param bus_idx_from:
-        :param bus_idx_to:
-        :param idx_br:
-        :param sense_br:
-        :param Pf:
-        :param idx_hvdc_br:
-        :param sense_hvdc_br:
-        :param Pf_hvdc:
-        :param dT:
-        :param threshold:
-        :param mode:
-        :param max_report_elements: maximum number of elements to show in the report (-1 for all)
-        :param use_clustering:
-        """
-        self.distributed_slack = distributed_slack
-        self.correct_values = correct_values
-        self.use_provided_flows = use_provided_flows
-        self.bus_idx_from = bus_idx_from
-        self.bus_idx_to = bus_idx_to
-        self.inter_area_branch_idx = idx_br
-        self.inter_area_branch_sense = sense_br
-        self.Pf = Pf
-
-        self.idx_hvdc_br = idx_hvdc_br
-        self.inter_area_hvdc_branch_sense = sense_hvdc_br
-        self.Pf_hvdc = Pf_hvdc
-
-        self.dT = dT
-        self.threshold = threshold
-        self.mode = mode
-
-        self.max_report_elements = max_report_elements
-
-        self.use_clustering = use_clustering
-        self.cluster_number = cluster_number
-
 
 class AvailableTransferCapacityDriver(DriverTemplate):
     tpe = SimulationTypes.NetTransferCapacity_run
@@ -545,15 +493,16 @@ class AvailableTransferCapacityDriver(DriverTemplate):
         Run thread
         """
         self.tic()
-        self.progress_text.emit('Analyzing')
-        self.progress_signal.emit(0)
+
+        self.report_text("Analyzing")
+        self.report_progress(0.0)
 
         # get the converted bus indices
         idx1b = self.options.bus_idx_from
         idx2b = self.options.bus_idx_to
 
         # declare the numerical circuit
-        nc = compile_numerical_circuit_at(circuit=self.grid, t_idx=None)
+        nc = compile_numerical_circuit_at(circuit=self.grid, t_idx=None, logger=self.logger)
 
         # declare the linear analysis
         linear = LinearAnalysis(
