@@ -1,38 +1,39 @@
-# This file is a python port of the routines included in MATPOWER to perform continuation power flow.
-# The license is the same BSD-style that is provided in LICENSE_MATPOWER
+# GridCal
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
+#
+# This program is free software; you can redistribute it and/or
+# modify it under the terms of the GNU Lesser General Public
+# License as published by the Free Software Foundation; either
+# version 3 of the License, or (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+# Lesser General Public License for more details.
+#
+# You should have received a copy of the GNU Lesser General Public License
+# along with this program; if not, write to the Free Software Foundation,
+# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 
 import numpy as np
-import scipy
-import scipy.sparse as sp
-from enum import Enum
-from GridCalEngine.basic_structures import ReactivePowerControlMode, Logger
+
+from GridCalEngine.enumerations import ReactivePowerControlMode, CpfParametrization, CpfStopAt
 from GridCalEngine.Simulations.PowerFlow.NumericalMethods.ac_jacobian import AC_jacobian
 from GridCalEngine.Simulations.PowerFlow.NumericalMethods.discrete_controls import control_q_direct
-from GridCalEngine.Core.topology import compile_types
-import GridCalEngine.Simulations.PowerFlow.NumericalMethods.common_functions as cf
-from GridCalEngine.Simulations.sparse_solve import get_sparse_type, get_linear_solver
+from GridCalEngine.Topology.simulation_indices import compile_types
+from GridCalEngine.Simulations.PowerFlow.NumericalMethods.common_functions import (polar_to_rect, compute_power)
+from GridCalEngine.basic_structures import Vec, CxVec, IntVec
+from GridCalEngine.Utils.Sparse.csc2 import spsolve_csc, extend
 
-linear_solver = get_linear_solver()
-sparse = get_sparse_type()
-scipy.ALLOW_THREADS = True
 np.set_printoptions(precision=8, suppress=True, linewidth=320)
 
 
-class CpfStopAt(Enum):
-    Nose = 'Nose'
-    ExtraOverloads = 'Extra overloads'
-    Full = 'Full curve'
-
-
-class CpfParametrization(Enum):
-    Natural = 'Natural'
-    ArcLength = 'Arc Length'
-    PseudoArcLength = 'Pseudo Arc Length'
-
-
 class CpfNumericResults:
+    """
+    CpfNumericResults
+    """
 
-    def __init__(self):
+    def __init__(self) -> None:
         self.V = list()
         self.Sbus = list()
         self.lmbda = list()
@@ -43,7 +44,21 @@ class CpfNumericResults:
         self.normF = list()
         self.success = list()
 
-    def add(self, v, sbus, Sf, St, lam, losses, loading, normf, converged):
+    def add(self, v: CxVec, sbus: CxVec, Sf: CxVec, St: CxVec, lam: float,
+            losses: CxVec, loading: CxVec, normf: float, converged: bool):
+        """
+
+        :param v:
+        :param sbus:
+        :param Sf:
+        :param St:
+        :param lam:
+        :param losses:
+        :param loading:
+        :param normf:
+        :param converged:
+        :return:
+        """
         self.V.append(v)
         self.Sbus.append(sbus)
         self.lmbda.append(lam)
@@ -54,11 +69,12 @@ class CpfNumericResults:
         self.normF.append(normf)
         self.success.append(converged)
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self.V)
 
 
-def cpf_p(parametrization: CpfParametrization, step, z, V, lam, V_prev, lamprv, pv, pq, pvpq):
+def cpf_p(parametrization: CpfParametrization, step: float, z: Vec, V: CxVec, lam: Vec,
+          V_prev: CxVec, lamprv: Vec, pv: IntVec, pq: IntVec, pvpq: IntVec):
     """
     Computes the value of the Current Parametrization Function
     :param parametrization: Value of  option (1: Natural, 2:Arc-length, 3: pseudo arc-length)
@@ -111,13 +127,13 @@ def cpf_p(parametrization: CpfParametrization, step, z, V, lam, V_prev, lamprv,
 
     ## evaluate P(x0, lambda0)
     """
-    if parametrization == CpfParametrization.Natural:        # natural
+    if parametrization == CpfParametrization.Natural:  # natural
         if lam >= lamprv:
             P = lam - lamprv - step
         else:
             P = lamprv - lam - step
 
-    elif parametrization == CpfParametrization.ArcLength:    # arc length
+    elif parametrization == CpfParametrization.ArcLength:  # arc length
         Va = np.angle(V)
         Vm = np.abs(V)
         Va_prev = np.angle(V_prev)
@@ -126,7 +142,7 @@ def cpf_p(parametrization: CpfParametrization, step, z, V, lam, V_prev, lamprv,
         b = np.r_[Va_prev[pvpq], Vm_prev[pq], lamprv]
         P = np.sum(np.power(a - b, 2)) - np.power(step, 2)
 
-    elif parametrization == CpfParametrization.PseudoArcLength:    # pseudo arc length
+    elif parametrization == CpfParametrization.PseudoArcLength:  # pseudo arc length
         nb = len(V)
         Va = np.angle(V)
         Vm = np.abs(V)
@@ -146,7 +162,8 @@ def cpf_p(parametrization: CpfParametrization, step, z, V, lam, V_prev, lamprv,
     return P
 
 
-def cpf_p_jac(parametrization: CpfParametrization, z, V, lam, Vprv, lamprv, pv, pq, pvpq):
+def cpf_p_jac(parametrization: CpfParametrization, z, V, lam, Vprv, lamprv,
+              pv: IntVec, pq: IntVec, pvpq: IntVec):
     """
     Computes partial derivatives of Current Parametrization Function (CPF).
     :param parametrization:
@@ -200,7 +217,7 @@ def cpf_p_jac(parametrization: CpfParametrization, z, V, lam, Vprv, lamprv, pv,
     #   See http://www.pserc.cornell.edu/matpower/ for more info.
     """
 
-    if parametrization == CpfParametrization.Natural:   # natural
+    if parametrization == CpfParametrization.Natural:  # natural
         npv = len(pv)
         npq = len(pq)
         dP_dV = np.zeros(npv + 2 * npq)
@@ -216,8 +233,8 @@ def cpf_p_jac(parametrization: CpfParametrization, z, V, lam, Vprv, lamprv, pv,
         Vmprv = np.abs(Vprv)
         dP_dV = 2.0 * (np.r_[Va[pvpq], Vm[pq]] - np.r_[Vaprv[pvpq], Vmprv[pq]])
 
-        if lam == lamprv:   # first step
-            dP_dlam = 1.0   # avoid singular Jacobian that would result from [dP_dV, dP_dlam] = 0
+        if lam == lamprv:  # first step
+            dP_dlam = 1.0  # avoid singular Jacobian that would result from [dP_dV, dP_dlam] = 0
         else:
             dP_dlam = 2.0 * (lam - lamprv)
 
@@ -235,7 +252,7 @@ def cpf_p_jac(parametrization: CpfParametrization, z, V, lam, Vprv, lamprv, pv,
     return dP_dV, dP_dlam
 
 
-def predictor(V, lam, Ybus, Sxfr, pv, pq, step, z, Vprv, lamprv,
+def predictor(V, lam, Ybus, Sxfr, pv: IntVec, pq: IntVec, step: float, z, Vprv, lamprv,
               parametrization: CpfParametrization):
     """
     Computes a prediction (approximation) to the next solution of the
@@ -275,8 +292,15 @@ def predictor(V, lam, Ybus, Sxfr, pv, pq, step, z, Vprv, lamprv,
     J2 = [   J   dF_dlam
            dP_dV dP_dlam ]
     '''
-    J2 = sp.vstack([sp.hstack([J, dF_dlam.reshape(nj, 1)]),
-                    sp.hstack([dP_dV, dP_dlam])], format="csc")
+
+    # last_col = np.empty((nj, 1))
+    # last_col[:, 0] = dF_dlam
+    #
+    # last_row = np.empty((1, nj + 1))
+    # last_row[0, :nj] = dP_dV
+    # last_row[0, nj] = dP_dlam
+
+    J2 = extend(J, dF_dlam, dP_dV, dP_dlam)
 
     Va_prev = np.angle(V)
     Vm_prev = np.abs(V)
@@ -288,7 +312,7 @@ def predictor(V, lam, Ybus, Sxfr, pv, pq, step, z, Vprv, lamprv,
     s[npv + 2 * npq] = 1
 
     # tangent vector
-    z[np.r_[pvpq, nb + pq, 2 * nb]] = linear_solver(J2, s)
+    z[np.r_[pvpq, nb + pq, 2 * nb]] = spsolve_csc(J2, s)
 
     # normalize_string tangent predictor  (dividing by the euclidean norm)
     z /= np.linalg.norm(z)
@@ -306,7 +330,7 @@ def predictor(V, lam, Ybus, Sxfr, pv, pq, step, z, Vprv, lamprv,
     return V0, lam0, z
 
 
-def corrector(Ybus, Sbus, V0, pv, pq, lam0, Sxfr, Vprv, lamprv, z, step, parametrization, tol, max_it,
+def corrector(Ybus, Sbus, V0, pv: IntVec, pq: IntVec, lam0, Sxfr, Vprv, lamprv, z, step, parametrization, tol, max_it,
               verbose, mu_0=1.0, acceleration_parameter=0.5):
     """
     Solves the corrector step of a continuation power flow using a full Newton method
@@ -340,6 +364,8 @@ def corrector(Ybus, Sbus, V0, pv, pq, lam0, Sxfr, Vprv, lamprv, z, step, paramet
     :param tol: Tolerance (p.u.)
     :param max_it: max iterations
     :param verbose: print information?
+    :param mu_0:
+    :param acceleration_parameter:
     :return: Voltage, converged, iterations, lambda, power error, calculated power
     """
 
@@ -348,10 +374,10 @@ def corrector(Ybus, Sbus, V0, pv, pq, lam0, Sxfr, Vprv, lamprv, z, step, paramet
     V = V0
     Va = np.angle(V)
     Vm = np.abs(V)
-    lam = lam0             # set lam to initial lam0
+    lam = lam0  # set lam to initial lam0
     dVa = np.zeros_like(Va)
     dVm = np.zeros_like(Vm)
-    dlam = 0
+    # dlam = 0
 
     # set up indexing for updating V
     npv = len(pv)
@@ -369,13 +395,13 @@ def corrector(Ybus, Sbus, V0, pv, pq, lam0, Sxfr, Vprv, lamprv, z, step, paramet
     Scalc = V * np.conj(Ybus * V)
     mismatch = Scalc - Sbus - lam * Sxfr
     # F = np.r_[mismatch[pvpq].real, mismatch[pq].imag]
-    
+
     # evaluate P(x0, lambda0)
     P = cpf_p(parametrization, step, z, V, lam, Vprv, lamprv, pv, pq, pvpq)
-    
+
     # augment F(x,lambda) with P(x,lambda)
     F = np.r_[mismatch[pvpq].real, mismatch[pq].imag, P]
-    
+
     # check tolerance
     normF = np.linalg.norm(F, np.Inf)
     converged = normF < tol
@@ -389,22 +415,31 @@ def corrector(Ybus, Sbus, V0, pv, pq, lam0, Sxfr, Vprv, lamprv, z, step, paramet
 
         # update iteration counter
         i += 1
-        
+
         # evaluate Jacobian
         J = AC_jacobian(Ybus, V, pvpq, pq)
 
         dP_dV, dP_dlam = cpf_p_jac(parametrization, z, V, lam, Vprv, lamprv, pv, pq, pvpq)
-    
+
         # augment J with real/imag - Sxfr and z^T
         '''
         J = [   J   dF_dlam 
               dP_dV dP_dlam ]
         '''
-        J = sp.vstack([sp.hstack([J, dF_dlam.reshape(nj, 1)]),
-                       sp.hstack([dP_dV, dP_dlam])], format="csc")
-    
+        # J = sp.vstack([sp.hstack([J, dF_dlam.reshape(nj, 1)]),
+        #                sp.hstack([dP_dV, dP_dlam])], format="csc")
+
+        # last_col = np.empty((nj, 1))
+        # last_col[:, 0] = dF_dlam
+        #
+        # last_row = np.empty((1, nj + 1))
+        # last_row[0, :nj] = dP_dV
+        # last_row[0, nj] = dP_dlam
+
+        J = extend(J, dF_dlam, dP_dV, dP_dlam)
+
         # compute update step
-        dx = linear_solver(J, F)
+        dx = spsolve_csc(J, F)
         dVa[pvpq] = dx[j1:j2]
         dVm[pq] = dx[j2:j3]
         dlam = dx[j3]
@@ -426,17 +461,17 @@ def corrector(Ybus, Sbus, V0, pv, pq, lam0, Sxfr, Vprv, lamprv, z, step, paramet
                 Va = prev_Va.copy()
                 Vm = prev_Vm.copy()
                 lam = prev_lam
-    
+
             # update the variables from the solution
             Va -= mu * dVa
             Vm -= mu * dVm
             lam -= mu * dlam
 
             # update Vm and Va again in case we wrapped around with a negative Vm
-            V = cf.polar_to_rect(Vm, Va)
+            V = polar_to_rect(Vm, Va)
 
             # evaluate F(x, lam)
-            Scalc = cf.compute_power(Ybus, V)
+            Scalc = compute_power(Ybus, V)
             mismatch = Scalc - Sbus - lam * Sxfr
 
             # evaluate the parametrization function P(x, lambda)
@@ -476,13 +511,13 @@ def corrector(Ybus, Sbus, V0, pv, pq, lam0, Sxfr, Vprv, lamprv, z, step, paramet
     return V, converged, i, lam, normF, Scalc
 
 
-def continuation_nr(Ybus, Cf, Ct, Yf, Yt, branch_rates, Sbase, Ibus_base, Ibus_target, Sbus_base, Sbus_target,
+def continuation_nr(Ybus, Cf, Ct, Yf, Yt, branch_rates, Sbase, Sbus_base, Sbus_target,
                     V, distributed_slack, bus_installed_power,
-                    vd, pv, pq, step, approximation_order: CpfParametrization,
+                    vd, pv: IntVec, pq: IntVec, step, approximation_order: CpfParametrization,
                     adapt_step, step_min, step_max, error_tol=1e-3, tol=1e-6, max_it=20,
                     stop_at=CpfStopAt.Nose, control_q=ReactivePowerControlMode.NoControl,
                     qmax_bus=None, qmin_bus=None, original_bus_types=None, base_overload_number=0,
-                    verbose=False, call_back_fx=None, logger=Logger()) -> CpfNumericResults:
+                    verbose=False, call_back_fx=None) -> CpfNumericResults:
     """
     Runs a full AC continuation power flow using a normalized tangent
     predictor and selected approximation_order scheme.
@@ -492,8 +527,7 @@ def continuation_nr(Ybus, Cf, Ct, Yf, Yt, branch_rates, Sbase, Ibus_base, Ibus_t
     :param Yf: Admittance matrix of the "from" nodes
     :param Yt: Admittance matrix of the "to" nodes
     :param branch_rates: array of branch rates to check the overload condition
-    :param Ibus_base:
-    :param Ibus_target:
+    :param Sbase:
     :param Sbus_base: Power array of the base solvable case
     :param Sbus_target: Power array of the case to be solved
     :param V: Voltage array of the base solved case
@@ -518,7 +552,6 @@ def continuation_nr(Ybus, Cf, Ct, Yf, Yt, branch_rates, Sbase, Ibus_base, Ibus_t
     :param base_overload_number: number of overloads in the base situation (used when stop_at=CpfStopAt.ExtraOverloads)
     :param verbose: Display additional intermediate information?
     :param call_back_fx: Function to call on every iteration passing the lambda parameter
-    :param logger: Logger instance
     :return: CpfNumericResults instance
 
 
@@ -541,8 +574,8 @@ def continuation_nr(Ybus, Cf, Ct, Yf, Yt, branch_rates, Sbase, Ibus_base, Ibus_t
     Sxfr = Sbus_target - Sbus_base
     nb = len(Sbus_base)
     lam = 0
-    lam_prev = lam   # lam at previous step
-    V_prev = V       # V at previous step
+    lam_prev = lam  # lam at previous step
+    V_prev = V  # V at previous step
     continuation = True
     cont_steps = 0
     pvpq = np.r_[pv, pq]
@@ -648,20 +681,19 @@ def continuation_nr(Ybus, Cf, Ct, Yf, Yt, branch_rates, Sbase, Ibus_base, Ibus_t
             # Check controls
             if control_q == ReactivePowerControlMode.Direct:
 
-
                 Vm = np.abs(V)
-                V, \
-                Qnew, \
-                types_new, \
-                any_q_control_issue = control_q_direct(V=V,
-                                                       Vm=Vm,
-                                                       Vset=Vm,
-                                                       Q=Scalc.imag,
-                                                       Qmax=qmax_bus,
-                                                       Qmin=qmin_bus,
-                                                       types=bus_types,
-                                                       original_types=original_bus_types,
-                                                       verbose=verbose)
+                (V,
+                 Qnew,
+                 types_new,
+                 any_q_control_issue) = control_q_direct(V=V,
+                                                         Vm=Vm,
+                                                         Vset=Vm,
+                                                         Q=Scalc.imag,
+                                                         Qmax=qmax_bus,
+                                                         Qmin=qmin_bus,
+                                                         types=bus_types,
+                                                         original_types=original_bus_types,
+                                                         verbose=verbose)
             else:
                 # did not check Q limits
                 any_q_control_issue = False
@@ -673,9 +705,9 @@ def continuation_nr(Ybus, Cf, Ct, Yf, Yt, branch_rates, Sbase, Ibus_base, Ibus_t
                 bus_types = types_new
                 Sbus = Scalc.real + 1j * Qnew
 
-                Sxfr = Sbus_target - Sbus  # TODO: really?
+                Sxfr = Sbus_target - Sbus
 
-                vd, pq, pv, pqpv = compile_types(Pbus=Sbus.real, types=types_new)
+                vd, pq, pv, pqv, p, pqpv = compile_types(Pbus=Sbus.real, types=types_new)
             else:
                 if verbose:
                     print('Q controls Ok')
@@ -715,7 +747,7 @@ def continuation_nr(Ybus, Cf, Ct, Yf, Yt, branch_rates, Sbase, Ibus_base, Ibus_t
                     continuation = False
 
             else:
-                raise Exception('Stop point ' + stop_at.value + ' not recognised.')
+                raise Exception(f'Stop point {stop_at.value} not recognised.')
 
             if adapt_step and continuation:
 
@@ -748,4 +780,3 @@ def continuation_nr(Ybus, Cf, Ct, Yf, Yt, branch_rates, Sbase, Ibus_base, Ibus_t
                 print('step ', cont_steps, ' : lambda = ', lam, ', corrector did not converge in ', i, ' iterations\n')
 
     return results
-
