@@ -1,21 +1,40 @@
-
+# GridCal
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
+#
+# This program is free software; you can redistribute it and/or
+# modify it under the terms of the GNU Lesser General Public
+# License as published by the Free Software Foundation; either
+# version 3 of the License, or (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+# Lesser General Public License for more details.
+#
+# You should have received a copy of the GNU Lesser General Public License
+# along with this program; if not, write to the Free Software Foundation,
+# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+from __future__ import annotations
+from typing import TYPE_CHECKING, Union
 import numpy as np
 import pandas as pd
 from GridCalEngine.Simulations.driver_template import DriverTemplate
-from GridCalEngine.Simulations.driver_types import SimulationTypes
 from GridCalEngine.Simulations.results_template import ResultsTemplate
-from GridCalEngine.Simulations.result_types import ResultTypes
 from GridCalEngine.Simulations.results_table import ResultsTable
-from GridCalEngine.Core.Devices.multi_circuit import MultiCircuit
-from GridCalEngine.enumerations import StudyResultsType
+from GridCalEngine.Devices.multi_circuit import MultiCircuit
+from GridCalEngine.enumerations import StudyResultsType, ResultTypes, DeviceType, SimulationTypes
+
+if TYPE_CHECKING:  # Only imports the below statements during type checking
+    from GridCalEngine.Simulations.OPF.opf_results import OptimalPowerFlowResults
+    from GridCalEngine.Simulations.OPF.opf_ts_results import OptimalPowerFlowTimeSeriesResults
 
 
 class InputsAnalysisResults(ResultsTemplate):
     tpe = 'Inputs Analysis'
 
     def __init__(self, grid: MultiCircuit,
-                 opf_results: "OptimalPowerFlowResults" = None,
-                 opf_time_series_results: "OptimalPowerFlowTimeSeriesResults" = None):
+                 opf_results: Union[None, OptimalPowerFlowResults] = None,
+                 opf_time_series_results: Union[None, OptimalPowerFlowTimeSeriesResults] = None):
         """
         Construct the analysis
         :param grid:
@@ -235,11 +254,11 @@ class InputsAnalysisResults(ResultsTemplate):
         d = {elm: i for i, elm in enumerate(self.grid.substations)}
         return np.array([d.get(bus.substation, "") for bus in self.grid.buses])
 
-    def get_collection_attr_series(self, elms, attr, aggregation="Area"):
+    def get_collection_attr_series(self, elms, magnitude: str, aggregation="Area"):
         """
 
         :param elms:
-        :param attr:
+        :param magnitude:snaphot property name
         :param aggregation:
         :return:
         """
@@ -272,7 +291,7 @@ class InputsAnalysisResults(ResultsTemplate):
         for elm in elms:
             i = self.bus_dict[elm.bus]
             i2 = d2[i]
-            x[:, i2] += getattr(elm, attr)
+            x[:, i2] += elm.get_profile(magnitude=magnitude).toarray()
 
         return x, headers
 
@@ -284,121 +303,150 @@ class InputsAnalysisResults(ResultsTemplate):
                 (or None if the result was not understood)
         """
 
-        columns = [result_type.value[0]]
-
         if result_type == ResultTypes.ZoneAnalysis:
             df = self.group_by('Zone')
-            columns = df.columns.values
-            labels = df.index.values
-            y = df.values
-            y_label = ''
-            title = result_type.value[0]
+
+            return ResultsTable(data=df.values,
+                                index=df.index.values,
+                                idx_device_type=DeviceType.ZoneDevice,
+                                columns=df.columns.values,
+                                cols_device_type=DeviceType.ZoneDevice,
+                                title=result_type.value)
 
         elif result_type == ResultTypes.AreaAnalysis:
             df = self.group_by('Area')
-            columns = df.columns.values
-            labels = df.index.values
-            y = df.values
-            y_label = ''
-            title = result_type.value[0]
+            return ResultsTable(data=df.values,
+                                index=df.index.values,
+                                idx_device_type=DeviceType.AreaDevice,
+                                columns=df.columns.values,
+                                cols_device_type=DeviceType.AreaDevice,
+                                title=result_type.value)
 
         elif result_type == ResultTypes.CountryAnalysis:
             df = self.group_by('Country')
-            columns = df.columns.values
-            labels = df.index.values
-            y = df.values
-            y_label = ''
-            title = result_type.value[0]
+            return ResultsTable(data=df.values,
+                                index=df.index.values,
+                                idx_device_type=DeviceType.CountryDevice,
+                                columns=df.columns.values,
+                                cols_device_type=DeviceType.CountryDevice,
+                                title=result_type.value)
 
         elif result_type == ResultTypes.AreaGenerationAnalysis:
             generators = self.grid.get_generators() + self.grid.get_batteries() + self.grid.get_static_generators()
-            y, columns = self.get_collection_attr_series(generators, 'P_prof', 'Area')
-            labels = pd.to_datetime(self.grid.time_profile)
-            y_label = 'MW'
-            title = result_type.value[0]
+            y, columns = self.get_collection_attr_series(generators, 'P', 'Area')
+
+            return ResultsTable(data=y,
+                                index=pd.to_datetime(self.grid.time_profile),
+                                idx_device_type=DeviceType.TimeDevice,
+                                columns=columns,
+                                cols_device_type=DeviceType.AreaDevice,
+                                title=result_type.value,
+                                units="(MW)")
 
         elif result_type == ResultTypes.ZoneGenerationAnalysis:
             generators = self.grid.get_generators() + self.grid.get_batteries() + self.grid.get_static_generators()
-            y, columns = self.get_collection_attr_series(generators, 'P_prof', 'Zone')
-            labels = pd.to_datetime(self.grid.time_profile)
-            y_label = 'MW'
-            title = result_type.value[0]
+            y, columns = self.get_collection_attr_series(generators, 'P', 'Zone')
+
+            return ResultsTable(data=y,
+                                index=pd.to_datetime(self.grid.time_profile),
+                                idx_device_type=DeviceType.TimeDevice,
+                                columns=columns,
+                                cols_device_type=DeviceType.ZoneDevice,
+                                title=result_type.value,
+                                units="(MW)")
 
         elif result_type == ResultTypes.CountryGenerationAnalysis:
             generators = self.grid.get_generators() + self.grid.get_batteries() + self.grid.get_static_generators()
-            y, columns = self.get_collection_attr_series(generators, 'P_prof', 'Country')
-            labels = pd.to_datetime(self.grid.time_profile)
-            y_label = 'MW'
-            title = result_type.value[0]
+            y, columns = self.get_collection_attr_series(generators, 'P', 'Country')
+
+            return ResultsTable(data=y,
+                                index=pd.to_datetime(self.grid.time_profile),
+                                idx_device_type=DeviceType.TimeDevice,
+                                columns=columns,
+                                cols_device_type=DeviceType.CountryDevice,
+                                title=result_type.value,
+                                units="(MW)")
 
         elif result_type == ResultTypes.AreaLoadAnalysis:
-            y, columns = self.get_collection_attr_series(self.grid.get_loads(), 'P_prof', 'Area')
-            labels = pd.to_datetime(self.grid.time_profile)
-            y_label = 'MW'
-            title = result_type.value[0]
+            y, columns = self.get_collection_attr_series(self.grid.get_loads(), 'P', 'Area')
+            return ResultsTable(data=y,
+                                index=pd.to_datetime(self.grid.time_profile),
+                                idx_device_type=DeviceType.TimeDevice,
+                                columns=columns,
+                                cols_device_type=DeviceType.AreaDevice,
+                                title=result_type.value,
+                                units="(MW)")
 
         elif result_type == ResultTypes.ZoneLoadAnalysis:
-            y, columns = self.get_collection_attr_series(self.grid.get_loads(), 'P_prof', 'Zone')
-            labels = pd.to_datetime(self.grid.time_profile)
-            y_label = 'MW'
-            title = result_type.value[0]
+            y, columns = self.get_collection_attr_series(self.grid.get_loads(), 'P', 'Zone')
+            return ResultsTable(data=y,
+                                index=pd.to_datetime(self.grid.time_profile),
+                                idx_device_type=DeviceType.TimeDevice,
+                                columns=columns,
+                                cols_device_type=DeviceType.ZoneDevice,
+                                title=result_type.value,
+                                units="(MW)")
 
         elif result_type == ResultTypes.CountryLoadAnalysis:
-            y, columns = self.get_collection_attr_series(self.grid.get_loads(), 'P_prof', 'Country')
-            labels = pd.to_datetime(self.grid.time_profile)
-            y_label = 'MW'
-            title = result_type.value[0]
+            y, columns = self.get_collection_attr_series(self.grid.get_loads(), 'P', 'Country')
+            return ResultsTable(data=y,
+                                index=pd.to_datetime(self.grid.time_profile),
+                                idx_device_type=DeviceType.TimeDevice,
+                                columns=columns,
+                                cols_device_type=DeviceType.CountryDevice,
+                                title=result_type.value,
+                                units="(MW)")
 
         elif result_type == ResultTypes.AreaBalanceAnalysis:
             generators = self.grid.get_generators() + self.grid.get_batteries() + self.grid.get_static_generators()
-            yg, columns = self.get_collection_attr_series(generators, 'P_prof', 'Area')
+            yg, columns = self.get_collection_attr_series(generators, 'P', 'Area')
 
-            yl, columns = self.get_collection_attr_series(self.grid.get_loads(), 'P_prof', 'Area')
+            yl, columns = self.get_collection_attr_series(self.grid.get_loads(), 'P', 'Area')
 
             y = yg - yl
 
-            labels = pd.to_datetime(self.grid.time_profile)
-            y_label = 'MW'
-            title = result_type.value[0]
+            return ResultsTable(data=y,
+                                index=pd.to_datetime(self.grid.time_profile),
+                                idx_device_type=DeviceType.TimeDevice,
+                                columns=columns,
+                                cols_device_type=DeviceType.AreaDevice,
+                                title=result_type.value,
+                                units="(MW)")
 
         elif result_type == ResultTypes.ZoneBalanceAnalysis:
             generators = self.grid.get_generators() + self.grid.get_batteries() + self.grid.get_static_generators()
-            yg, columns = self.get_collection_attr_series(generators, 'P_prof', 'Zone')
+            yg, columns = self.get_collection_attr_series(generators, 'P', 'Zone')
 
-            yl, columns = self.get_collection_attr_series(self.grid.get_loads(), 'P_prof', 'Zone')
+            yl, columns = self.get_collection_attr_series(self.grid.get_loads(), 'P', 'Zone')
 
             y = yg - yl
 
-            labels = pd.to_datetime(self.grid.time_profile)
-            y_label = 'MW'
-            title = result_type.value[0]
+            return ResultsTable(data=y,
+                                index=pd.to_datetime(self.grid.time_profile),
+                                idx_device_type=DeviceType.TimeDevice,
+                                columns=columns,
+                                cols_device_type=DeviceType.ZoneDevice,
+                                title=result_type.value,
+                                units="(MW)")
 
         elif result_type == ResultTypes.CountryBalanceAnalysis:
             generators = self.grid.get_generators() + self.grid.get_batteries() + self.grid.get_static_generators()
-            yg, columns = self.get_collection_attr_series(generators, 'P_prof', 'Country')
+            yg, columns = self.get_collection_attr_series(generators, 'P', 'Country')
 
-            yl, columns = self.get_collection_attr_series(self.grid.get_loads(), 'P_prof', 'Country')
+            yl, columns = self.get_collection_attr_series(self.grid.get_loads(), 'P', 'Country')
 
             y = yg - yl
 
-            labels = pd.to_datetime(self.grid.time_profile)
-            y_label = 'MW'
-            title = result_type.value[0]
+            return ResultsTable(data=y,
+                                index=pd.to_datetime(self.grid.time_profile),
+                                idx_device_type=DeviceType.TimeDevice,
+                                columns=columns,
+                                cols_device_type=DeviceType.CountryDevice,
+                                title=result_type.value,
+                                units="(MW)")
 
         else:
-            labels = []
-            y = np.zeros(0)
-            y_label = '(MW)'
-            title = ''
-
-        mdl = ResultsTable(data=y,
-                           index=labels,
-                           columns=columns,
-                           title=title,
-                           ylabel=y_label,
-                           xlabel='',
-                           units=y_label)
-        return mdl
+            raise Exception('Result type not understood:' + str(result_type))
 
 
 class InputsAnalysisDriver(DriverTemplate):
@@ -432,5 +480,3 @@ class InputsAnalysisDriver(DriverTemplate):
 
     def cancel(self):
         self.__cancel__ = True
-
-
