@@ -1,5 +1,5 @@
 # GridCal
-# Copyright (C) 2015 - 2023 Santiago Peñate Vera
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -18,15 +18,18 @@
 import numpy as np
 import numba as nb
 import scipy.sparse as sp
-from typing import Union, List
-from scipy.sparse.linalg import spsolve
+from typing import Union, List, Tuple
+from scipy.sparse.linalg import spsolve as scipy_spsolve
 
 from GridCalEngine.basic_structures import Logger, Vec, IntVec, CxVec, Mat, ObjVec, CxMat
-from GridCalEngine.Core.DataStructures.numerical_circuit import NumericalCircuit
-from GridCalEngine.Core.Devices.multi_circuit import MultiCircuit
+from GridCalEngine.DataStructures.numerical_circuit import NumericalCircuit
+from GridCalEngine.Devices.multi_circuit import MultiCircuit
+from GridCalEngine.Devices.Aggregation.contingency_group import ContingencyGroup
+from GridCalEngine.Devices.Aggregation.contingency import Contingency
 from GridCalEngine.Simulations.PowerFlow.NumericalMethods.ac_jacobian import AC_jacobian
 from GridCalEngine.Simulations.PowerFlow.NumericalMethods.derivatives import dSf_dV_csc
 from GridCalEngine.Utils.Sparse.csc import dense_to_csc
+import GridCalEngine.Utils.Sparse.csc2 as csc
 from GridCalEngine.Utils.MIP.selected_interface import lpDot
 
 
@@ -125,7 +128,7 @@ def make_acptdf(Ybus: sp.csc_matrix,
     dS = np.r_[dP[pvpq, :], dQ]
 
     # solve the voltage increments
-    dx = spsolve(J, dS)
+    dx = csc.spsolve_csc(J, dS)
 
     # compute branch derivatives
     dSf_dVm, dSf_dVa = dSf_dV_csc(Yf.tocsc(), V, F, T)
@@ -138,23 +141,23 @@ def make_acptdf(Ybus: sp.csc_matrix,
     return PTDF
 
 
-def make_ptdf(Bbus: sp.csc_matrix,
+def make_ptdf(Bpqpv: sp.csc_matrix,
               Bf: sp.csc_matrix,
               pqpv: IntVec,
               distribute_slack: bool = True) -> Mat:
     """
     Build the PTDF matrix
-    :param Bbus: DC-linear susceptance matrix
+    :param Bpqpv: DC-linear susceptance matrix already sliced
     :param Bf: Bus-branch "from" susceptance matrix
     :param pqpv: array of sorted pq and pv node indices
     :param distribute_slack: distribute the slack?
     :return: PTDF matrix. It is a full matrix of dimensions Branches x buses
     """
 
-    n = Bbus.shape[0]
+    n = Bf.shape[1]
     nb = n
     nbi = n
-    noref = np.arange(1, nb)
+    noref = pqpv  # np.arange(1, nb)
     noslack = pqpv
 
     if distribute_slack:
@@ -166,8 +169,8 @@ def make_ptdf(Bbus: sp.csc_matrix,
 
     # solve for change in voltage angles
     dTheta = np.zeros((nb, nbi))
-    Bref = Bbus[noslack, :][:, noref].tocsc()
-    dtheta_ref = spsolve(Bref, dP[noslack, :])
+    # Bref = Bbus[noslack, :][:, noref].tocsc()
+    dtheta_ref = scipy_spsolve(Bpqpv, dP[noslack, :])
 
     if sp.issparse(dtheta_ref):
         dTheta[noref, :] = dtheta_ref.toarray()
@@ -224,78 +227,6 @@ def make_lodf(Cf: sp.csc_matrix,
     return LODF
 
 
-# @nb.njit(cache=True)
-def make_mlodf(circuit, lodf):
-    """
-    Function to build an MLODF matrix.
-    Deprecated, use LinearMultiContingency instead
-    :param circuit:
-    :param lodf:
-    :return:
-    """
-    lodf_nx_list = list()
-
-    # Create dictionaries to speed up the access
-    cg_dict = {cg.idtag: cg for cg in circuit.contingency_groups}
-    idx_dict = {e.idtag: i for i, e in enumerate(circuit.get_branches())}
-
-    # Initialize c_idx list for each contingency groups
-    for cg in circuit.contingency_groups:
-        cg.c_idx = list()
-
-    # Loop for contingencies to fill group c_idx
-    for c in circuit.contingencies:
-        cg_dict[c.group.idtag].c_idx.append(idx_dict[c.device_idtag])
-
-    for cg in circuit.contingency_groups:
-
-        # Sort unique c_idx
-        cg.c_idx = list(sorted(set(cg.c_idx), reverse=False))
-
-        # Compute LODF vector
-        L = lodf[:, cg.c_idx]  # Take the columns of the LODF associated with the contingencies
-
-        # Compute M matrix [n, n] (lodf relating the outaged lines to each other)
-        M = np.ones((len(cg.c_idx), len(cg.c_idx)))
-        for i in range(len(cg.c_idx)):
-            for j in range(len(cg.c_idx)):
-                if not (i == j):
-                    M[i, j] = -lodf[cg.c_idx[i], cg.c_idx[j]]
-
-        # Compute LODF_NX
-        lodf_nx = np.matmul(L, np.linalg.inv(M))
-
-        # store tuple (c_idx, lodf_nx)
-        lodf_nx_list.append(
-            (cg.c_idx, lodf_nx)
-        )
-
-    return lodf_nx_list
-
-
-@nb.njit(cache=True)
-def make_otdf(ptdf: Mat,
-              lodf: Mat,
-              j: int) -> Mat:
-    """
-    Outage sensitivity of the Branches when transferring power from the bus j to the slack
-        LODF: outage transfer distribution factors
-    :param ptdf: power transfer distribution factors matrix (n-branch, n-bus)
-    :param lodf: line outage distribution factors matrix (n-branch, n-branch)
-    :param j: index of the bus injection
-    :return: LODF matrix (n-branch, n-branch)
-    """
-    nk = ptdf.shape[0]
-    nl = nk
-    otdf = np.empty((nk, nl))
-
-    for k in range(nk):
-        for l in range(nl):
-            otdf[k, l] = ptdf[k, j] + lodf[k, l] * ptdf[l, j]
-
-    return otdf
-
-
 @nb.njit(cache=True)
 def make_transfer_limits(ptdf: Mat,
                          flows: Vec,
@@ -351,23 +282,35 @@ class LinearMultiContingency:
                  branch_indices: IntVec,
                  bus_indices: IntVec,
                  mlodf_factors: sp.csc_matrix,
-                 ptdf_factors: sp.csc_matrix,
+                 compensated_ptdf_factors: sp.csc_matrix,
                  injections_factor: Vec):
         """
         Linear multi contingency object
         :param branch_indices: contingency branch indices.
         :param bus_indices: contingency bus indices.
+
         :param mlodf_factors: MLODF factors applicable (all_branches, contingency branches).
-        :param ptdf_factors: PTDF factors applicable (all_branches, contingency buses)
-        :param injections_factor: Injection contingency factors, i.e percentage to decrease an injection (len(bus indices))
+                             Should be: MLODF[k, βδ]
+
+        :param compensated_ptdf_factors: compensated PTDF factors applicable (all_branches, contingency buses)
+                                         should be: MLODF[k, βδ] x PTDF[βδ, i] + PTDF[k, i]
+
+        :param injections_factor: Injection contingency factors,
+                                  i.e percentage to decrease an injection (len(bus indices))
         """
 
         assert len(bus_indices) == len(injections_factor)
 
         self.branch_indices: IntVec = branch_indices
         self.bus_indices: IntVec = bus_indices
+
+        # MLODF[k, βδ]
         self.mlodf_factors: sp.csc_matrix = mlodf_factors
-        self.ptdf_factors: sp.csc_matrix = ptdf_factors
+
+        # MLODF[k, βδ] x PTDF[βδ, i] + PTDF[k, i]
+        self.compensated_ptdf_factors: sp.csc_matrix = compensated_ptdf_factors
+
+        # percentage to decrease an injection, used to compute ΔP
         self.injections_factor: Vec = injections_factor
 
     def has_injection_contingencies(self) -> bool:
@@ -377,22 +320,26 @@ class LinearMultiContingency:
         """
         return len(self.bus_indices) > 0
 
-    def get_contingency_flows(self, base_flow: Vec, injections: Vec) -> Vec:
+    def get_contingency_flows(self, base_flow: Vec, injections: Vec, tau: Vec = None) -> Vec:
         """
         Get contingency flows
         :param base_flow: Base branch flows (nbranch)
         :param injections: Bus injections increments (nbus)
+        :param tau: Phase shifter angles (rad)
         :return: New flows (nbranch)
         """
 
         flow = base_flow.copy()
 
         if len(self.branch_indices):
+            # MLODF[k, βδ] x Pf0[βδ]
             flow += self.mlodf_factors @ base_flow[self.branch_indices]
 
         if len(self.bus_indices):
-            injection_delta = self.injections_factor * injections[self.bus_indices]
-            flow += self.ptdf_factors @ injection_delta[self.bus_indices]  # TODO: is the slicing necesary here too?
+            injection_delta = injections[self.bus_indices]
+
+            # (MLODF[k, βδ] x PTDF[βδ, i] + PTDF[k, i]) x ΔP[i]
+            flow += self.compensated_ptdf_factors @ injection_delta
 
         return flow
 
@@ -413,106 +360,258 @@ class LinearMultiContingency:
 
         if len(self.bus_indices):
             injection_delta = self.injections_factor * injections[self.bus_indices]
-            flow += lpDot(self.ptdf_factors, injection_delta[self.bus_indices])
+            flow += lpDot(self.compensated_ptdf_factors, injection_delta[self.bus_indices])
 
         return flow
 
 
+class ContingencyIndices:
+    """
+    Contingency indices
+    """
+
+    def __init__(self, contingency_group: ContingencyGroup, contingency_group_dict, branches_dict,
+                 generator_dict, bus_index_dict):
+
+        (self.branch_contingency_indices,
+         self.bus_contingency_indices,
+         self.injections_factors) = self.get_contingencies_info(contingency_group=contingency_group,
+                                                                contingency_group_dict=contingency_group_dict,
+                                                                branches_dict=branches_dict,
+                                                                generator_dict=generator_dict,
+                                                                bus_index_dict=bus_index_dict)
+
+    @staticmethod
+    def try_find_indices(cnt: Contingency, branches_dict, generator_dict, bus_index_dict,
+                         branch_contingency_indices, bus_contingency_indices, injections_factors):
+        """
+        Try to find the contingency indices f the device in the contingency
+        :param cnt: Contingency
+        :param branches_dict:
+        :param generator_dict:
+        :param bus_index_dict:
+        :param branch_contingency_indices:
+        :param bus_contingency_indices:
+        :param injections_factors:
+        :return:
+        """
+        # search for the contingency in the Branches
+        br_idx = branches_dict.get(cnt.device_idtag, None)
+        branch_found = False
+        if br_idx is not None:
+            if cnt.prop == 'active':
+                branch_contingency_indices.append(br_idx)
+                return
+            else:
+                branch_found = True
+                print(f'Unknown branch contingency property {cnt.prop} at {cnt.name} {cnt.idtag}')
+        else:
+            branch_found = False
+
+        gen = generator_dict.get(cnt.device_idtag, None)
+        gen_found = False
+        if gen is not None:
+            if cnt.prop == '%':
+                bus_contingency_indices.append(bus_index_dict[gen.bus])
+                injections_factors.append(cnt.value / 100.0)
+                return
+            else:
+                branch_found = True
+                print(f'Unknown generator contingency property {cnt.prop} at {cnt.name} {cnt.idtag}')
+        else:
+            gen_found = False
+
+        if not branch_found:
+            print(f"contingency branch {cnt.device_idtag} not found")
+        if not gen_found:
+            print(f"contingency generator {cnt.device_idtag} not found")
+
+    def get_contingencies_info(self, contingency_group: ContingencyGroup,
+                               contingency_group_dict, branches_dict,
+                               generator_dict, bus_index_dict) -> Tuple[IntVec, IntVec, Vec]:
+        """
+        Get the indices from a contingency group
+        :param contingency_group:
+        :param contingency_group_dict:
+        :param branches_dict:
+        :param generator_dict:
+        :param bus_index_dict:
+        :return: branch_contingency_indices, bus_contingency_indices, injections_factors
+        """
+
+        # get the group's contingencies
+        contingencies = contingency_group_dict[contingency_group.idtag]
+
+        branch_contingency_indices = list()
+        bus_contingency_indices = list()
+        injections_factors = list()
+
+        # apply the contingencies
+        for cnt in contingencies:
+            self.try_find_indices(cnt, branches_dict, generator_dict, bus_index_dict,
+                                  branch_contingency_indices, bus_contingency_indices, injections_factors)
+
+        branch_contingency_indices = np.array(branch_contingency_indices)
+        bus_contingency_indices = np.array(bus_contingency_indices)
+        injections_factors = np.array(injections_factors)
+
+        return branch_contingency_indices, bus_contingency_indices, injections_factors
+
+
 class LinearMultiContingencies:
     """
     LinearMultiContingencies
     """
 
-    def __init__(self, grid: MultiCircuit):
+    def __init__(self, grid: MultiCircuit, contingency_groups_used: List[ContingencyGroup]):
         """
         Constructor
         :param grid: MultiCircuit
         """
         self.grid: MultiCircuit = grid
 
+        self.contingency_groups_used = contingency_groups_used
+
         # auxiliary structures
         self.__contingency_group_dict = grid.get_contingency_group_dict()
         self.__bus_index_dict = grid.get_bus_index_dict()
         self.__branches_dict = {b.idtag: i for i, b in enumerate(grid.get_branches_wo_hvdc())}
         self.__generator_dict = {g.idtag: g for g in grid.get_contingency_devices()}
 
+        self.contingency_indices = list()
+
+        # for each contingency group
+        for ic, contingency_group in enumerate(self.contingency_groups_used):
+            self.contingency_indices.append(ContingencyIndices(contingency_group=contingency_group,
+                                                               contingency_group_dict=self.__contingency_group_dict,
+                                                               branches_dict=self.__branches_dict,
+                                                               generator_dict=self.__generator_dict,
+                                                               bus_index_dict=self.__bus_index_dict))
+
         # list of LinearMultiContingency objects that are used later to compute the contingency flows
         self.multi_contingencies: List[LinearMultiContingency] = list()
 
-    def update(self, lodf: Mat, ptdf: Mat, threshold: float = 0.0001) -> None:
+    def get_contingency_group_names(self) -> List[str]:
+        """
+        Returns a list of of the names of the used contingency groups
+        :return:
+        """
+        return [elm.name for elm in self.contingency_groups_used]
+
+    def compute(self,
+                lodf: Mat,
+                ptdf: Mat,
+                ptdf_threshold: float = 0.0001,
+                lodf_threshold: float = 0.0001,
+                prepare_for_srap: bool = False) -> None:
         """
         Make the LODF with any contingency combination using the declared contingency objects
         :param lodf: original LODF matrix (nbr, nbr)
         :param ptdf: original PTDF matrix (nbr, nbus)
-        :param threshold: threshold to discard values
+        :param ptdf_threshold: threshold to discard values
+        :param lodf_threshold: Threshold for LODF conversion to sparse
+        :param prepare_for_srap:
         :return: None
         """
 
         self.multi_contingencies = list()
 
         # for each contingency group
-        for ic, contingency_group in enumerate(self.grid.contingency_groups):
+        for ic, contingency_group in enumerate(self.contingency_groups_used):
 
-            # get the group's contingencies
-            contingencies = self.__contingency_group_dict[contingency_group.idtag]
+            contingency_indices = self.contingency_indices[ic]
 
-            branch_contingency_indices = list()
-            bus_contingency_indices = list()
-            injections_factors = list()
+            if len(contingency_indices.branch_contingency_indices) > 1:
 
-            # apply the contingencies
-            for cnt in contingencies:
-
-                # search for the contingency in the Branches
-                br_idx = self.__branches_dict.get(cnt.device_idtag, None)
-                if br_idx is not None:
-                    if cnt.prop == 'active':
-                        branch_contingency_indices.append(br_idx)
-                    else:
-                        print(f'Unknown contingency property {cnt.prop} at {cnt.name} {cnt.idtag}')
-
-                gen = self.__generator_dict.get(cnt.device_idtag, None)
-                if gen is not None:
-                    if cnt.prop == '%':
-                        bus_contingency_indices.append(self.__bus_index_dict[gen.bus])
-                        injections_factors.append(cnt.value / 100.0)
-
-            branch_contingency_indices = np.array(branch_contingency_indices)
-            bus_contingency_indices = np.array(bus_contingency_indices)
-            injections_factors = np.array(injections_factors)
-
-            if len(branch_contingency_indices) > 1:
+                # Flow =
+                #   Pf0[k]
+                # + MLODF[k, bd] * Pf0[bd]
+                # + MLODF[k, bd] * PTDF[bd, i] * dP[i]
+                # + PTDF[k, i] * dPi
 
                 # Compute M matrix [n, n] (lodf relating the outaged lines to each other)
-                M = create_M_numba(lodf=lodf, branch_contingency_indices=branch_contingency_indices)
-                L = lodf[:, branch_contingency_indices]
+                M = create_M_numba(lodf=lodf,
+                                   branch_contingency_indices=contingency_indices.branch_contingency_indices)
+                L = lodf[:, contingency_indices.branch_contingency_indices]
 
-                # Compute LODF for the multiple failure
-                mlodf_factors = dense_to_csc(mat=L @ np.linalg.inv(M),
-                                             threshold=threshold)
+                try:
+                    # Compute LODF for the multiple failure MLODF[k, βδ]
+                    mlodf_factors = dense_to_csc(mat=L @ np.linalg.inv(M),
+                                                 threshold=lodf_threshold)
+
+                except np.linalg.LinAlgError:
+                    # Done to capture antenna when computing multiples contingencies
+                    mlodf_factors = dense_to_csc(mat=L @ np.linalg.pinv(M),
+                                                 threshold=lodf_threshold)
+
+                if len(contingency_indices.bus_contingency_indices) > 0:
+                    # this is PTDF[k, i]
+                    ptdf_k_i = dense_to_csc(mat=ptdf[:, contingency_indices.bus_contingency_indices],
+                                            threshold=ptdf_threshold)
+                    # PTDF[βδ, i]
+                    ptdf_bd_i = dense_to_csc(
+                        mat=ptdf[np.ix_(contingency_indices.branch_contingency_indices,
+                                        contingency_indices.bus_contingency_indices)],
+                        threshold=ptdf_threshold
+                    )
+
+                else:
+                    ptdf_k_i = sp.csc_matrix((ptdf.shape[0], ptdf.shape[1]))
+
+                    # PTDF[βδ, i]
+                    ptdf_bd_i = dense_to_csc(mat=ptdf[contingency_indices.branch_contingency_indices, :],
+                                             threshold=ptdf_threshold)
+
+                # must compute: MLODF[k, βδ] x PTDF[βδ, i] + PTDF[k, i]
+                compensated_ptdf_factors = mlodf_factors @ ptdf_bd_i + ptdf_k_i
+
+            elif len(contingency_indices.branch_contingency_indices) == 1:
+
+                # Pf0[k]
+                # + LODF[k, c] * Pf0[c]
+                # + LODF[k, c] * PTDF[c, i] * dPi
+                # + PTDF[k, i] * dPi
 
-            elif len(branch_contingency_indices) == 1:
                 # append values
-                mlodf_factors = dense_to_csc(mat=lodf[:, branch_contingency_indices],
-                                             threshold=threshold)
+                mlodf_factors = dense_to_csc(mat=lodf[:, contingency_indices.branch_contingency_indices],
+                                             threshold=lodf_threshold)
+
+                if len(contingency_indices.bus_contingency_indices) > 0:
+                    # single branch and single bus contingency
+
+                    # this is PTDF[k, i]
+                    ptdf_k_i = dense_to_csc(mat=ptdf[:, contingency_indices.bus_contingency_indices],
+                                            threshold=ptdf_threshold)
+                    # PTDF[βδ, i]
+                    ptdf_bd_i = dense_to_csc(
+                        mat=ptdf[np.ix_(contingency_indices.branch_contingency_indices,
+                                        contingency_indices.bus_contingency_indices)],
+                        threshold=ptdf_threshold
+                    )
+
+                    # must compute: MLODF[k, βδ] x PTDF[βδ, i] + PTDF[k, i]
+                    compensated_ptdf_factors = mlodf_factors @ ptdf_bd_i + ptdf_k_i
+                else:
+                    # single branch contingency, no bus contingency
+                    compensated_ptdf_factors = sp.csc_matrix(([], [], [0]), shape=(lodf.shape[0], 0))
 
             else:
                 mlodf_factors = sp.csc_matrix(([], [], [0]), shape=(lodf.shape[0], 0))
-
-            if len(bus_contingency_indices):
-                ptdf_factors = dense_to_csc(mat=ptdf[:, bus_contingency_indices],
-                                            threshold=threshold)
-            else:
-                ptdf_factors = sp.csc_matrix(([], [], [0]), shape=(lodf.shape[0], 0))
+                if len(contingency_indices.bus_contingency_indices) > 0:
+                    # only bus contingencies
+                    compensated_ptdf_factors = ptdf[:, contingency_indices.bus_contingency_indices]
+                else:
+                    # no bus or branch contingencies
+                    compensated_ptdf_factors = sp.csc_matrix(([], [], [0]), shape=(lodf.shape[0], 0))
 
             # append values
             self.multi_contingencies.append(
                 LinearMultiContingency(
-                    branch_indices=branch_contingency_indices,
-                    bus_indices=bus_contingency_indices,
+                    branch_indices=contingency_indices.branch_contingency_indices,
+                    bus_indices=contingency_indices.bus_contingency_indices,
                     mlodf_factors=mlodf_factors,
-                    ptdf_factors=ptdf_factors,
-                    injections_factor=injections_factors
+                    compensated_ptdf_factors=compensated_ptdf_factors,
+                    injections_factor=contingency_indices.injections_factors
                 )
             )
 
@@ -562,9 +661,9 @@ class LinearAnalysis:
                 # no slacks will make it impossible to compute the PTDF analytically
                 if len(island.vd) == 1:
                     if len(island.pqpv) > 0:
-
+                        # island.Btau
                         # compute the PTDF of the island
-                        ptdf_island = make_ptdf(Bbus=island.Bbus,
+                        ptdf_island = make_ptdf(Bpqpv=island.Bpqpv,
                                                 Bf=island.Bf,
                                                 pqpv=island.pqpv,
                                                 distribute_slack=self.distributed_slack)
@@ -591,7 +690,7 @@ class LinearAnalysis:
         else:
 
             # there is only 1 island, compute the PTDF
-            self.PTDF = make_ptdf(Bbus=islands[0].Bbus,
+            self.PTDF = make_ptdf(Bpqpv=islands[0].Bpqpv,
                                   Bf=islands[0].Bf,
                                   pqpv=islands[0].pqpv,
                                   distribute_slack=self.distributed_slack)
