@@ -1,5 +1,5 @@
 # GridCal
-# Copyright (C) 2015 - 2023 Santiago Peñate Vera
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -21,18 +21,14 @@ That means that solves the OPF problem for a complete time series at once
 """
 import numpy as np
 from typing import List, Union, Tuple, Callable
-from GridCalEngine.enumerations import TransformerControlType, HvdcControlType, GenerationNtcFormulation
-
-from GridCalEngine.basic_structures import ZonalGrouping
-from GridCalEngine.basic_structures import MIPSolvers
-from GridCalEngine.Core.Devices.multi_circuit import MultiCircuit
-from GridCalEngine.Core.DataStructures.numerical_circuit import NumericalCircuit, compile_numerical_circuit_at
-from GridCalEngine.Core.DataStructures.generator_data import GeneratorData
-from GridCalEngine.Core.DataStructures.battery_data import BatteryData
-from GridCalEngine.Core.DataStructures.load_data import LoadData
-from GridCalEngine.Core.DataStructures.branch_data import BranchData
-from GridCalEngine.Core.DataStructures.hvdc_data import HvdcData
-from GridCalEngine.Core.DataStructures.bus_data import BusData
+from GridCalEngine.enumerations import MIPSolvers, ZonalGrouping
+from GridCalEngine.Devices.multi_circuit import MultiCircuit
+from GridCalEngine.DataStructures.numerical_circuit import NumericalCircuit, compile_numerical_circuit_at
+from GridCalEngine.DataStructures.generator_data import GeneratorData
+from GridCalEngine.DataStructures.load_data import LoadData
+from GridCalEngine.DataStructures.branch_data import BranchData
+from GridCalEngine.DataStructures.hvdc_data import HvdcData
+from GridCalEngine.DataStructures.bus_data import BusData
 from GridCalEngine.basic_structures import Logger, Vec, IntVec, BoolVec, StrVec, CxMat
 from GridCalEngine.Utils.MIP.selected_interface import LpExp, LpVar, LpModel, lpDot, set_var_bounds, join
 from GridCalEngine.enumerations import TransformerControlType, HvdcControlType, AvailableTransferMode
@@ -40,34 +36,40 @@ from GridCalEngine.Simulations.LinearFactors.linear_analysis import LinearAnalys
 from GridCalEngine.Simulations.ATC.available_transfer_capacity_driver import compute_alpha
 
 
-def get_structural_ntc(inter_area_branches, inter_area_hvdcs, branch_ratings, hvdc_ratings):
-    '''
-
-    :param inter_area_branches:
-    :param inter_area_hvdcs:
-    :param branch_ratings:
-    :param hvdc_ratings:
-    :return:
-    '''
-    if len(inter_area_branches):
-        idx_branch, b = list(zip(*inter_area_branches))
-        idx_branch = list(idx_branch)
-        sum_ratings = sum(branch_ratings[idx_branch])
-    else:
-        sum_ratings = 0.0
-
-    if len(inter_area_hvdcs):
-        idx_hvdc, b = list(zip(*inter_area_hvdcs))
-        idx_hvdc = list(idx_hvdc)
-        sum_ratings += sum(hvdc_ratings[idx_hvdc])
-
-    return sum_ratings
+# def get_structural_ntc(inter_area_branches, inter_area_hvdcs, branch_ratings, hvdc_ratings):
+#     """
+#
+#     :param inter_area_branches:
+#     :param inter_area_hvdcs:
+#     :param branch_ratings:
+#     :param hvdc_ratings:
+#     :return:
+#     """
+#     if len(inter_area_branches):
+#         idx_branch, b = list(zip(*inter_area_branches))
+#         idx_branch = list(idx_branch)
+#         sum_ratings = sum(branch_ratings[idx_branch])
+#     else:
+#         sum_ratings = 0.0
+#
+#     if len(inter_area_hvdcs):
+#         idx_hvdc, b = list(zip(*inter_area_hvdcs))
+#         idx_hvdc = list(idx_hvdc)
+#         sum_ratings += sum(hvdc_ratings[idx_hvdc])
+#
+#     return sum_ratings
 
 
-def formulate_monitorization_logic(
-        monitor_only_sensitive_branches: bool, monitor_only_ntc_load_rule_branches: bool,
-        monitor_loading: BoolVec, alpha: Vec, alpha_n1: Vec, branch_sensitivity_threshold: float,
-        base_flows: Vec, structural_ntc: float, ntc_load_rule: float, rates: Vec) -> Tuple[BoolVec, StrVec, Vec, Vec]:
+def formulate_monitorization_logic(monitor_only_sensitive_branches: bool,
+                                   monitor_only_ntc_load_rule_branches: bool,
+                                   monitor_loading: BoolVec,
+                                   alpha: Vec,
+                                   alpha_n1: Vec,
+                                   branch_sensitivity_threshold: float,
+                                   base_flows: Vec,
+                                   structural_ntc: float,
+                                   ntc_load_rule: float,
+                                   rates: Vec) -> Tuple[BoolVec, StrVec, Vec, Vec]:
     """
     Function to formulate branch monitor status due the given logic
     :param monitor_only_sensitive_branches: boolean to apply sensitivity threshold to the monitorization logic.
@@ -227,6 +229,13 @@ def get_transfer_power_scaling_per_bus(bus_data_t: BusData,
 def get_sensed_proportions(power: Vec,
                            idx: IntVec,
                            logger: Logger) -> Vec:
+    """
+
+    :param power:
+    :param idx:
+    :param logger:
+    :return:
+    """
     nelem = len(power)
 
     # bus area mask
@@ -424,6 +433,13 @@ class BranchNtcVars:
         """
         self.contingency_flow_data.append((t, m, c, flow_var, neg_slack, pos_slack))
 
+    def get_total_flow_slack(self):
+        """
+        Get total flow slacks
+        :return:
+        """
+        return self.flow_slacks_pos - self.flow_slacks_neg
+
 
 class HvdcNtcVars:
     """
@@ -500,6 +516,8 @@ class NtcVars:
     def get_values(self, Sbase: float, model: LpModel) -> "NtcVars":
         """
         Return an instance of this class where the arrays content are not LP vars but their value
+        :param Sbase
+        :param model:
         :return: OpfVars instance
         """
 
@@ -530,6 +548,10 @@ class NtcVars:
         return data
 
     def get_voltages(self) -> CxMat:
+        """
+
+        :return:
+        """
         return np.ones((self.nt, self.nbus)) * np.exp(1j * self.bus_vars.theta)
 
 
@@ -670,7 +692,7 @@ def add_linear_branches_formulation(t_idx: int,
                 bk = 1.0 / branch_data_t.X[m]
 
             # compute the flow
-            if branch_data_t.control_mode[m] == TransformerControlType.Pt:
+            if branch_data_t.control_mode[m] == TransformerControlType.Pf:
 
                 # add angle
                 branch_vars.tap_angles[t_idx, m] = prob.add_var(
@@ -943,6 +965,7 @@ def add_linear_node_balance(t_idx: int,
     Add the kirchoff nodal equality
     :param t_idx: time step
     :param Bbus: susceptance matrix (complete)
+    :param vd: Array of slack indices
     :param bus_data: BusData
     :param bus_vars: BusVars
     :param prob: LpModel
@@ -1026,7 +1049,7 @@ def run_linear_ntc_opf_ts(grid: MultiCircuit,
     nbr = grid.get_branch_number_wo_hvdc()
     ng = grid.get_generators_number()
     nb = grid.get_batteries_number()
-    nl = grid.get_calculation_loads_number()
+    nl = grid.get_load_like_device_number()
     n_hvdc = grid.get_hvdc_number()
 
     lp_model: LpModel = LpModel(solver_type)
@@ -1045,7 +1068,8 @@ def run_linear_ntc_opf_ts(grid: MultiCircuit,
         nc: NumericalCircuit = compile_numerical_circuit_at(circuit=grid,
                                                             t_idx=t,  # yes, this is not a bug
                                                             bus_dict=bus_dict,
-                                                            areas_dict=areas_dict)
+                                                            areas_dict=areas_dict,
+                                                            logger=logger)
 
         # formulate the bus angles ---------------------------------------------------------------------------------
         for k in range(nc.bus_data.nbus):
@@ -1140,7 +1164,7 @@ def run_linear_ntc_opf_ts(grid: MultiCircuit,
 
                 # Compute the more generalistic contingency structures
                 mctg = LinearMultiContingencies(grid=grid)
-                mctg.update(lodf=ls.LODF, ptdf=ls.PTDF, threshold=lodf_threshold)
+                mctg.compute(lodf=ls.LODF, ptdf=ls.PTDF, ptdf_threshold=lodf_threshold, lodf_threshold=lodf_threshold)
 
                 # formulate the contingencies
                 f_obj += add_linear_branches_contingencies_formulation(
@@ -1183,16 +1207,16 @@ def run_linear_ntc_opf_ts(grid: MultiCircuit,
 
     # gather the results
     if status == LpModel.OPTIMAL:
-        print('Solution:')
-        print('Objective value =', lp_model.fobj_value())
+        # print('Solution:')
+        # print('Objective value =', lp_model.fobj_value())
         mip_vars.acceptable_solution = True
     else:
-        print('The problem does not have an optimal solution.')
+        logger.add_error('The problem does not have an optimal solution.')
         mip_vars.acceptable_solution = False
-        lp_file_name = grid.name + "_debug.lp"
-        lp_model.save_model(file_name=lp_file_name)
-        print("Debug LP model saved as:", lp_file_name)
+        # lp_file_name = grid.name + "_debug.lp"
+        # lp_model.save_model(file_name=lp_file_name)
+        # print("Debug LP model saved as:", lp_file_name)
 
-    vars_v = mip_vars.get_values(grid.Sbase)
+    vars_v = mip_vars.get_values(grid.Sbase, model=lp_model)
 
     return vars_v
