@@ -1,5 +1,5 @@
 # GridCal
-# Copyright (C) 2015 - 2023 Santiago Peñate Vera
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -16,11 +16,10 @@
 # Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 import numpy as np
 
-from GridCalEngine.Simulations.result_types import ResultTypes
 from GridCalEngine.Simulations.results_table import ResultsTable
 from GridCalEngine.Simulations.results_template import ResultsTemplate
 from GridCalEngine.basic_structures import DateVec, IntVec, Vec, StrVec, CxMat
-from GridCalEngine.enumerations import StudyResultsType, TransformerControlType
+from GridCalEngine.enumerations import StudyResultsType, TransformerControlType, ResultTypes, DeviceType
 
 
 def add_shifter_data(y, columns, controlled_shifters, phase_shift):
@@ -284,12 +283,6 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
         ResultsTemplate.__init__(self,
                                  name='NTC',
                                  available_results={
-                                     ResultTypes.FlowReports: [
-                                         ResultTypes.ContingencyFlowsReport,
-                                         ResultTypes.ContingencyFlowsBranchReport,
-                                         ResultTypes.ContingencyFlowsGenerationReport,
-                                         ResultTypes.ContingencyFlowsHvdcReport,
-                                     ],
                                      ResultTypes.BusResults: [
                                          ResultTypes.BusVoltageModule,
                                          ResultTypes.BusVoltageAngle,
@@ -303,16 +296,22 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
                                      ResultTypes.HvdcResults: [
                                          ResultTypes.HvdcPowerFrom,
                                      ],
-                                     ResultTypes.DispatchResults: [
-                                         ResultTypes.BatteryPower,
-                                         ResultTypes.GeneratorPower,
-                                         ResultTypes.GenerationDelta,
-                                     ],
+                                     # ResultTypes.DispatchResults: [
+                                     #     ResultTypes.BatteryPower,
+                                     #     ResultTypes.GeneratorPower,
+                                     #     ResultTypes.GenerationDelta,
+                                     # ],
                                      ResultTypes.AreaResults: [
                                          ResultTypes.AvailableTransferCapacityAlpha,
                                          ResultTypes.AvailableTransferCapacityAlphaN1,
                                          ResultTypes.InterAreaExchange,
-                                     ]
+                                     ],
+                                     ResultTypes.FlowReports: [
+                                         ResultTypes.ContingencyFlowsReport,
+                                         ResultTypes.ContingencyFlowsBranchReport,
+                                         ResultTypes.ContingencyFlowsGenerationReport,
+                                         ResultTypes.ContingencyFlowsHvdcReport,
+                                     ],
                                  },
                                  time_array=None,
                                  clustering_results=None,
@@ -401,10 +400,6 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
 
         self.reports = dict()
 
-    def apply_new_rates(self, nc: "NumericalCircuit"):
-        rates = nc.Rates
-        self.loading = self.Sf / (rates + 1e-9)
-
     def initialize(self, n, m):
         """
         Initialize the arrays
@@ -439,7 +434,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
 
     def create_contingency_report(self, loading_threshold=0.98, reverse=True):
 
-        title = f'{ResultTypes.ContingencyFlowsReport.value[0]}. ' \
+        title = f'{ResultTypes.ContingencyFlowsReport.value}. ' \
                 f'Loading threshold: {str(loading_threshold)}. ' \
                 f'Reverse: {str(reverse)}'
 
@@ -498,7 +493,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
         Get flow report
         """
 
-        title = ResultTypes.BranchMonitoring.value[0]
+        title = ResultTypes.BranchMonitoring.value
 
         y = np.array([
             self.monitor,  # Monitor result
@@ -561,7 +556,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
         :param reverse: Boolean to get ordered results. None to keep original .
         """
 
-        title = f'{ResultTypes.BaseFlowReport.value[0]}. ' \
+        title = f'{ResultTypes.BaseFlowReport.value}. ' \
                 f'Loading threshold: {str(loading_threshold)}. ' \
                 f'Reverse: {str(reverse)}'
 
@@ -649,7 +644,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
         :param reverse: Boolean to get ordered results. None to keep original .
         """
 
-        title = f'{ResultTypes.ContingencyFlowsBranchReport.value[0]}. ' \
+        title = f'{ResultTypes.ContingencyFlowsBranchReport.value}. ' \
                 f'Loading threshold: {str(loading_threshold)}. ' \
                 f'Reverse: {str(reverse)}'
 
@@ -757,7 +752,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
         :param loading_threshold: threshold to filter results,
         :param reverse: Boolean to get ordered results. None to keep original .
         """
-        title = f'{ResultTypes.ContingencyFlowsGenerationReport.value[0]}. ' \
+        title = f'{ResultTypes.ContingencyFlowsGenerationReport.value}. ' \
                 f'Loading threshold: {str(loading_threshold)}. ' \
                 f'Reverse: {str(reverse)}'
 
@@ -866,7 +861,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
         :param reverse: Boolean to get ordered results. None to keep original .
         """
 
-        title = f'{ResultTypes.ContingencyFlowsHvdcReport.value[0]}. ' \
+        title = f'{ResultTypes.ContingencyFlowsHvdcReport.value}. ' \
                 f'Loading threshold: {str(loading_threshold)}. ' \
                 f'Reverse: {str(reverse)}'
 
@@ -970,7 +965,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
 
     def create_interarea_exchange_report(self):
 
-        title = ResultTypes.InterAreaExchange.value[0]
+        title = ResultTypes.InterAreaExchange.value
 
         labels = list()
         y = list()
@@ -1021,7 +1016,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
             self.alpha_n1 = None
 
     def get_controlled_shifters_as_pt(self):
-        shifter_idx = np.where(self.branch_control_modes == TransformerControlType.Pt)
+        shifter_idx = np.where(self.branch_control_modes == TransformerControlType.Pf)
         shifter_names = self.branch_names[shifter_idx]
 
         return shifter_idx, shifter_names
@@ -1060,8 +1055,10 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
                 data=np.abs(self.voltage),
                 index=self.bus_names,
                 columns=['V (p.u.)'],
-                title=result_type.value[0],
+                title=str(result_type.value),
                 ylabel='(p.u.)',
+                cols_device_type=DeviceType.NoDevice,
+                idx_device_type=DeviceType.BusDevice
             )
 
         elif result_type == ResultTypes.BusVoltageAngle:
@@ -1069,8 +1066,10 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
                 data=np.angle(self.voltage),
                 index=self.bus_names,
                 columns=['V (radians)'],
-                title=result_type.value[0],
+                title=str(result_type.value),
                 ylabel='(radians)',
+                cols_device_type=DeviceType.NoDevice,
+                idx_device_type=DeviceType.BusDevice
             )
 
         elif result_type == ResultTypes.BranchPower:
@@ -1078,8 +1077,10 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
                 data=self.Sf.real,
                 columns=['Sf'],
                 index=self.branch_names,
-                title=result_type.value[0],
+                title=str(result_type.value),
                 ylabel='(MW)',
+                cols_device_type=DeviceType.NoDevice,
+                idx_device_type=DeviceType.BranchDevice
             )
 
         elif result_type == ResultTypes.BusPower:
@@ -1087,8 +1088,10 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
                 data=self.loading * 100.0,
                 index=self.Sbus.real,
                 columns=['Sb'],
-                title=result_type.value[0],
+                title=str(result_type.value),
                 ylabel='(MW)',
+                cols_device_type=DeviceType.NoDevice,
+                idx_device_type=DeviceType.BusDevice
             )
 
         elif result_type == ResultTypes.BranchLoading:
@@ -1096,8 +1099,10 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
                 data=self.loading * 100.0,
                 index=self.branch_names,
                 columns=['Loading'],
-                title=result_type.value[0],
+                title=str(result_type.value),
                 ylabel='(%)',
+                cols_device_type=DeviceType.NoDevice,
+                idx_device_type=DeviceType.BranchDevice
             )
 
         elif result_type == ResultTypes.BranchLosses:
@@ -1105,8 +1110,10 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
                 data=self.losses.real,
                 index=self.branch_names,
                 columns=['PLosses'],
-                title=result_type.value[0],
+                title=str(result_type.value),
                 ylabel='(MW)',
+                cols_device_type=DeviceType.NoDevice,
+                idx_device_type=DeviceType.BranchDevice
             )
 
         elif result_type == ResultTypes.BranchTapAngle:
@@ -1114,26 +1121,10 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
                 data=np.rad2deg(self.phase_shift),
                 index=self.branch_names,
                 columns=['V (deg)'],
-                title=result_type.value[0],
+                title=str(result_type.value),
                 ylabel='(deg)',
-            )
-
-        elif result_type == ResultTypes.GeneratorPower:
-            return ResultsTable(
-                data=self.generator_power,
-                index=self.generator_names,
-                columns=['P'],
-                title=result_type.value[0],
-                ylabel='(MW)',
-            )
-
-        elif result_type == ResultTypes.BatteryPower:
-            return ResultsTable(
-                data=self.battery_power,
-                index=self.battery_names,
-                columns=['P'],
-                title=result_type.value[0],
-                ylabel='(MW)',
+                cols_device_type=DeviceType.NoDevice,
+                idx_device_type=DeviceType.BranchDevice
             )
 
         elif result_type == ResultTypes.HvdcPowerFrom:
@@ -1141,19 +1132,23 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
                 data=self.hvdc_Pf,
                 index=self.hvdc_names,
                 columns=['Pf'],
-                title=result_type.value[0],
+                title=str(result_type.value),
                 ylabel='(MW)',
+                cols_device_type=DeviceType.NoDevice,
+                idx_device_type=DeviceType.HVDCLineDevice
             )
 
         elif result_type == ResultTypes.AvailableTransferCapacityAlpha:
             return ResultsTable(
                 data=self.alpha,
                 index=self.branch_names,
-                title=result_type.value[0],
+                title=str(result_type.value),
                 columns=['Sensitivity'],
                 ylabel='(p.u.)',
                 xlabel='',
                 units='',
+                cols_device_type=DeviceType.NoDevice,
+                idx_device_type=DeviceType.BranchDevice
             )
 
         elif result_type == ResultTypes.AvailableTransferCapacityAlphaN1:
@@ -1161,21 +1156,12 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
                 data=self.alpha_n1,
                 index=self.branch_names,
                 columns=self.branch_names,
-                title=result_type.value[0],
+                title=str(result_type.value),
                 ylabel='(p.u.)',
                 xlabel='',
                 units='',
-            )
-
-        elif result_type == ResultTypes.GenerationDelta:
-            return ResultsTable(
-                data=self.generation_delta,
-                index=self.generator_names,
-                columns=['P'],
-                title=result_type.value[0],
-                ylabel='(MW)',
-                xlabel='',
-                units='',
+                cols_device_type=DeviceType.BranchDevice,
+                idx_device_type=DeviceType.BranchDevice
             )
 
         elif result_type == ResultTypes.BranchMonitoring:
@@ -1209,18 +1195,14 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
             )
 
         else:
-            return ResultsTable(
-                data=np.zeros(0),
-                index=np.zeros(0).shape[0] * [''],
-                columns=np.zeros(0).shape[0] * [''],
-            )
+            raise Exception(f"Unknown NTC result type {result_type}")
 
     def get_monitoring_logic_report(self):
         """
 
         :return:
         """
-        title = ResultTypes.BranchMonitoring.value[0]
+        title = ResultTypes.BranchMonitoring.value
 
         if title not in self.reports.keys():
             self.create_monitoring_logic_report()
@@ -1234,7 +1216,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
         :param reverse:
         :return:
         """
-        title = f'{ResultTypes.BaseFlowReport.value[0]}. ' \
+        title = f'{ResultTypes.BaseFlowReport.value}. ' \
                 f'Loading threshold: {str(loading_threshold)}. ' \
                 f'Reverse: {str(reverse)}'
 
@@ -1252,7 +1234,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
         :param reverse:
         :return:
         """
-        title = f'{ResultTypes.ContingencyFlowsReport.value[0]}. ' \
+        title = f'{ResultTypes.ContingencyFlowsReport.value}. ' \
                 f'Loading threshold: {str(loading_threshold)}. ' \
                 f'Reverse: {str(reverse)}'
 
@@ -1270,7 +1252,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
         :param reverse:
         :return:
         """
-        title = f'{ResultTypes.ContingencyFlowsBranchReport.value[0]}. ' \
+        title = f'{ResultTypes.ContingencyFlowsBranchReport.value}. ' \
                 f'Loading threshold: {str(loading_threshold)}. ' \
                 f'Reverse: {str(reverse)}'
 
@@ -1288,7 +1270,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
         :param reverse:
         :return:
         """
-        title = f'{ResultTypes.ContingencyFlowsGenerationReport.value[0]}. ' \
+        title = f'{ResultTypes.ContingencyFlowsGenerationReport.value}. ' \
                 f'Loading threshold: {str(loading_threshold)}. ' \
                 f'Reverse: {str(reverse)}'
 
@@ -1307,7 +1289,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
         :param reverse:
         :return:
         """
-        title = f'{ResultTypes.ContingencyFlowsHvdcReport.value[0]}. ' \
+        title = f'{ResultTypes.ContingencyFlowsHvdcReport.value}. ' \
                 f'Loading threshold: {str(loading_threshold)}. ' \
                 f'Reverse: {str(reverse)}'
 
@@ -1324,7 +1306,7 @@ class OptimalNetTransferCapacityResults(ResultsTemplate):
 
         :return:
         """
-        title = ResultTypes.InterAreaExchange.value[0]
+        title = ResultTypes.InterAreaExchange.value
 
         if title not in self.reports.keys():
             self.create_interarea_exchange_report()
