@@ -1,5 +1,5 @@
 # GridCal
-# Copyright (C) 2015 - 2023 Santiago Peñate Vera
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -19,19 +19,18 @@ import datetime
 import numpy as np
 import pandas as pd
 from typing import Union
-from GridCalEngine.basic_structures import TimeGrouping, get_time_groups
-from GridCalEngine.Core.Devices.multi_circuit import MultiCircuit
-from GridCalEngine.basic_structures import SolverType
+from GridCalEngine.Devices.multi_circuit import MultiCircuit
+from GridCalEngine.enumerations import SolverType, TimeGrouping, EngineType, SimulationTypes
 from GridCalEngine.Simulations.OPF.opf_options import OptimalPowerFlowOptions
 from GridCalEngine.Simulations.OPF.linear_opf_ts import run_linear_opf_ts
 from GridCalEngine.Simulations.OPF.simple_dispatch_ts import run_simple_dispatch_ts
+from GridCalEngine.Simulations.OPF.NumericalMethods.ac_opf import run_nonlinear_opf
 from GridCalEngine.Simulations.OPF.opf_ts_results import OptimalPowerFlowTimeSeriesResults
-from GridCalEngine.Simulations.driver_types import SimulationTypes
 from GridCalEngine.Simulations.PowerFlow.power_flow_options import PowerFlowOptions
 from GridCalEngine.Simulations.driver_template import TimeSeriesDriverTemplate
-from GridCalEngine.Core.Compilers.circuit_to_newton_pa import newton_pa_linear_opf, newton_pa_nonlinear_opf
+from GridCalEngine.Compilers.circuit_to_newton_pa import newton_pa_linear_opf, newton_pa_nonlinear_opf
 from GridCalEngine.Simulations.Clustering.clustering_results import ClusteringResults
-import GridCalEngine.basic_structures as bs
+from GridCalEngine.basic_structures import IntVec, Vec, get_time_groups
 
 
 class OptimalPowerFlowTimeSeriesDriver(TimeSeriesDriverTemplate):
@@ -41,9 +40,9 @@ class OptimalPowerFlowTimeSeriesDriver(TimeSeriesDriverTemplate):
     def __init__(self,
                  grid: MultiCircuit,
                  options: Union[OptimalPowerFlowOptions, None] = None,
-                 time_indices: Union[bs.IntVec, None] = None,
+                 time_indices: Union[IntVec, None] = None,
                  clustering_results: Union[ClusteringResults, None] = None,
-                 engine: bs.EngineType = bs.EngineType.GridCal):
+                 engine: EngineType = EngineType.GridCal):
         """
         OptimalPowerFlowTimeSeriesDriver class constructor
         :param grid: MultiCircuit Object
@@ -72,6 +71,11 @@ class OptimalPowerFlowTimeSeriesDriver(TimeSeriesDriverTemplate):
             generator_names=self.grid.get_generator_names(),
             battery_names=self.grid.get_battery_names(),
             hvdc_names=self.grid.get_hvdc_names(),
+            fuel_names=self.grid.get_fuel_names(),
+            emission_names=self.grid.get_emission_names(),
+            fluid_node_names=self.grid.get_fluid_node_names(),
+            fluid_path_names=self.grid.get_fluid_path_names(),
+            fluid_injection_names=self.grid.get_fluid_injection_names(),
             n=self.grid.get_bus_number(),
             m=self.grid.get_branch_number_wo_hvdc(),
             nt=nt,
@@ -79,6 +83,9 @@ class OptimalPowerFlowTimeSeriesDriver(TimeSeriesDriverTemplate):
             nbat=self.grid.get_batteries_number(),
             nload=self.grid.get_loads_number(),
             nhvdc=self.grid.get_hvdc_number(),
+            n_fluid_node=self.grid.get_fluid_nodes_number(),
+            n_fluid_path=self.grid.get_fluid_paths_number(),
+            n_fluid_injection=self.grid.get_fluid_injection_number(),
             time_array=self.grid.time_profile[self.time_indices] if self.time_indices is not None else [datetime.datetime.now()],
             bus_types=np.ones(self.grid.get_bus_number(), dtype=int),
             clustering_results=clustering_results)
@@ -108,10 +115,10 @@ class OptimalPowerFlowTimeSeriesDriver(TimeSeriesDriverTemplate):
         """
 
         if not remote:
-            self.progress_signal.emit(0.0)
-            self.progress_text.emit('Formulating problem...')
+            self.report_progress(0.0)
+            self.report_text('Formulating problem...')
 
-        if self.options.solver == SolverType.DC_OPF:
+        if self.options.solver == SolverType.LINEAR_OPF:
 
             # DC optimal power flow
             opf_vars = run_linear_opf_ts(grid=self.grid,
@@ -120,37 +127,118 @@ class OptimalPowerFlowTimeSeriesDriver(TimeSeriesDriverTemplate):
                                          zonal_grouping=self.options.zonal_grouping,
                                          skip_generation_limits=self.options.skip_generation_limits,
                                          consider_contingencies=self.options.consider_contingencies,
+                                         contingency_groups_used=self.options.contingency_groups_used,
+                                         unit_Commitment=self.options.unit_commitment,
+                                         ramp_constraints=self.options.unit_commitment,
+                                         all_generators_fixed=False,
                                          lodf_threshold=self.options.lodf_tolerance,
                                          maximize_inter_area_flow=self.options.maximize_flows,
                                          areas_from=self.options.areas_from,
                                          areas_to=self.options.areas_to,
                                          logger=self.logger,
-                                         progress_text=self.progress_text.emit,
-                                         progress_func=self.progress_signal.emit,
+                                         progress_text=self.report_text,
+                                         progress_func=self.report_progress,
                                          export_model_fname=self.options.export_model_fname)
 
             self.results.voltage = np.ones((opf_vars.nt, opf_vars.nbus)) * np.exp(1j * opf_vars.bus_vars.theta)
             self.results.bus_shadow_prices = opf_vars.bus_vars.shadow_prices
+
             self.results.load_shedding = opf_vars.load_vars.shedding
+
             self.results.battery_power = opf_vars.batt_vars.p
             self.results.battery_energy = opf_vars.batt_vars.e
+
             self.results.generator_power = opf_vars.gen_vars.p
+            self.results.generator_shedding = opf_vars.gen_vars.shedding
+            self.results.generator_cost = opf_vars.gen_vars.cost
+            # self.results.generator_fuel = opf_vars.gen_vars.fuel
+            # self.results.generator_emissions = opf_vars.gen_vars.emissions
+            self.results.generator_producing = opf_vars.gen_vars.producing
+            self.results.generator_starting_up = opf_vars.gen_vars.starting_up
+            self.results.generator_shutting_down = opf_vars.gen_vars.shedding
+
             self.results.Sf = opf_vars.branch_vars.flows
             self.results.St = -opf_vars.branch_vars.flows
             self.results.overloads = opf_vars.branch_vars.flow_slacks_pos - opf_vars.branch_vars.flow_slacks_neg
             self.results.loading = opf_vars.branch_vars.loading
             self.results.phase_shift = opf_vars.branch_vars.tap_angles
-            # self.results.Sbus = problem.get_power_injections()
+
             self.results.hvdc_Pf = opf_vars.hvdc_vars.flows
             self.results.hvdc_loading = opf_vars.hvdc_vars.loading
 
-        elif self.options.solver == SolverType.Simple_OPF:
+            self.results.fluid_node_current_level = opf_vars.fluid_node_vars.current_level
+            self.results.fluid_node_flow_in = opf_vars.fluid_node_vars.flow_in
+            self.results.fluid_node_flow_out = opf_vars.fluid_node_vars.flow_out
+            self.results.fluid_node_p2x_flow = opf_vars.fluid_node_vars.p2x_flow
+            self.results.fluid_node_spillage = opf_vars.fluid_node_vars.spillage
+            self.results.fluid_path_flow = opf_vars.fluid_path_vars.flow
+            self.results.fluid_injection_flow = opf_vars.fluid_inject_vars.flow
+
+            self.results.system_fuel = opf_vars.sys_vars.system_fuel
+            self.results.system_emissions = opf_vars.sys_vars.system_emissions
+            self.results.system_energy_cost = opf_vars.sys_vars.system_energy_cost
+
+            # set converged for all t to the value of acceptable solution
+            self.results.converged = np.array([opf_vars.acceptable_solution] * opf_vars.nt)
+
+        elif self.options.solver == SolverType.NONLINEAR_OPF:
+
+            self.report_progress(0.0)
+            for it, t in enumerate(self.time_indices):
+
+                # report progress
+                self.report_text('Nonlinear OPF at ' + str(self.grid.time_profile[t]) + '...')
+                self.report_progress2(it, len(self.time_indices))
+
+                # run opf
+                res = run_nonlinear_opf(grid=self.grid,
+                                        opf_options=self.options,
+                                        pf_options=self.pf_options,
+                                        t_idx=t,
+                                        # for the first power flow, use the given strategy
+                                        # for the succesive ones, use the previous solution
+                                        pf_init=self.options.ips_init_with_pf if it == 0 else True,
+                                        Sbus_pf0=self.results.Sbus[it-1, :] if it > 0 else None,
+                                        voltage_pf0=self.results.voltage[it - 1, :] if it > 0 else None,
+                                        logger=self.logger)
+                Sbase = self.grid.Sbase
+                self.results.voltage[it, :] = res.V
+                self.results.Sbus[it, :] = res.S * Sbase
+                self.results.bus_shadow_prices[it, :] = res.lam_p
+                # self.results.load_shedding = npa_res.load_shedding[0, :]
+                # self.results.battery_power = npa_res.battery_p[0, :]
+                # self.results.battery_energy = npa_res.battery_energy[0, :]
+                self.results.generator_power[it, :] = res.Pg * Sbase
+                self.results.generator_cost[it, :] = res.Pcost
+
+                self.results.Sf[it, :] = res.Sf * Sbase
+                self.results.St[it, :] = res.St * Sbase
+                self.results.overloads[it, :] = (res.sl_sf - res.sl_st) * Sbase
+                self.results.loading[it, :] = res.loading
+                self.results.phase_shift[it, :] = res.tap_phase
+
+                self.results.hvdc_Pf[it, :] = res.hvdc_Pf
+                self.results.hvdc_loading[it, :] = res.hvdc_loading
+                self.results.converged[it] = res.converged
+
+                if self.__cancel__:
+                    return self.results
+
+            # Compute the emissions, fuel costs and energy used
+            (self.results.system_fuel,
+             self.results.system_emissions,
+             self.results.system_energy_cost) = self.get_fuel_emissions_energy_calculations(
+                gen_p=self.results.generator_power,
+                gen_cost=self.results.generator_cost
+            )
+
+        elif self.options.solver == SolverType.SIMPLE_OPF:
 
             # AC optimal power flow
             Pl, Pg = run_simple_dispatch_ts(grid=self.grid,
                                             time_indices=self.time_indices,
-                                            text_prog=self.progress_text.emit,
-                                            prog_func=self.progress_signal.emit)
+                                            text_prog=self.report_text,
+                                            prog_func=self.report_progress)
 
             self.results.generator_power[self.time_indices, :] = Pg  # already in MW
 
@@ -159,34 +247,41 @@ class OptimalPowerFlowTimeSeriesDriver(TimeSeriesDriverTemplate):
             return
 
         if not remote:
-            self.progress_signal.emit(0.0)
-            self.progress_text.emit('Running all in an external solver, this may take a while...')
+            self.report_progress(0.0)
+            self.report_text('Running all in an external solver, this may take a while...')
 
         return self.results
 
-    def opf_by_groups(self):
+    def opf_by_groups(self) -> None:
         """
         Run the OPF by groups
         """
 
-        self.progress_signal.emit(0.0)
-        self.progress_text.emit('Making groups...')
+        self.report_progress(0.0)
+        self.report_text('Making groups...')
 
         # get the partition points of the time series
-        groups = get_time_groups(t_array=self.grid.time_profile[self.time_indices], grouping=self.options.grouping)
+        groups = get_time_groups(t_array=self.grid.time_profile[self.time_indices], grouping=self.options.time_grouping)
 
         n = len(groups)
         i = 1
-        energy_0: Union[bs.Vec, None] = None  # at the beginning
+        energy_0: Union[Vec, None] = None  # at the beginning
+        fluid_level_0: Union[Vec, None] = None
 
         while i < n and not self.__cancel__:
             start_ = groups[i - 1]
             end_ = groups[i]
-            time_indices = np.arange(start_, end_)
+
+            # Grab the last time index in the last group
+            if i == n - 1:
+                time_indices = np.arange(start_, end_ + 1)
+            else:
+                time_indices = np.arange(start_, end_)
+
             # show progress message
             print(start_, ':', end_, ' [', end_ - start_, ']')
-            self.progress_text.emit('Running OPF for the time group {0} '
-                                    'start {1} - end {2} in external solver...'.format(i, start_, end_))
+            self.report_text('Running OPF for the time group {0} '
+                             'start {1} - end {2} in external solver...'.format(i, start_, end_))
 
             # run an opf for the group interval only if the group is within the start:end boundaries
             # DC optimal power flow
@@ -196,67 +291,106 @@ class OptimalPowerFlowTimeSeriesDriver(TimeSeriesDriverTemplate):
                                          zonal_grouping=self.options.zonal_grouping,
                                          skip_generation_limits=self.options.skip_generation_limits,
                                          consider_contingencies=self.options.consider_contingencies,
+                                         contingency_groups_used=self.options.contingency_groups_used,
+                                         unit_Commitment=self.options.unit_commitment,
+                                         ramp_constraints=self.options.unit_commitment,
+                                         all_generators_fixed=False,
                                          lodf_threshold=self.options.lodf_tolerance,
                                          maximize_inter_area_flow=self.options.maximize_flows,
                                          areas_from=self.options.areas_from,
                                          areas_to=self.options.areas_to,
                                          energy_0=energy_0,
+                                         fluid_level_0=fluid_level_0,
                                          logger=self.logger,
                                          export_model_fname=self.options.export_model_fname)
 
-            self.results.voltage[time_indices, :] = np.ones((opf_vars.nt, opf_vars.nbus)) * np.exp(1j * opf_vars.bus_vars.theta)
+            self.results.voltage[time_indices, :] = (np.ones((opf_vars.nt, opf_vars.nbus))
+                                                     * np.exp(1j * opf_vars.bus_vars.theta))
             self.results.bus_shadow_prices[time_indices, :] = opf_vars.bus_vars.shadow_prices
+
             self.results.load_shedding[time_indices, :] = opf_vars.load_vars.shedding
+
             self.results.battery_power[time_indices, :] = opf_vars.batt_vars.p
             self.results.battery_energy[time_indices, :] = opf_vars.batt_vars.e
+
             self.results.generator_power[time_indices, :] = opf_vars.gen_vars.p
+            self.results.generator_shedding[time_indices, :] = opf_vars.gen_vars.shedding
+            self.results.generator_cost[time_indices, :] = opf_vars.gen_vars.cost
+            self.results.generator_producing[time_indices, :] = opf_vars.gen_vars.producing
+            self.results.generator_starting_up[time_indices, :] = opf_vars.gen_vars.starting_up
+            self.results.generator_shutting_down[time_indices, :] = opf_vars.gen_vars.shedding
+
             self.results.Sf[time_indices, :] = opf_vars.branch_vars.flows
             self.results.St[time_indices, :] = -opf_vars.branch_vars.flows
-            self.results.overloads[time_indices, :] = opf_vars.branch_vars.flow_slacks_pos - opf_vars.branch_vars.flow_slacks_neg
+            self.results.overloads[time_indices, :] = (opf_vars.branch_vars.flow_slacks_pos
+                                                       - opf_vars.branch_vars.flow_slacks_neg)
             self.results.loading[time_indices, :] = opf_vars.branch_vars.loading
             self.results.phase_shift[time_indices, :] = opf_vars.branch_vars.tap_angles
-            # self.results.Sbus[time_indices, :] = problem.get_power_injections()
+
             self.results.hvdc_Pf[time_indices, :] = opf_vars.hvdc_vars.flows
             self.results.hvdc_loading[time_indices, :] = opf_vars.hvdc_vars.loading
 
+            self.results.fluid_node_current_level[time_indices, :] = opf_vars.fluid_node_vars.current_level
+            self.results.fluid_node_flow_in[time_indices, :] = opf_vars.fluid_node_vars.flow_in
+            self.results.fluid_node_flow_out[time_indices, :] = opf_vars.fluid_node_vars.flow_out
+            self.results.fluid_node_p2x_flow[time_indices, :] = opf_vars.fluid_node_vars.p2x_flow
+            self.results.fluid_node_spillage[time_indices, :] = opf_vars.fluid_node_vars.spillage
+            self.results.fluid_path_flow[time_indices, :] = opf_vars.fluid_path_vars.flow
+            self.results.fluid_injection_flow[time_indices, :] = opf_vars.fluid_inject_vars.flow
+
+            self.results.system_fuel[time_indices, :] = opf_vars.sys_vars.system_fuel
+            self.results.system_emissions[time_indices, :] = opf_vars.sys_vars.system_emissions
+            self.results.system_energy_cost[time_indices] = opf_vars.sys_vars.system_energy_cost
+
+            # set converged for all t to the value of acceptable solution
+            self.results.converged[time_indices] = np.array([opf_vars.acceptable_solution] * opf_vars.nt)
+
             energy_0 = self.results.battery_energy[end_ - 1, :]
+            fluid_level_0 = self.results.fluid_node_current_level[end_ - 1, :]
 
             # update progress bar
-            self.progress_signal.emit((i / len(groups)) * 100)
+            self.report_progress2(i, len(groups))
 
             i += 1
 
-    def add_report(self):
+    def add_report(self, eps: float = 1e-6) -> None:
         """
         Add a report of the results (in-place)
         """
         if self.progress_text:
-            self.progress_text.emit("Creating report")
+            self.report_text("Creating report")
 
         nt = len(self.time_indices)
         for t, t_idx in enumerate(self.time_indices):
 
-            if self.progress_signal:
-                self.progress_signal.emit((t + 1) / nt * 100)
+            self.report_progress2(t, nt)
 
             t_name = str(self.results.time_array[t])
 
             for gen_name, gen_shedding in zip(self.results.generator_names, self.results.generator_shedding[t, :]):
-                if gen_shedding > 0:
+                if gen_shedding > eps:
                     self.logger.add_warning("Generation shedding {}".format(t_name),
                                             device=gen_name,
                                             value=gen_shedding,
                                             expected_value=0.0)
 
             for load_name, load_shedding in zip(self.results.load_names, self.results.load_shedding[t, :]):
-                if load_shedding > 0:
+                if load_shedding > eps:
                     self.logger.add_warning("Load shedding {}".format(t_name),
                                             device=load_name,
                                             value=load_shedding,
                                             expected_value=0.0)
 
+            for fluid_node_name, fluid_node_spillage in zip(self.results.fluid_node_names,
+                                                            self.results.fluid_node_spillage[t, :]):
+                if fluid_node_spillage > eps:
+                    self.logger.add_warning("Fluid node spillage {}".format(t_name),
+                                            device=fluid_node_name,
+                                            value=fluid_node_spillage,
+                                            expected_value=0.0)
+
             for name, val in zip(self.results.branch_names, self.results.loading[t, :]):
-                if val > 1:
+                if val > (1.0 + eps):
                     self.logger.add_warning("Overload {}".format(t_name),
                                             device=name,
                                             value=val * 100,
@@ -283,9 +417,9 @@ class OptimalPowerFlowTimeSeriesDriver(TimeSeriesDriverTemplate):
 
         self.tic()
 
-        if self.engine == bs.EngineType.GridCal:
+        if self.engine == EngineType.GridCal:
 
-            if self.options.grouping == TimeGrouping.NoGrouping:
+            if self.options.time_grouping == TimeGrouping.NoGrouping:
                 self.opf()
             else:
                 if self.time_indices is None:
@@ -296,7 +430,7 @@ class OptimalPowerFlowTimeSeriesDriver(TimeSeriesDriverTemplate):
                     else:
                         self.opf_by_groups()
 
-        elif self.engine == bs.EngineType.NewtonPA:
+        elif self.engine == EngineType.NewtonPA:
 
             if self.time_indices is None:
                 ti = 0
@@ -308,8 +442,8 @@ class OptimalPowerFlowTimeSeriesDriver(TimeSeriesDriverTemplate):
                 else:
                     ti = self.time_indices
 
-            if self.options.solver == SolverType.DC_OPF:
-                self.progress_text.emit('Running Linear OPF with Newton...')
+            if self.options.solver == SolverType.LINEAR_OPF:
+                self.report_text('Running Linear OPF with Newton...')
 
                 npa_res = newton_pa_linear_opf(circuit=self.grid,
                                                opf_options=self.options,
@@ -333,8 +467,16 @@ class OptimalPowerFlowTimeSeriesDriver(TimeSeriesDriverTemplate):
                 self.results.hvdc_Pf[ti, :] = npa_res.hvdc_flows
                 self.results.hvdc_loading[ti, :] = npa_res.hvdc_loading
 
-            if self.options.solver == SolverType.AC_OPF:
-                self.progress_text.emit('Running Non-Linear OPF with Newton...')
+                self.results.fluid_node_current_level[ti, :] = npa_res.fluid_node_vars.current_level
+                self.results.fluid_node_flow_in[ti, :] = npa_res.fluid_node_vars.flow_in
+                self.results.fluid_node_flow_out[ti, :] = npa_res.fluid_node_vars.flow_out
+                self.results.fluid_node_p2x_flow[ti, :] = npa_res.fluid_node_vars.p2x_flow
+                self.results.fluid_node_spillage[ti, :] = npa_res.fluid_node_vars.spillage
+                self.results.fluid_path_flow[ti, :] = npa_res.fluid_path_vars.flow
+                self.results.fluid_injection_flow[ti, :] = npa_res.fluid_inject_vars.flow
+
+            if self.options.solver == SolverType.NONLINEAR_OPF:
+                self.report_text('Running Non-Linear OPF with Newton...')
 
                 # pack the results
                 npa_res = newton_pa_nonlinear_opf(circuit=self.grid,
@@ -360,6 +502,14 @@ class OptimalPowerFlowTimeSeriesDriver(TimeSeriesDriverTemplate):
                 self.results.hvdc_Pf[ti, :] = npa_res.hvdc_Pf
                 self.results.hvdc_loading[ti, :] = npa_res.hvdc_loading
 
+                self.results.fluid_node_current_level[ti, :] = npa_res.fluid_node_vars.current_level
+                self.results.fluid_node_flow_in[ti, :] = npa_res.fluid_node_vars.flow_in
+                self.results.fluid_node_flow_out[ti, :] = npa_res.fluid_node_vars.flow_out
+                self.results.fluid_node_p2x_flow[ti, :] = npa_res.fluid_node_vars.p2x_flow
+                self.results.fluid_node_spillage[ti, :] = npa_res.fluid_node_vars.spillage
+                self.results.fluid_path_flow[ti, :] = npa_res.fluid_path_vars.flow
+                self.results.fluid_injection_flow[ti, :] = npa_res.fluid_inject_vars.flow
+
         if self.options.generate_report:
             self.add_report()
 
