@@ -1,5 +1,5 @@
 # GridCal
-# Copyright (C) 2015 - 2023 Santiago Peñate Vera
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
 # 
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -20,7 +20,7 @@ import numba as nb
 from typing import Tuple
 import scipy.sparse as sp
 from scipy.sparse import lil_matrix, diags, csc_matrix, csr_matrix
-from GridCalEngine.basic_structures import Vec, CxVec, IntVec
+from GridCalEngine.basic_structures import CxVec, IntVec
 
 
 def dSbus_dV(Ybus: csc_matrix, V: CxVec) -> Tuple[csc_matrix, csc_matrix]:
@@ -42,13 +42,14 @@ def dSbus_dV(Ybus: csc_matrix, V: CxVec) -> Tuple[csc_matrix, csc_matrix]:
 
 
 @nb.njit(cache=True)
-def dSbus_dV_numba_sparse_csc(Yx: CxVec, Yp: IntVec, Yi: IntVec, V: CxVec, E: CxVec) -> Tuple[CxVec, CxVec]:
+def dSbus_dV_numba_sparse_csc(Yx: CxVec, Yp: IntVec, Yi: IntVec, V: CxVec, Vm: CxVec) -> Tuple[CxVec, CxVec]:
     """
     Compute the power injection derivatives w.r.t the voltage module and angle
     :param Yx: data of Ybus in CSC format
     :param Yp: indptr of Ybus in CSC format
     :param Yi: indices of Ybus in CSC format
     :param V: Voltages vector
+    :param Vm: voltage modules vector
     :return: dS_dVm, dS_dVa data ordered in the CSC format to match the indices of Ybus
     """
 
@@ -66,24 +67,32 @@ def dSbus_dV_numba_sparse_csc(Yx: CxVec, Yp: IntVec, Yi: IntVec, V: CxVec, E: Cx
 
     # init buffer vector
     n = len(Yp) - 1
-    Ibus = np.zeros(n, dtype=np.complex128)
+    Ibus = np.zeros(n, dtype=nb.complex128)
     dS_dVm = Yx.copy()
     dS_dVa = Yx.copy()
+    E = V.copy()
 
     # pass 1: perform the matrix-vector products
     for j in range(n):  # for each column ...
+
+        # compute the unitary vector of the voltage
+        E[j] /= Vm[j]
+
         for k in range(Yp[j], Yp[j + 1]):  # for each row ...
             # row index
             i = Yi[k]
 
             # Ibus = Ybus * V
-            Ibus[i] += Yx[k] * V[j]  # Yx[k] -> Y(i,j)
+            I = Yx[k] * V[j]
+
+            # store in the Ibus vector
+            Ibus[i] += I  # Yx[k] -> Y(i,j)
 
             # Ybus * diagE
             dS_dVm[k] = Yx[k] * E[j]
 
-            # Ybus * diag(V)
-            dS_dVa[k] = Yx[k] * V[j]
+            # - Ybus * diag(V)
+            dS_dVa[k] = -I
 
     # pass 2: finalize the operations
     for j in range(n):  # for each column ...
@@ -103,17 +112,18 @@ def dSbus_dV_numba_sparse_csc(Yx: CxVec, Yp: IntVec, Yi: IntVec, V: CxVec, E: Cx
 
             if j == i:
                 # diagonal elements
-                dS_dVa[k] -= Ibus[j]
-                dS_dVm[k] += buffer
+                dS_dVa[k] += Ibus[j]  # diagIbus, after this it contains: diagIbus - Ybus * diagV
+                dS_dVm[k] += buffer  # conj(I(j)) * E(j), after this it contains; diag(V) * conj(Ybus * diagE) + conj(diagIbus) * diagE
 
             # 1j * diagV * conj(diagIbus - Ybus * diagV)
-            dS_dVa[k] = np.conj(-dS_dVa[k]) * (1j * V[i])
+            dS_dVa[k] = (1j * V[i]) * np.conj(dS_dVa[k])
 
     return dS_dVm, dS_dVa
 
 
 @nb.jit(nopython=True, cache=True)
-def dSbus_dV_numba_sparse_csr(Yx: CxVec, Yp: IntVec, Yj: IntVec, V: CxVec, E: CxVec) -> Tuple[CxVec, CxVec]:  # pragma: no cover
+def dSbus_dV_numba_sparse_csr(Yx: CxVec, Yp: IntVec, Yj: IntVec, V: CxVec, E: CxVec) -> Tuple[
+    CxVec, CxVec]:  # pragma: no cover
     """
     partial derivatives of power injection w.r.t. voltage.
     :param Yx: Ybus data in CSC format
@@ -168,16 +178,16 @@ def dSbus_dV_numba_sparse_csr(Yx: CxVec, Yp: IntVec, Yj: IntVec, V: CxVec, E: Cx
     return dS_dVm, dS_dVa
 
 
-def dSbus_dV_csc(Ybus: csc_matrix, V: CxVec, E: CxVec) -> Tuple[csc_matrix, csc_matrix]:
+def dSbus_dV_csc(Ybus: csc_matrix, V: CxVec, Vm) -> Tuple[csc_matrix, csc_matrix]:
     """
     Call the numba sparse constructor of the derivatives
     :param Ybus: Ybus in CSC format
     :param V: Voltages vector
-    :param E: Voltages unitary vector
+    :param Vm: Voltages modules
     :return: dS_dVm, dS_dVa in CSC format
     """
     # compute the derivatives' data fast
-    dS_dVm, dS_dVa = dSbus_dV_numba_sparse_csc(Ybus.data, Ybus.indptr, Ybus.indices, V, E)
+    dS_dVm, dS_dVa = dSbus_dV_numba_sparse_csc(Ybus.data, Ybus.indptr, Ybus.indices, V, Vm)
 
     # generate sparse CSC matrices with computed data and return them
     return (sp.csc_matrix((dS_dVa, Ybus.indices, Ybus.indptr)),
@@ -437,8 +447,8 @@ def dSf_dV_csc(Yf, V, F, T):
                                               idx_f=idx_f,
                                               idx_t=idx_t)
 
-    return csc_matrix((dSf_dVm_data, Yf.indices, Yf.indptr), shape=Yf.shape), \
-           csc_matrix((dSf_dVa_data, Yf.indices, Yf.indptr), shape=Yf.shape)
+    return (csc_matrix((dSf_dVm_data, Yf.indices, Yf.indptr), shape=Yf.shape),
+            csc_matrix((dSf_dVa_data, Yf.indices, Yf.indptr), shape=Yf.shape))
 
 
 def dSt_dV_csc(Yt, V, F, T):
@@ -468,8 +478,8 @@ def dSt_dV_csc(Yt, V, F, T):
                                               idx_f=idx_f,
                                               idx_t=idx_t)
 
-    return csc_matrix((dSt_dVm_data, Yt.indices, Yt.indptr), shape=Yt.shape), \
-           csc_matrix((dSt_dVa_data, Yt.indices, Yt.indptr), shape=Yt.shape)
+    return (csc_matrix((dSt_dVm_data, Yt.indices, Yt.indptr), shape=Yt.shape),
+            csc_matrix((dSt_dVa_data, Yt.indices, Yt.indptr), shape=Yt.shape))
 
 
 # ----------------------------------------------------------------------------------------------------------------------
@@ -594,9 +604,9 @@ def derivatives_tau_csc_numba(iPxsh, F: IntVec, T: IntVec, Ys: CxVec, k2, tap, V
     dSf_dsh_indptr[ndev] = ndev
     dSt_dsh_indptr[ndev] = ndev
 
-    return dSbus_dsh_data, dSbus_dsh_indices, dSbus_dsh_indptr, \
-           dSf_dsh_data, dSf_dsh_indices, dSf_dsh_indptr, \
-           dSt_dsh_data, dSt_dsh_indices, dSt_dsh_indptr
+    return (dSbus_dsh_data, dSbus_dsh_indices, dSbus_dsh_indptr,
+            dSf_dsh_data, dSf_dsh_indices, dSf_dsh_indptr,
+            dSt_dsh_data, dSt_dsh_indices, dSt_dsh_indptr)
 
 
 def derivatives_sh_csc_fast(nb, nl, iPxsh, F, T, Ys, k2, tap, V):
@@ -620,9 +630,9 @@ def derivatives_sh_csc_fast(nb, nl, iPxsh, F, T, Ys, k2, tap, V):
     """
     ndev = len(iPxsh)
 
-    dSbus_dsh_data, dSbus_dsh_indices, dSbus_dsh_indptr, \
-    dSf_dsh_data, dSf_dsh_indices, dSf_dsh_indptr, \
-    dSt_dsh_data, dSt_dsh_indices, dSt_dsh_indptr = derivatives_tau_csc_numba(iPxsh, F, T, Ys, k2, tap, V)
+    (dSbus_dsh_data, dSbus_dsh_indices, dSbus_dsh_indptr,
+     dSf_dsh_data, dSf_dsh_indices, dSf_dsh_indptr,
+     dSt_dsh_data, dSt_dsh_indices, dSt_dsh_indptr) = derivatives_tau_csc_numba(iPxsh, F, T, Ys, k2, tap, V)
 
     dSbus_dsh = sp.csc_matrix((dSbus_dsh_data, dSbus_dsh_indices, dSbus_dsh_indptr), shape=(nb, ndev))
     dSf_dsh = sp.csc_matrix((dSf_dsh_data, dSf_dsh_indices, dSf_dsh_indptr), shape=(nl, ndev))
@@ -765,9 +775,9 @@ def derivatives_ma_csc_numba(iXxma, F, T, Ys, k2, tap, ma, Bc, Beq, V):
     dSf_dma_indptr[ndev] = ndev
     dSt_dma_indptr[ndev] = ndev
 
-    return dSbus_dma_data, dSbus_dma_indices, dSbus_dma_indptr, \
-           dSf_dma_data, dSf_dma_indices, dSf_dma_indptr, \
-           dSt_dma_data, dSt_dma_indices, dSt_dma_indptr
+    return (dSbus_dma_data, dSbus_dma_indices, dSbus_dma_indptr,
+            dSf_dma_data, dSf_dma_indices, dSf_dma_indptr,
+            dSt_dma_data, dSt_dma_indices, dSt_dma_indptr)
 
 
 def derivatives_ma_csc_fast(nb, nl, iXxma, F, T, Ys, k2, tap, ma, Bc, Beq, V):
@@ -798,9 +808,9 @@ def derivatives_ma_csc_fast(nb, nl, iXxma, F, T, Ys, k2, tap, ma, Bc, Beq, V):
     # Declare the derivative
     ndev = len(iXxma)
 
-    dSbus_dma_data, dSbus_dma_indices, dSbus_dma_indptr, \
-    dSf_dma_data, dSf_dma_indices, dSf_dma_indptr, \
-    dSt_dma_data, dSt_dma_indices, dSt_dma_indptr = derivatives_ma_csc_numba(iXxma, F, T, Ys, k2, tap, ma, Bc, Beq, V)
+    (dSbus_dma_data, dSbus_dma_indices, dSbus_dma_indptr,
+     dSf_dma_data, dSf_dma_indices, dSf_dma_indptr,
+     dSt_dma_data, dSt_dma_indices, dSt_dma_indptr) = derivatives_ma_csc_numba(iXxma, F, T, Ys, k2, tap, ma, Bc, Beq, V)
 
     dSbus_dma = sp.csc_matrix((dSbus_dma_data, dSbus_dma_indices, dSbus_dma_indptr), shape=(nb, ndev))
     dSf_dma = sp.csc_matrix((dSf_dma_data, dSf_dma_indices, dSf_dma_indptr), shape=(nl, ndev))
@@ -945,12 +955,11 @@ def derivatives_Beq_csc_fast(nb, nl, iBeqx, F, T, V, ma, k2):
 
     ndev = len(iBeqx)
 
-    dSbus_dBeq_data, dSbus_dBeq_indices, dSbus_dBeq_indptr, \
-    dSf_dBeqx_data, dSf_dBeqx_indices, dSf_dBeqx_indptr = derivatives_Beq_csc_numba(iBeqx, F, V, ma, k2)
+    (dSbus_dBeq_data, dSbus_dBeq_indices, dSbus_dBeq_indptr,
+     dSf_dBeqx_data, dSf_dBeqx_indices, dSf_dBeqx_indptr) = derivatives_Beq_csc_numba(iBeqx, F, V, ma, k2)
 
     dSbus_dBeqx = sp.csc_matrix((dSbus_dBeq_data, dSbus_dBeq_indices, dSbus_dBeq_indptr), shape=(nb, ndev))
     dSf_dBeqx = sp.csc_matrix((dSf_dBeqx_data, dSf_dBeqx_indices, dSf_dBeqx_indptr), shape=(nl, ndev))
     dSt_dBeqx = sp.csc_matrix((nl, ndev), dtype=complex)
 
     return dSbus_dBeqx, dSf_dBeqx, dSt_dBeqx
-
