@@ -1,5 +1,5 @@
 # GridCal
-# Copyright (C) 2015 - 2023 Santiago Peñate Vera
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -212,7 +212,6 @@ def conv1(A, B, c):
     :param A: Coefficients matrix 1 (orders, buses)
     :param B: Coefficients matrix 2 (orders, buses)
     :param c: order of the coefficients
-    :param indices: bus indices array
     :return: Array with the convolution for the buses given by "indices"
     """
     suma = np.zeros(A.shape[1], dtype=nb.complex128)
@@ -320,7 +319,7 @@ def helm_coefficients_josep(Ybus, Yseries, V0, S0, Ysh0, pq, pv, sl, pqpv, toler
 
     pq_ = pq - nsl_counted[pq]
     pv_ = pv - nsl_counted[pv]
-    pqpv_ = np.sort(np.r_[pq_, pv_])
+    no_slack_ = np.sort(np.r_[pq_, pv_])
 
     # .......................CALCULATION OF TERMS [0] ------------------------------------------------------------------
 
@@ -335,16 +334,23 @@ def helm_coefficients_josep(Ybus, Yseries, V0, S0, Ysh0, pq, pv, sl, pqpv, toler
     valor = np.zeros(npqpv, dtype=complex)
 
     # get the current Injections that appear due to the slack buses reduction
-    I_inj_slack = Yslack[pqpv_, :] * Vslack
+    I_inj_slack = Yslack[no_slack_, :] * Vslack
 
-    valor[pq_] = I_inj_slack[pq_] - Yslack[pq_].sum(axis=1).A1 + (vec_P[pq_] - vec_Q[pq_] * 1j) * X[0, pq_] - U[
-        0, pq_] * Ysh[pq_]
-    valor[pv_] = I_inj_slack[pv_] - Yslack[pv_].sum(axis=1).A1 + (vec_P[pv_]) * X[0, pv_] - U[0, pv_] * Ysh[pv_]
+    valor[pq_] = (I_inj_slack[pq_]
+                  - Yslack[pq_].sum(axis=1).A1
+                  + (vec_P[pq_] - vec_Q[pq_] * 1j) * X[0, pq_]
+                  - U[0, pq_] * Ysh[pq_])
+
+    valor[pv_] = (I_inj_slack[pv_]
+                  - Yslack[pv_].sum(axis=1).A1
+                  + (vec_P[pv_]) * X[0, pv_]
+                  - U[0, pv_] * Ysh[pv_])
 
     # compose the right-hand side vector
-    RHS = np.r_[valor.real,
-    valor.imag,
-    vec_W[pv_] - (U[0, pv_] * U[0, pv_]).real  # vec_W[pv_] - 1.0
+    RHS = np.r_[
+        valor.real,
+        valor.imag,
+        vec_W[pv_] - (U[0, pv_] * U[0, pv_]).real  # vec_W[pv_] - 1.0
     ]
 
     # Form the system matrix (MAT)
