@@ -1,5 +1,5 @@
 # GridCal
-# Copyright (C) 2015 - 2023 Santiago Peñate Vera
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -19,16 +19,17 @@ import time
 
 import numpy as np
 
-from GridCalEngine.Core.DataStructures.numerical_circuit import NumericalCircuit
-from GridCalEngine.Core.admittance_matrices import compile_y_acdc
+from GridCalEngine.DataStructures.numerical_circuit import NumericalCircuit
+from GridCalEngine.Topology.admittance_matrices import compile_y_acdc
 from GridCalEngine.Simulations.PowerFlow.power_flow_results import NumericPowerFlowResults
 from GridCalEngine.Simulations.PowerFlow.NumericalMethods.discrete_controls import control_q_inside_method
 from GridCalEngine.Simulations.PowerFlow.NumericalMethods.acdc_jacobian import fubm_jacobian, AcDcSolSlicer
 from GridCalEngine.Simulations.PowerFlow.NumericalMethods.common_functions import (compute_acdc_fx,
                                                                                    compute_converter_losses,
                                                                                    compute_power, compute_zip_power)
-from GridCalEngine.basic_structures import ReactivePowerControlMode, CxVec
-import GridCalEngine.Simulations.sparse_solve as gcsp
+from GridCalEngine.basic_structures import CxVec
+from GridCalEngine.enumerations import ReactivePowerControlMode
+import GridCalEngine.Utils.NumericalMethods.sparse_solve as gcsp
 
 
 def NR_LS_ACDC(nc: NumericalCircuit,
@@ -122,7 +123,7 @@ def NR_LS_ACDC(nc: NumericalCircuit,
     # compute initial admittances
     Ybus, Yf, Yt, tap = compile_y_acdc(Cf=Cf, Ct=Ct,
                                        C_bus_shunt=nc.shunt_data.C_bus_elm.tocsc(),
-                                       shunt_admittance=nc.shunt_data.admittance,
+                                       shunt_admittance=nc.shunt_data.Y,
                                        shunt_active=nc.shunt_data.active,
                                        ys=Ys,
                                        B=Bc,
@@ -224,7 +225,7 @@ def NR_LS_ACDC(nc: NumericalCircuit,
                 # compute admittances
                 Ybus, Yf, Yt, tap = compile_y_acdc(Cf=Cf, Ct=Ct,
                                                    C_bus_shunt=nc.shunt_data.C_bus_elm.tocsc(),
-                                                   shunt_admittance=nc.shunt_data.admittance,
+                                                   shunt_admittance=nc.shunt_data.Y,
                                                    shunt_active=nc.shunt_data.active,
                                                    ys=Ys,
                                                    B=Bc,
@@ -323,16 +324,14 @@ def NR_LS_ACDC(nc: NumericalCircuit,
                         # check and adjust the reactive power
                         # this function passes pv buses to pq when the limits are violated,
                         # but not pq to pv because that is unstable
-                        n_changes, Scalc, S0, pv, pq, pvpq, messages = control_q_inside_method(Scalc, S0, pv, pq,
-                                                                                               pvpq, Qmin, Qmax)
+                        changed, pv, pq, pqv, p = control_q_inside_method(Scalc, S0, pv, pq, pqv, p, Qmin, Qmax)
 
-                        # compute the ZIP power injection
-                        Sbus = compute_zip_power(S0=S0, I0=I0, Y0=Y0, Vm=Vm)
-
-                        if n_changes > 0:
+                        if len(changed) > 0:
                             # adjust internal variables to the new pq|pv values
-                            npv = len(pv)
-                            npq = len(pq)
+                            blck1_idx = np.r_[pv, pq, p, pqv]
+                            blck2_idx = np.r_[pq, p]
+                            blck3_idx = np.r_[pq, pqv]
+                            n_block1 = len(blck1_idx)
 
                             # re declare the slicer because the indices of pq and pv changed
                             sol_slicer = AcDcSolSlicer(pvpq=pvpq,
@@ -369,11 +368,6 @@ def NR_LS_ACDC(nc: NumericalCircuit,
                                                  i_vt_m=nc.i_vt_m)
                             norm_f_new = np.max(np.abs(fx))
 
-                            # if verbose > 0:
-                            #     for sense, idx, var in messages:
-                            #         msg = "Bus " + str(idx) + " changed to PQ, limited to " + str(var * 100) + " MVAr"
-                            #         logger.add_debug(msg)
-
                 # set the mismatch to the new mismatch
                 norm_f = norm_f_new
 
