@@ -1,5 +1,5 @@
 # GridCal
-# Copyright (C) 2015 - 2023 Santiago Peñate Vera
+# Copyright (C) 2015 - 2024 Santiago Peñate Vera
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -14,39 +14,47 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with this program; if not, write to the Free Software Foundation,
 # Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
-
+from __future__ import annotations
 import numpy as np
-from typing import Union, Dict, Tuple
+from typing import Union, Dict, Tuple, TYPE_CHECKING
 
 import GridCalEngine.Simulations.PowerFlow as pflw
-from GridCalEngine.basic_structures import Logger, ConvergenceReport, SolverType
+from GridCalEngine.enumerations import SolverType
+from GridCalEngine.basic_structures import Logger, ConvergenceReport
 from GridCalEngine.Simulations.PowerFlow.power_flow_results import PowerFlowResults
 from GridCalEngine.Simulations.PowerFlow.power_flow_options import PowerFlowOptions
 from GridCalEngine.Simulations.PowerFlow.power_flow_results import NumericPowerFlowResults
-from GridCalEngine.Core.DataStructures.numerical_circuit import NumericalCircuit
-from GridCalEngine.Core.Devices.multi_circuit import MultiCircuit
-from GridCalEngine.Core.DataStructures.numerical_circuit import compile_numerical_circuit_at
-from GridCalEngine.Core.Devices.Substation.bus import Bus
-from GridCalEngine.Core.Devices.Aggregation.area import Area
-from GridCalEngine.basic_structures import CxVec, Vec, IntVec
+from GridCalEngine.DataStructures.numerical_circuit import NumericalCircuit
+from GridCalEngine.Devices.multi_circuit import MultiCircuit
+from GridCalEngine.DataStructures.numerical_circuit import compile_numerical_circuit_at
+from GridCalEngine.DataStructures.numerical_circuit import (
+    compile_numerical_circuit_at as compile_numerical_circuit_at_generalised_pf)
+from GridCalEngine.Devices.Substation.bus import Bus
+from GridCalEngine.Devices.Aggregation.area import Area
+from GridCalEngine.basic_structures import CxVec, Vec, IntVec, CscMat
+
+if TYPE_CHECKING:  # Only imports the below statements during type checking
+    from GridCalEngine.Simulations.OPF.opf_results import OptimalPowerFlowResults
 
 
 def solve(circuit: NumericalCircuit,
           options: PowerFlowOptions,
           report: ConvergenceReport,
-          V0: CxVec, 
-          S0: CxVec, 
-          I0: CxVec, 
+          V0: CxVec,
+          S0: CxVec,
+          I0: CxVec,
           Y0: CxVec,
-          tap_modules: Vec, 
-          tap_angles: Vec, 
+          tap_modules: Vec,
+          tap_angles: Vec,
           Beq: Vec,
-          pq: IntVec, 
-          pv: IntVec, 
-          ref: IntVec, 
+          pq: IntVec,
+          pv: IntVec,
+          pqv: IntVec,
+          p: IntVec,
+          ref: IntVec,
           pqpv: IntVec,
-          Qmin: Vec, 
-          Qmax: Vec, 
+          Qmin: Vec,
+          Qmax: Vec,
           logger=Logger()) -> NumericPowerFlowResults:
     """
     Run a power flow simulation using the selected method (no outer loop controls).
@@ -62,6 +70,8 @@ def solve(circuit: NumericalCircuit,
     :param Beq: Array of branch equivalent susceptances
     :param pq: Array of pq nodes
     :param pv: Array of pv nodes
+    :param pqv: Array of pqv nodes
+    :param p: Array of p values
     :param ref: Array of slack nodes
     :param pqpv: Array of (sorted) pq and pv nodes
     :param Qmin: Array of minimum reactive power capability per bus
@@ -136,12 +146,13 @@ def solve(circuit: NumericalCircuit,
             solution = pflw.dcpf(Ybus=circuit.Ybus,
                                  Bpqpv=circuit.Bpqpv,
                                  Bref=circuit.Bref,
-                                 Btheta=circuit.Btheta,
+                                 Bf=circuit.Bf,
                                  S0=S0,
                                  I0=I0,
+                                 Y0=Y0,
                                  V0=V0,
-                                 theta=tap_angles,
-                                 ref=ref,
+                                 tau=tap_angles,
+                                 vd=ref,
                                  pvpq=pqpv,
                                  pq=pq,
                                  pv=pv)
@@ -188,6 +199,8 @@ def solve(circuit: NumericalCircuit,
                                                        Y0=Y0,
                                                        pv_=pv,
                                                        pq_=pq,
+                                                       pqv_=pqv,
+                                                       p_=p,
                                                        Qmin=Qmin,
                                                        Qmax=Qmax,
                                                        tol=options.tolerance,
@@ -205,15 +218,18 @@ def solve(circuit: NumericalCircuit,
                                  Ybus=circuit.Ybus,
                                  B1=circuit.B1,
                                  B2=circuit.B2,
-                                 pq=pq,
-                                 pv=pv,
-                                 pqpv=pqpv,
+                                 pv_=pv,
+                                 pq_=pq,
+                                 pqv_=pqv,
+                                 p_=p,
+                                 Qmin=Qmin,
+                                 Qmax=Qmax,
                                  tol=options.tolerance,
-                                 max_it=options.max_iter)
+                                 max_it=options.max_iter,
+                                 control_q=options.control_Q)
 
         # Newton-Raphson (full)
         elif solver_type == SolverType.NR:
-
             if circuit.any_control:
                 # Solve NR with the AC/DC algorithm
                 solution = pflw.NR_LS_ACDC(nc=circuit,
@@ -224,7 +240,7 @@ def solve(circuit: NumericalCircuit,
                                            tolerance=options.tolerance,
                                            max_iter=options.max_iter,
                                            acceleration_parameter=options.backtracking_parameter,
-                                           mu_0=options.mu,
+                                           mu_0=options.trust_radius,
                                            control_q=options.control_Q)
             else:
                 # Solve NR with the AC algorithm
@@ -235,11 +251,13 @@ def solve(circuit: NumericalCircuit,
                                       Y0=Y0,
                                       pv_=pv,
                                       pq_=pq,
+                                      pqv_=pqv,
+                                      p_=p,
                                       Qmin=Qmin,
                                       Qmax=Qmax,
                                       tol=options.tolerance,
                                       max_it=options.max_iter,
-                                      mu_0=options.mu,
+                                      mu_0=options.trust_radius,
                                       acceleration_parameter=options.backtracking_parameter,
                                       control_q=options.control_Q,
                                       verbose=options.verbose,
@@ -268,12 +286,15 @@ def solve(circuit: NumericalCircuit,
                                       Y0=Y0,
                                       pv_=pv,
                                       pq_=pq,
+                                      pqv_=pqv,
+                                      p_=p,
                                       Qmin=Qmin,
                                       Qmax=Qmax,
                                       tol=options.tolerance,
                                       max_it=options.max_iter,
                                       control_q=options.control_Q,
-                                      robust=True)
+                                      robust=True,
+                                      logger=logger)
 
         # Newton-Raphson in current equations
         elif solver_type == SolverType.NRI:
@@ -325,8 +346,22 @@ def solve(circuit: NumericalCircuit,
 
 
 def single_island_pf(circuit: NumericalCircuit, options: PowerFlowOptions,
-                     voltage_solution, S0, I0, Y0, tap_modules, tap_angles, Beq, branch_rates,
-                     pq, pv, vd, pqpv, Qmin, Qmax, logger=Logger()) -> "PowerFlowResults":
+                     voltage_solution: CxVec,
+                     S0: CxVec,
+                     I0: CxVec,
+                     Y0: CxVec,
+                     tap_modules: Vec,
+                     tap_angles: Vec,
+                     Beq: Vec,
+                     branch_rates: Vec,
+                     pq: IntVec,
+                     pv: IntVec,
+                     pqv: IntVec,
+                     p: IntVec,
+                     vd: IntVec,
+                     pqpv: IntVec,
+                     Qmin: Vec,
+                     Qmax: Vec, logger=Logger()) -> "PowerFlowResults":
     """
     Run a power flow simulation for a single circuit using the
     selected outer loop controls.
@@ -343,6 +378,8 @@ def single_island_pf(circuit: NumericalCircuit, options: PowerFlowOptions,
     :param branch_rates: Array of branch rates
     :param pq: Array of pq nodes
     :param pv: Array of pv nodes
+    :param pqv: Array of pqv nodes
+    :param p: Array of p nodes
     :param vd: Array of slack nodes
     :param pqpv: Array of (sorted) pq and pv nodes
     :param Qmin: Array of minimum reactive power capability per bus
@@ -388,6 +425,8 @@ def single_island_pf(circuit: NumericalCircuit, options: PowerFlowOptions,
                          Beq=Beq,
                          pq=pq,
                          pv=pv,
+                         pqv=pqv,
+                         p=p,
                          ref=vd,
                          pqpv=pqpv,
                          Qmin=Qmin,
@@ -415,6 +454,8 @@ def single_island_pf(circuit: NumericalCircuit, options: PowerFlowOptions,
                                  Beq=Beq,
                                  pq=pq,
                                  pv=pv,
+                                 pqv=pqv,
+                                 p=p,
                                  ref=vd,
                                  pqpv=pqpv,
                                  Qmin=Qmin,
@@ -426,6 +467,7 @@ def single_island_pf(circuit: NumericalCircuit, options: PowerFlowOptions,
                                                                              Sbus=solution.Scalc,
                                                                              V=solution.V,
                                                                              branch_rates=branch_rates,
+                                                                             Ybus=solution.Ybus,
                                                                              Yf=solution.Yf,
                                                                              Yt=solution.Yt,
                                                                              method=solution.method)
@@ -459,18 +501,23 @@ def single_island_pf(circuit: NumericalCircuit, options: PowerFlowOptions,
     return results
 
 
-def power_flow_post_process(calculation_inputs: NumericalCircuit,
-                            Sbus: CxVec,
-                            V: CxVec,
-                            branch_rates: CxVec,
-                            Yf=None, Yt=None,
-                            method: SolverType = None) -> Tuple[CxVec, CxVec, CxVec, CxVec, CxVec, CxVec, CxVec, CxVec]:
+def power_flow_post_process(
+        calculation_inputs: NumericalCircuit,
+        Sbus: CxVec,
+        V: CxVec,
+        branch_rates: CxVec,
+        Ybus: Union[CscMat, None] = None,
+        Yf: Union[CscMat, None] = None,
+        Yt: Union[CscMat, None] = None,
+        method: Union[None, SolverType] = None
+) -> Tuple[CxVec, CxVec, CxVec, CxVec, CxVec, CxVec, CxVec, CxVec]:
     """
     Compute the power Sf trough the Branches.
     :param calculation_inputs: NumericalCircuit
     :param Sbus: Array of computed nodal injections
     :param V: Array of computed nodal voltages
     :param branch_rates: Array of branch rates
+    :param Ybus: Admittance matrix
     :param Yf: Admittance-from matrix
     :param Yt: Admittance-to matrix
     :param method: SolverType (the non-linear and Linear flow calculations differ)
@@ -480,26 +527,27 @@ def power_flow_post_process(calculation_inputs: NumericalCircuit,
     vd = calculation_inputs.vd
     pv = calculation_inputs.pv
 
-    Vf = calculation_inputs.Cf * V
-    Vt = calculation_inputs.Ct * V
-
     if method not in [SolverType.DC]:
-        # power at the slack nodes
-        Sbus[vd] = V[vd] * np.conj(calculation_inputs.Ybus[vd, :].dot(V))
-
-        # Reactive power at the pv nodes
-        P = Sbus[pv].real
-        Q = (V[pv] * np.conj(calculation_inputs.Ybus[pv, :].dot(V))).imag
-        Sbus[pv] = P + 1j * Q  # keep the original P injection and set the calculated reactive power
-
+        if Ybus is None:
+            Ybus = calculation_inputs.Ybus
         if Yf is None:
             Yf = calculation_inputs.Yf
         if Yt is None:
             Yt = calculation_inputs.Yt
 
+        # power at the slack nodes
+        Sbus[vd] = V[vd] * np.conj(Ybus[vd, :] @ V)
+
+        # Reactive power at the pv nodes
+        P = Sbus[pv].real
+        Q = (V[pv] * np.conj(Ybus[pv, :] @ V)).imag
+        Sbus[pv] = P + 1j * Q  # keep the original P injection and set the calculated reactive power
+
         # Branches current, loading, etc
-        If = Yf * V
-        It = Yt * V
+        Vf = V[calculation_inputs.branch_data.F]
+        Vt = V[calculation_inputs.branch_data.T]
+        If = Yf @ V
+        It = Yt @ V
         Sf = Vf * np.conj(If)
         St = Vt * np.conj(It)
 
@@ -515,16 +563,25 @@ def power_flow_post_process(calculation_inputs: NumericalCircuit,
 
     else:
         # DC power flow
-        theta_f = np.angle(Vf, deg=False)
-        theta_t = np.angle(Vt, deg=False)
-        Vbranch = theta_f - theta_t
-        Sf = (1.0 / calculation_inputs.branch_data.X) * Vbranch
-        Sfb = Sf * calculation_inputs.Sbase
-        Stb = -Sf * calculation_inputs.Sbase
-        If = Sf / (Vf + 1e-20)
+        theta = np.angle(V, deg=False)
+        theta_f = theta[calculation_inputs.F]
+        theta_t = theta[calculation_inputs.T]
+
+        b = 1.0 / (calculation_inputs.branch_data.X * calculation_inputs.branch_data.tap_module)
+        # Pf = calculation_inputs.Bf @ theta - b * calculation_inputs.branch_data.tap_angle
+
+        Pf = b * (theta_f - theta_t - calculation_inputs.branch_data.tap_angle)
+
+        Sfb = Pf * calculation_inputs.Sbase
+        Stb = -Pf * calculation_inputs.Sbase
+
+        Vf = V[calculation_inputs.branch_data.F]
+        Vt = V[calculation_inputs.branch_data.T]
+        Vbranch = Vf - Vt
+        If = Pf / (Vf + 1e-20)
         It = -If
         # losses are not considered in the power flow computation
-        losses = np.zeros(Sf.shape)  # If * If * calculation_inputs.branch_data.R * calculation_inputs.Sbase
+        losses = np.zeros(calculation_inputs.nbr)
 
     # Branch loading in p.u.
     loading = Sfb / (branch_rates + 1e-9)
@@ -536,7 +593,7 @@ def multi_island_pf_nc(nc: NumericalCircuit,
                        options: PowerFlowOptions,
                        logger=Logger(),
                        V_guess: Union[CxVec, None] = None,
-                       Sbus_input: Union[CxVec, None] = None) -> "PowerFlowResults":
+                       Sbus_input: Union[CxVec, None] = None) -> PowerFlowResults:
     """
     Multiple islands power flow (this is the most generic power flow function)
     :param nc: SnapshotData instance
@@ -592,28 +649,7 @@ def multi_island_pf_nc(nc: NumericalCircuit,
                 if Sbus_input is None:
                     Sbus = island.Sbus + Shvdc[island.original_bus_idx]
                 else:
-                    Sbus = Sbus_input + Shvdc[island.original_bus_idx]
-
-                # run circuit power flow
-                # res = single_island_pf(
-                #     circuit=island,
-                #     Vbus=island.Vbus if V_guess is None else V_guess[island.original_bus_idx],
-                #     Sbus=Sbus,
-                #     Ibus=island.Ibus,
-                #     Yloadbus=island.YLoadBus,
-                #     ma=island.branch_data.tap_module,
-                #     theta=island.branch_data.tap_angle,
-                #     Beq=island.branch_data.Beq,
-                #     branch_rates=island.Rates,
-                #     pq=island.pq,
-                #     pv=island.pv,
-                #     vd=island.vd,
-                #     pqpv=island.pqpv,
-                #     Qmin=island.Qmin_bus,
-                #     Qmax=island.Qmax_bus,
-                #     options=options,
-                #     logger=logger,
-                # )
+                    Sbus = (Sbus_input + Shvdc)[island.original_bus_idx]
 
                 res = single_island_pf(
                     circuit=island,
@@ -628,11 +664,14 @@ def multi_island_pf_nc(nc: NumericalCircuit,
                     branch_rates=island.Rates,
                     pq=island.pq,
                     pv=island.pv,
+                    pqv=island.pqv,
+                    p=island.p,
                     vd=island.vd,
                     pqpv=island.pqpv,
                     Qmin=island.Qmin_bus,
                     Qmax=island.Qmax_bus,
-                    logger=logger)
+                    logger=logger
+                )
 
                 # merge the results from this island
                 results.apply_from_island(
@@ -710,11 +749,11 @@ def multi_island_pf_nc(nc: NumericalCircuit,
 
 def multi_island_pf(multi_circuit: MultiCircuit,
                     options: PowerFlowOptions,
-                    opf_results: Union["OptimalPowerFlowResults", None] = None,
+                    opf_results: Union[OptimalPowerFlowResults, None] = None,
                     t: Union[int, None] = None,
                     logger: Logger = Logger(),
                     bus_dict: Union[Dict[Bus, int], None] = None,
-                    areas_dict: Union[Dict[Area, int], None] = None) -> "PowerFlowResults":
+                    areas_dict: Union[Dict[Area, int], None] = None) -> PowerFlowResults:
     """
     Multiple islands power flow (this is the most generic power flow function)
     :param multi_circuit: MultiCircuit instance
@@ -727,16 +766,34 @@ def multi_island_pf(multi_circuit: MultiCircuit,
     :return: PowerFlowResults instance
     """
 
-    nc = compile_numerical_circuit_at(
-        circuit=multi_circuit,
-        t_idx=t,
-        apply_temperature=options.apply_temperature_correction,
-        branch_tolerance_mode=options.branch_impedance_tolerance_mode,
-        opf_results=opf_results,
-        use_stored_guess=options.use_stored_guess,
-        bus_dict=bus_dict,
-        areas_dict=areas_dict
-    )
+    # Generalised PowerFlow
+    if options.generalised_pf:
+        nc = compile_numerical_circuit_at_generalised_pf(
+            circuit=multi_circuit,
+            t_idx=t,
+            apply_temperature=options.apply_temperature_correction,
+            branch_tolerance_mode=options.branch_impedance_tolerance_mode,
+            opf_results=opf_results,
+            use_stored_guess=options.use_stored_guess,
+            bus_dict=bus_dict,
+            areas_dict=areas_dict
+        )
+        # print("Generalised PowerFlow")
+
+    # Normal PowerFlow
+    else:
+        nc = compile_numerical_circuit_at(
+            circuit=multi_circuit,
+            t_idx=t,
+            apply_temperature=options.apply_temperature_correction,
+            branch_tolerance_mode=options.branch_impedance_tolerance_mode,
+            opf_results=opf_results,
+            use_stored_guess=options.use_stored_guess,
+            bus_dict=bus_dict,
+            areas_dict=areas_dict,
+            logger=logger,
+        )
+        # print("Normal PowerFlow")
 
     res = multi_island_pf_nc(nc=nc, options=options, logger=logger)
 
